#!/usr/bin/env python
# -*- coding: utf8 -*-

"""
Generated from file 4 of a prod VistA (with some manual amendments)

See webReportUtils: reduce4VISNVistASNO
"""
byVISNByVistAStation = {
        "1": {
            "402": [
                {
                    "name": "VA MAINE HCS",
                    "stationNumber": "402",
                    "officialName": "MAINE VA HCS",
                    "npi": "1184674509",
                    "visn": "1",
                    "ien": "402"
                },
                {
                    "name": "TOGUS MAINE NHCU",
                    "stationNumber": "4029AA",
                    "visn": "1",
                    "ien": 40201
                },
                {
                    "name": "MACHIAS STATE VETERANS HOME",
                    "stationNumber": "402DY",
                    "visn": "1",
                    "ien": 40202
                },
                {
                    "name": "CARIBOU MAINE CBOC",
                    "stationNumber": "402GA",
                    "npi": "1609829514",
                    "visn": "1",
                    "ien": 40203
                },
                {
                    "name": "CALAIS MAINE CBOC",
                    "stationNumber": "402GB",
                    "npi": "1528019130",
                    "visn": "1",
                    "ien": 40204
                },
                {
                    "name": "RUMFORD MAINE CBOC",
                    "stationNumber": "402GC",
                    "npi": "1982646360",
                    "visn": "1",
                    "ien": 40205
                },
                {
                    "name": "SACO MAINE CBOC",
                    "stationNumber": "402GD",
                    "npi": "1710929369",
                    "visn": "1",
                    "ien": 40206
                },
                {
                    "name": "LEWISTON AUBURN CBOC",
                    "stationNumber": "402GE",
                    "npi": "1770809691",
                    "visn": "1",
                    "ien": 40207
                },
                {
                    "name": "LINCOLN VA CLINIC",
                    "stationNumber": "402GF",
                    "visn": "1",
                    "ien": 40208
                },
                {
                    "name": "BANGOR MAINE CBOC",
                    "stationNumber": "402HB",
                    "npi": "1033160064",
                    "visn": "1",
                    "ien": 40209
                },
                {
                    "name": "PORTLAND MAINE MH ORC",
                    "stationNumber": "402HC",
                    "npi": "1114961992",
                    "visn": "1",
                    "ien": 40210
                },
                {
                    "name": "RURAL AND COMNTY HLTCARE MOC",
                    "stationNumber": "402HL",
                    "npi": "1942675103",
                    "visn": "1",
                    "ien": 40211
                },
                {
                    "name": "FORT KENT VA CLINIC",
                    "stationNumber": "402QA",
                    "npi": "1821496886",
                    "visn": "1",
                    "ien": 40212
                },
                {
                    "name": "HOULTON VA CLINIC",
                    "stationNumber": "402QB",
                    "npi": "1407254188",
                    "visn": "1",
                    "ien": 40213
                }
            ],
            "405": [
                {
                    "name": "WHITE RIVER JCT VAMROC",
                    "stationNumber": "405",
                    "npi": "1134179146",
                    "visn": "1",
                    "ien": "405"
                },
                {
                    "name": "BENNINGTON STATE NURSING HOME",
                    "stationNumber": "4059AF",
                    "visn": "1",
                    "ien": 40501
                },
                {
                    "name": "WHITE RIVER JUNCTION SARRTP",
                    "stationNumber": "405BU",
                    "npi": "1356706147",
                    "visn": "1",
                    "ien": 40502
                },
                {
                    "name": "BENNINGTON STATE HOME",
                    "stationNumber": "405DT",
                    "officialName": "BENNINGTON VERMONT STATE HOME",
                    "visn": "1",
                    "ien": 40503
                },
                {
                    "name": "BENNINGTON CBOC",
                    "stationNumber": "405GA",
                    "officialName": "BENNINGTON VERMONT OUTREACH CLINIC",
                    "npi": "1801847447",
                    "visn": "1",
                    "ien": 40504
                },
                {
                    "name": "BRATTLEBORO CBOC",
                    "stationNumber": "405GC",
                    "npi": "1760702765",
                    "visn": "1",
                    "ien": 40505
                },
                {
                    "name": "BURLINGTON LAKESIDE CBOC",
                    "stationNumber": "405HA",
                    "npi": "1184675795",
                    "visn": "1",
                    "ien": 40506
                },
                {
                    "name": "MONTPELIER ORC",
                    "stationNumber": "405HB",
                    "officialName": "MONTPELIER OUTREACH CLINIC",
                    "visn": "1",
                    "ien": 40507
                },
                {
                    "name": "ST. JOHNSBURY CBOC",
                    "stationNumber": "405HC",
                    "officialName": "ST. JOHNSBURY VERMONT OUTREACH CLINIC",
                    "npi": "1497794432",
                    "visn": "1",
                    "ien": 40508
                },
                {
                    "name": "KEENE CBOC",
                    "stationNumber": "405HE",
                    "officialName": "KEENE OUT REACH CENTER",
                    "npi": "1124397211",
                    "visn": "1",
                    "ien": 40509
                },
                {
                    "name": "RUTLAND CBOC",
                    "stationNumber": "405HF",
                    "officialName": "RUTLAND VERMONT OUTREACH CLINIC",
                    "npi": "1861435083",
                    "visn": "1",
                    "ien": 40510
                },
                {
                    "name": "WHITE RIVER JCT MORC",
                    "stationNumber": "405HK",
                    "officialName": "WHITE RIVER JUNCTION MORC",
                    "visn": "1",
                    "ien": 40511
                },
                {
                    "name": "ST JOHNSBURY VA CLINIC",
                    "stationNumber": "405QA",
                    "npi": "1316345093",
                    "visn": "1",
                    "ien": 40512
                },
                {
                    "name": "NEWPORT VA CLINIC",
                    "stationNumber": "405QB",
                    "npi": "1952709537",
                    "visn": "1",
                    "ien": 40513
                }
            ],
            "518": [
                {
                    "name": "EDITH NOURSE ROGERS VAMC",
                    "stationNumber": "518",
                    "npi": "1568412617",
                    "domain": "BEDFORD.MED.VA.GOV",
                    "visn": "1",
                    "ien": "518",
                    "mn": "BED"
                },
                {
                    "name": "BEDFORD NURSING HOME",
                    "stationNumber": "5189AA",
                    "visn": "1",
                    "ien": 51801
                },
                {
                    "name": "BEDFORD DOMICILIARY",
                    "stationNumber": "518BU",
                    "officialName": "BEDFORD DOMICILIARY 518BU",
                    "visn": "1",
                    "ien": 51802
                },
                {
                    "name": "LYNN VA CLINIC",
                    "stationNumber": "518GA",
                    "npi": "1073550588",
                    "visn": "1",
                    "ien": 51803
                },
                {
                    "name": "HAVERHILL VA CLINIC",
                    "stationNumber": "518GB",
                    "npi": "1750336194",
                    "visn": "1",
                    "ien": 51804
                },
                {
                    "name": "GLOUCESTER VA CLINIC",
                    "stationNumber": "518GE",
                    "npi": "1932153368",
                    "visn": "1",
                    "ien": 51805
                },
                {
                    "name": "FITCHBURG CBOC",
                    "stationNumber": "518GG",
                    "npi": "1629021506",
                    "visn": "1",
                    "ien": 51806
                },
                {
                    "name": "IDES HANSCOM AFB",
                    "stationNumber": "518MA",
                    "visn": "1",
                    "ien": 51807
                },
                {
                    "name": "BEDFORD-PRRTP",
                    "stationNumber": "518PA",
                    "visn": "1",
                    "ien": 51808
                }
            ],
            "523": [
                {
                    "name": "BOSTON HCS VAMC",
                    "stationNumber": "523",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - BOSTON DIVISION",
                    "npi": "1497851802",
                    "visn": "1",
                    "ien": "523"
                },
                {
                    "name": "WEST ROXBURY",
                    "stationNumber": "523A4",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - WEST ROXBURY DIVISION",
                    "npi": "1508815549",
                    "visn": "1",
                    "ien": 52301
                },
                {
                    "name": "BROCKTON VAMC",
                    "stationNumber": "523A5",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - BROCKTON DIVISION",
                    "npi": "1306894613",
                    "visn": "1",
                    "ien": 52302
                },
                {
                    "name": "BROCKTON DOM",
                    "stationNumber": "523BU",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - BROCKTON DOMICILARY",
                    "visn": "1",
                    "ien": 52303
                },
                {
                    "name": "BOSTON VA FACILITY DOM",
                    "stationNumber": "523BV",
                    "officialName": "BOSTON VA FACILITY DOMICILIARY",
                    "visn": "1",
                    "ien": 52304
                },
                {
                    "name": "LOWELL OPC",
                    "stationNumber": "523BY",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - LOWELL DIVISION",
                    "npi": "1447290176",
                    "visn": "1",
                    "ien": 52305
                },
                {
                    "name": "BOSTON SOC",
                    "stationNumber": "523BZ",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - BOSTON OPC DIVISION",
                    "npi": "1912958596",
                    "visn": "1",
                    "ien": 52306
                },
                {
                    "name": "CHELSEA",
                    "stationNumber": "523DT",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - CHELSEA SOLDIERS HOME",
                    "visn": "1",
                    "ien": 52307
                },
                {
                    "name": "FRAMINGHAM",
                    "stationNumber": "523GA",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - FRAMINGHAM CBOC",
                    "npi": "1184677213",
                    "visn": "1",
                    "ien": 52308
                },
                {
                    "name": "WORCESTER",
                    "stationNumber": "523GB",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - WORCESTER DIVISION",
                    "npi": "1447294707",
                    "visn": "1",
                    "ien": 52309
                },
                {
                    "name": "QUINCY",
                    "stationNumber": "523GC",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - QUINCY CBOC",
                    "npi": "1730127952",
                    "visn": "1",
                    "ien": 52310
                },
                {
                    "name": "PLYMOUTH CBOC",
                    "stationNumber": "523GD",
                    "npi": "1952650525",
                    "visn": "1",
                    "ien": 52311
                },
                {
                    "name": "IDES HANSCOM AFB",
                    "stationNumber": "523MA",
                    "visn": "1",
                    "ien": 52312
                },
                {
                    "name": "BOSTON PRRTP",
                    "stationNumber": "523PA",
                    "officialName": "VA BOSTON HEALTHCARE SYSTEM - BOSTON PRRTP",
                    "visn": "1",
                    "ien": 52313
                },
                {
                    "name": "BOSTON VA MOBILE CLINIC",
                    "stationNumber": "523QA",
                    "npi": "1861887721",
                    "visn": "1",
                    "ien": 52314
                },
                {
                    "name": "BOSTON SPINAL CORD OUTCOMES",
                    "stationNumber": "523SCI",
                    "officialName": "BOSTON SPINAL CORD OUTCOMES (SCI)",
                    "visn": "1",
                    "ien": 52315
                }
            ],
            "608": [
                {
                    "name": "MANCHESTER VAMC",
                    "stationNumber": "608",
                    "npi": "1477653731",
                    "visn": "1",
                    "ien": "608"
                },
                {
                    "name": "MANCHESTER NHC",
                    "stationNumber": "6089AA",
                    "visn": "1",
                    "ien": 60801
                },
                {
                    "name": "MANCHESTER STATE HOME",
                    "stationNumber": "6089AF",
                    "visn": "1",
                    "ien": 60802
                },
                {
                    "name": "TILTON",
                    "stationNumber": "608DT",
                    "visn": "1",
                    "ien": 60803
                },
                {
                    "name": "PORTSMOUTH CBOC",
                    "stationNumber": "608GA",
                    "npi": "1861439101",
                    "visn": "1",
                    "ien": 60804
                },
                {
                    "name": "SOMERSWORTH CBOC",
                    "stationNumber": "608GC",
                    "npi": "1154487585",
                    "visn": "1",
                    "ien": 60805
                },
                {
                    "name": "CONWAY CBOC",
                    "stationNumber": "608GD",
                    "npi": "1215989363",
                    "visn": "1",
                    "ien": 60806
                },
                {
                    "name": "TILTON CBOC",
                    "stationNumber": "608HA",
                    "npi": "1699716035",
                    "visn": "1",
                    "ien": 60807
                },
                {
                    "name": "IDES MANCHESTER VAMC",
                    "stationNumber": "608MA",
                    "visn": "1",
                    "ien": 60808
                }
            ],
            "631": [
                {
                    "name": "VA CNTRL WSTRN MASSCHUSETS HCS",
                    "stationNumber": "631",
                    "npi": "1780634162",
                    "domain": "NORTHAMPTON.MED.VA.GOV",
                    "visn": "1",
                    "ien": "631",
                    "mn": "NOR"
                },
                {
                    "name": "NORTHAMPTON",
                    "stationNumber": "6319AA",
                    "officialName": "Northampton NHC",
                    "visn": "1",
                    "ien": 63101
                },
                {
                    "name": "HOLYOKE (STNB)",
                    "stationNumber": "6319AF",
                    "officialName": "Holyoke (STNB)",
                    "visn": "1",
                    "ien": 63102
                },
                {
                    "name": "SPRINGFIELD",
                    "stationNumber": "631BY",
                    "officialName": "SPRINGFIELD OPC",
                    "npi": "1982642823",
                    "visn": "1",
                    "ien": 63103
                },
                {
                    "name": "HOLYOKE (STNB)",
                    "stationNumber": "631DT",
                    "officialName": "HOLYOKE SOLDIERS NURSING HOME",
                    "visn": "1",
                    "ien": 63104
                },
                {
                    "name": "HOLYOKE (STATE DOM)",
                    "stationNumber": "631EL",
                    "officialName": "HOLYOKE SOLDIERS HOME DOMICILIARY",
                    "visn": "1",
                    "ien": 63105
                },
                {
                    "name": "PITTSFIELD CBOC",
                    "stationNumber": "631GC",
                    "npi": "1033156518",
                    "visn": "1",
                    "ien": 63106
                },
                {
                    "name": "GREENFIELD (CBOC)",
                    "stationNumber": "631GD",
                    "officialName": "GREENFIELD CBOC",
                    "npi": "1083767503",
                    "visn": "1",
                    "ien": 63107
                },
                {
                    "name": "WORCESTER CBOC ",
                    "stationNumber": "631GE",
                    "officialName": "WORCESTER CBOC",
                    "npi": "1669766564",
                    "visn": "1",
                    "ien": 63108
                },
                {
                    "name": "FITCHBURG CBOC",
                    "stationNumber": "631GF",
                    "npi": "1659665537",
                    "visn": "1",
                    "ien": 63109
                },
                {
                    "name": "IDES KELLER ARMY COMNTY HOSPTL",
                    "stationNumber": "631MA",
                    "visn": "1",
                    "ien": 63110
                },
                {
                    "name": "NORTH-MAIN-MASSCHST-PRRTP",
                    "stationNumber": "631PA",
                    "visn": "1",
                    "ien": 63111
                },
                {
                    "name": "PLANTATION STREET VA CLINIC",
                    "stationNumber": "631QA",
                    "npi": "1417301714",
                    "visn": "1",
                    "ien": 63112
                },
                {
                    "name": "LAKE AVENUE VA CLINIC",
                    "stationNumber": "631QB",
                    "npi": "1407399298",
                    "visn": "1",
                    "ien": 63113
                }
            ],
            "650": [
                {
                    "name": "PROVIDENCE VAMC",
                    "stationNumber": "650",
                    "npi": "1306896733",
                    "domain": "PROVIDENCE.MED.VA.GOV",
                    "visn": "1",
                    "ien": "650",
                    "mn": "PRO"
                },
                {
                    "name": "BRISTOL",
                    "stationNumber": "650DT",
                    "officialName": "BRISTOL SVH",
                    "visn": "1",
                    "ien": 65001
                },
                {
                    "name": "NEW BEDFORD",
                    "stationNumber": "650GA",
                    "officialName": "NEW BEDFORD CBOC",
                    "npi": "1225076664",
                    "visn": "1",
                    "ien": 65002
                },
                {
                    "name": "HYANNIS",
                    "stationNumber": "650GB",
                    "officialName": "HYANNIS CBOC",
                    "npi": "1902851306",
                    "visn": "1",
                    "ien": 65003
                },
                {
                    "name": "MIDDLETOWN",
                    "stationNumber": "650GD",
                    "officialName": "MIDDLETOWN CBOC",
                    "npi": "1831136050",
                    "visn": "1",
                    "ien": 65004
                },
                {
                    "name": "NEWPORT NAVAL BASE",
                    "stationNumber": "650MA",
                    "visn": "1",
                    "ien": 65005
                },
                {
                    "name": "EAGLE SQUARE VA CLINIC",
                    "stationNumber": "650QA",
                    "npi": "1679971311",
                    "visn": "1",
                    "ien": 65006
                },
                {
                    "name": "EAGLE STREET VA CLINIC",
                    "stationNumber": "650QB",
                    "npi": "1710461215",
                    "visn": "1",
                    "ien": 65007
                }
            ],
            "689": [
                {
                    "name": "CONNECTICUT HCS",
                    "stationNumber": "689",
                    "officialName": "VA CONNECTICUT HEALTHCARE SYSTEM - WEST HAVEN DIVISION",
                    "npi": "1245289610",
                    "domain": "WEST-HAVEN.MED.VA.GOV",
                    "visn": "1",
                    "ien": "689",
                    "mn": "WES"
                },
                {
                    "name": "WEST HAVEN NHC",
                    "stationNumber": "6899AA",
                    "visn": "1",
                    "ien": 68901
                },
                {
                    "name": "NEWINGTON",
                    "stationNumber": "689A4",
                    "npi": "1871787150",
                    "visn": "1",
                    "ien": 68902
                },
                {
                    "name": "WEST HAVEN VA FACILITY DOM",
                    "stationNumber": "689BU",
                    "officialName": "WEST HAVEN VA FACILITY DOMICILARY",
                    "visn": "1",
                    "ien": 68903
                },
                {
                    "name": "NEWINGTON VA FACILITY DOM",
                    "stationNumber": "689BV",
                    "officialName": "NEWINGTON VA FACILITY DOMICILIARY",
                    "visn": "1",
                    "ien": 68904
                },
                {
                    "name": "ROCKY HILL",
                    "stationNumber": "689DT",
                    "visn": "1",
                    "ien": 68905
                },
                {
                    "name": "ROCKY HILL",
                    "stationNumber": "689EL",
                    "visn": "1",
                    "ien": 68906
                },
                {
                    "name": "WATERBURY",
                    "stationNumber": "689GA",
                    "npi": "1952345068",
                    "visn": "1",
                    "ien": 68907
                },
                {
                    "name": "STAMFORD VA PRIMARY CARE CTR",
                    "stationNumber": "689GB",
                    "npi": "1255484796",
                    "visn": "1",
                    "ien": 68908
                },
                {
                    "name": "Willimantic VA Clinic",
                    "stationNumber": "689GC",
                    "npi": "1760421630",
                    "visn": "1",
                    "ien": 68909
                },
                {
                    "name": "WINSTED",
                    "stationNumber": "689GD",
                    "npi": "1063452878",
                    "visn": "1",
                    "ien": 68910
                },
                {
                    "name": "DANBURY CBOC",
                    "stationNumber": "689GE",
                    "officialName": "DANBURY",
                    "npi": "1285686147",
                    "visn": "1",
                    "ien": 68911
                },
                {
                    "name": "NEW LONDON",
                    "stationNumber": "689HC",
                    "npi": "1942247820",
                    "visn": "1",
                    "ien": 68912
                },
                {
                    "name": "WEST HAVEN MORC",
                    "stationNumber": "689HK",
                    "visn": "1",
                    "ien": 68913
                },
                {
                    "name": "IDES SUB BASE NEW LONDON",
                    "stationNumber": "689MA",
                    "visn": "1",
                    "ien": 68914
                },
                {
                    "name": "WH-PRRTP",
                    "stationNumber": "689PA",
                    "visn": "1",
                    "ien": 68915
                },
                {
                    "name": "NEW HAVEN CWT/TR",
                    "stationNumber": "689PB",
                    "visn": "1",
                    "ien": 68916
                },
                {
                    "name": "ERRERA VA CLINIC",
                    "stationNumber": "689QA",
                    "npi": "1730587478",
                    "visn": "1",
                    "ien": 68917
                }
            ]
        },
        "19": {
            "436": [
                {
                    "name": "FORT HARRISON MEDICAL CENTER",
                    "stationNumber": "436",
                    "npi": "1144279910",
                    "visn": "19",
                    "ien": "436"
                },
                {
                    "name": "MONTANA HCS VA NURSING HOME",
                    "stationNumber": "4369AA",
                    "visn": "19",
                    "ien": 43601
                },
                {
                    "name": "COLUMBIA FALLS ST NURSING HOME",
                    "stationNumber": "4369AF",
                    "visn": "19",
                    "ien": 43602
                },
                {
                    "name": "MILES CITY VA MEDICAL CENTER",
                    "stationNumber": "436A4",
                    "npi": "1053728014",
                    "visn": "19",
                    "ien": 43603
                },
                {
                    "name": "FORT HARRISON DOM",
                    "stationNumber": "436BU",
                    "visn": "19",
                    "ien": 43604
                },
                {
                    "name": "FT. HARRISON VET STATE HOME",
                    "stationNumber": "436DT",
                    "officialName": "FT HARRISON VET STATE HOME",
                    "visn": "19",
                    "ien": 43605
                },
                {
                    "name": "FT. HARRISON STATE HOME",
                    "stationNumber": "436DU",
                    "officialName": "FT HARRISON STATE HOME",
                    "visn": "19",
                    "ien": 43606
                },
                {
                    "name": "ANACONDA CBOC",
                    "stationNumber": "436GA",
                    "npi": "1063463222",
                    "visn": "19",
                    "ien": 43607
                },
                {
                    "name": "GREAT FALLS CBOC",
                    "stationNumber": "436GB",
                    "npi": "1518912120",
                    "visn": "19",
                    "ien": 43608
                },
                {
                    "name": "DAVID J. THATCHER VA CLINIC",
                    "stationNumber": "436GC",
                    "npi": "1245277391",
                    "visn": "19",
                    "ien": 43609
                },
                {
                    "name": "BOZEMAN VA CLINIC",
                    "stationNumber": "436GD",
                    "npi": "1720039464",
                    "visn": "19",
                    "ien": 43610
                },
                {
                    "name": "KALISPELL VA CLINIC",
                    "stationNumber": "436GF",
                    "npi": "1376599787",
                    "visn": "19",
                    "ien": 43611
                },
                {
                    "name": "BENJAMIN C. STEELE VA CLINIC",
                    "stationNumber": "436GH",
                    "npi": "1639122823",
                    "visn": "19",
                    "ien": 43612
                },
                {
                    "name": "GLASGOW VA CLINIC",
                    "stationNumber": "436GI",
                    "npi": "1285688770",
                    "visn": "19",
                    "ien": 43613
                },
                {
                    "name": "GLENDIVE VA CLINIC",
                    "stationNumber": "436GK",
                    "npi": "1386682433",
                    "visn": "19",
                    "ien": 43614
                },
                {
                    "name": "CUT BANK VA CLINIC",
                    "stationNumber": "436GL",
                    "npi": "1487831392",
                    "visn": "19",
                    "ien": 43615
                },
                {
                    "name": "LEWISTOWN CBOC",
                    "stationNumber": "436GM",
                    "officialName": "LEWISTON CBOC",
                    "npi": "1124296991",
                    "visn": "19",
                    "ien": 43616
                },
                {
                    "name": "DR. J MEDICINE CROW VA CLINIC",
                    "stationNumber": "436GN",
                    "npi": "1770040172",
                    "visn": "19",
                    "ien": 43617
                },
                {
                    "name": "MERRIL LUNDMAN VA OPC",
                    "stationNumber": "436HC",
                    "npi": "1326403650",
                    "visn": "19",
                    "ien": 43618
                },
                {
                    "name": "IDES MALMSTROM AFB",
                    "stationNumber": "436MA",
                    "visn": "19",
                    "ien": 43619
                },
                {
                    "name": "HAMILTON VA CLINIC",
                    "stationNumber": "436QA",
                    "npi": "1346638640",
                    "visn": "19",
                    "ien": 43620
                },
                {
                    "name": "PLENTYWOOD VA CLINIC",
                    "stationNumber": "436QB",
                    "npi": "1518355817",
                    "visn": "19",
                    "ien": 43621
                },
                {
                    "name": "HELENA VA CLINIC",
                    "stationNumber": "436QC",
                    "npi": "1801271127",
                    "visn": "19",
                    "ien": 43622
                },
                {
                    "name": "BROWNING VA CLINIC",
                    "stationNumber": "436QD",
                    "npi": "1346870714",
                    "visn": "19",
                    "ien": 43623
                }
            ],
            "442": [
                {
                    "name": "CHEYENNE VA MEDICAL",
                    "stationNumber": "442",
                    "npi": "1164471991",
                    "visn": "19",
                    "ien": "442"
                },
                {
                    "name": "CHEYENNE NHCU",
                    "stationNumber": "4429AA",
                    "visn": "19",
                    "ien": 44201
                },
                {
                    "name": "CHEYENNE VA DOMICILIARY",
                    "stationNumber": "442BU",
                    "npi": "1013448273",
                    "visn": "19",
                    "ien": 44202
                },
                {
                    "name": "SCOTTSBLUFF ANNEX",
                    "stationNumber": "442DT",
                    "officialName": "SCOTTSBLUFF STNB",
                    "visn": "19",
                    "ien": 44203
                },
                {
                    "name": "SIDNEY VA CLINC",
                    "stationNumber": "442GB",
                    "npi": "1891733945",
                    "visn": "19",
                    "ien": 44204
                },
                {
                    "name": "FORT COLLINS VA CLINIC",
                    "stationNumber": "442GC",
                    "npi": "1528013083",
                    "visn": "19",
                    "ien": 44205
                },
                {
                    "name": "LOVELAND VA CLINIC",
                    "stationNumber": "442GD",
                    "npi": "1194770107",
                    "visn": "19",
                    "ien": 44206
                },
                {
                    "name": "WHEATLAND VA MOBILE CLINIC",
                    "stationNumber": "442HK",
                    "npi": "1073995734",
                    "visn": "19",
                    "ien": 44207
                },
                {
                    "name": "IDES F.E. WARREN AFB",
                    "stationNumber": "442MA",
                    "visn": "19",
                    "ien": 44208
                },
                {
                    "name": "IDES SHERIDAN VAMC",
                    "stationNumber": "442MB",
                    "visn": "19",
                    "ien": 44209
                },
                {
                    "name": "RAWLINS VA CLINIC",
                    "stationNumber": "442QA",
                    "npi": "1720477904",
                    "visn": "19",
                    "ien": 44210
                },
                {
                    "name": "TORRINGTON VA MOBILE CLINIC",
                    "stationNumber": "442QB",
                    "npi": "1548659733",
                    "visn": "19",
                    "ien": 44211
                },
                {
                    "name": "WHEATLAND FRONTRANGE 3 MOBILE",
                    "stationNumber": "442QC",
                    "officialName": "WHEATLAND FRONTRANGE 3 MOB CLN",
                    "npi": "1114477197",
                    "ien": 44212,
                    "visn": "19"
                },
                {
                    "name": "LARAMIE VA MOBILE CLINIC",
                    "stationNumber": "442QD",
                    "npi": "1922586692",
                    "visn": "19",
                    "ien": 44213
                },
                {
                    "name": "STERLING VA CLINIC",
                    "stationNumber": "442QE",
                    "npi": "1063990737",
                    "visn": "19",
                    "ien": 44214
                }
            ],
            "554": [
                {
                    "name": "ROCKY MOUNTAIN REGIONAL VAMC",
                    "stationNumber": "554",
                    "npi": "1568418333",
                    "domain": "DENVER.MED.VA.GOV",
                    "visn": "19",
                    "ien": "554",
                    "mn": "DEN"
                },
                {
                    "name": "DENVER NHC",
                    "stationNumber": "5549AA",
                    "visn": "19",
                    "ien": 55401
                },
                {
                    "name": "PUEBLO NHCU",
                    "stationNumber": "5549AB",
                    "officialName": "PUEBLO NURSING HOME CARE UNIT",
                    "visn": "19",
                    "ien": 55402
                },
                {
                    "name": "HAOME LAKE STATE HOME",
                    "stationNumber": "5549AF",
                    "visn": "19",
                    "ien": 55403
                },
                {
                    "name": "FLORENCE STATE HOME",
                    "stationNumber": "5549AG",
                    "visn": "19",
                    "ien": 55404
                },
                {
                    "name": "RIFLE STATE HOME",
                    "stationNumber": "5549AH",
                    "visn": "19",
                    "ien": 55405
                },
                {
                    "name": "PUEBLO VA CLC",
                    "stationNumber": "554A4",
                    "npi": "1750741211",
                    "visn": "19",
                    "ien": 55406
                },
                {
                    "name": "DENVER VA DOMICILLIARY (554A5)",
                    "stationNumber": "554A5",
                    "npi": "1891284964",
                    "visn": "19",
                    "ien": 55407
                },
                {
                    "name": "VALOR POINT VA DOMICILIARY",
                    "stationNumber": "554BU",
                    "npi": "1376903740",
                    "visn": "19",
                    "ien": 55408
                },
                {
                    "name": "DENVER VA DOMICILLIARY (554BV)",
                    "stationNumber": "554BV",
                    "visn": "19",
                    "ien": 55409
                },
                {
                    "name": "DENVER (DENT)",
                    "stationNumber": "554BY",
                    "visn": "19",
                    "ien": 55410
                },
                {
                    "name": "EVANS ARMY COMMUNITY HOSPITAL",
                    "stationNumber": "554CP",
                    "visn": "19",
                    "ien": 55411
                },
                {
                    "name": "PETERSON AFB,CO",
                    "stationNumber": "554CZ",
                    "visn": "19",
                    "ien": 55412
                },
                {
                    "name": "HOMELAKE STATE VETS CENTER",
                    "stationNumber": "554DT",
                    "visn": "19",
                    "ien": 55413
                },
                {
                    "name": "FLORENCE STATE VETS NURS. HOME",
                    "stationNumber": "554DU",
                    "visn": "19",
                    "ien": 55414
                },
                {
                    "name": "RIFLE STATE VETS NURSING HOME",
                    "stationNumber": "554DV",
                    "visn": "19",
                    "ien": 55415
                },
                {
                    "name": "WALSENBURG STATE VETS NH",
                    "stationNumber": "554DW",
                    "visn": "19",
                    "ien": 55416
                },
                {
                    "name": "AURORA VA CLINIC",
                    "stationNumber": "554GB",
                    "npi": "1851342042",
                    "visn": "19",
                    "ien": 55417
                },
                {
                    "name": "GOLDEN VA CLINIC",
                    "stationNumber": "554GC",
                    "npi": "1992752877",
                    "visn": "19",
                    "ien": 55418
                },
                {
                    "name": "PUEBLO VA CLINIC",
                    "stationNumber": "554GD",
                    "officialName": "PFC JAMES DUNN VA CLINIC",
                    "npi": "1205874526",
                    "visn": "19",
                    "ien": 55419
                },
                {
                    "name": "FLOYD K. LINDSTROM VA CLINIC",
                    "stationNumber": "554GE",
                    "npi": "1578515631",
                    "visn": "19",
                    "ien": 55420
                },
                {
                    "name": "ALAMOSA VA CLINIC",
                    "stationNumber": "554GF",
                    "npi": "1003866872",
                    "visn": "19",
                    "ien": 55421
                },
                {
                    "name": "LA JUNTA CBOC",
                    "stationNumber": "554GG",
                    "officialName": "LA JUNTA COMMUNITY BASED OUTPATIENT CLINIC (554GG)",
                    "npi": "1023065067",
                    "visn": "19",
                    "ien": 55422
                },
                {
                    "name": "LAMAR VA CLINIC",
                    "stationNumber": "554GH",
                    "npi": "1447207345",
                    "visn": "19",
                    "ien": 55423
                },
                {
                    "name": "BURLINGTON VA CLINIC",
                    "stationNumber": "554GI",
                    "npi": "1922297266",
                    "visn": "19",
                    "ien": 55424
                },
                {
                    "name": "DENVER VA CLINIC (9TH AVE)",
                    "stationNumber": "554GJ",
                    "npi": "1497307516",
                    "visn": "19",
                    "ien": 55425
                },
                {
                    "name": "IDES FORT CARSON",
                    "stationNumber": "554MA",
                    "visn": "19",
                    "ien": 55426
                },
                {
                    "name": "IDES US AFB ACADEMY",
                    "stationNumber": "554MB",
                    "visn": "19",
                    "ien": 55427
                },
                {
                    "name": "IDES BUCKLEY AFB",
                    "stationNumber": "554MC",
                    "visn": "19",
                    "ien": 55428
                },
                {
                    "name": "IDES PETERSON AFB",
                    "stationNumber": "554MD",
                    "visn": "19",
                    "ien": 55429
                },
                {
                    "name": "DENVER",
                    "stationNumber": "554PA",
                    "officialName": "DENVER PRRTP",
                    "visn": "19",
                    "ien": 55430
                },
                {
                    "name": "DENVER VA CLINIC (YORK ST)",
                    "stationNumber": "554QA",
                    "officialName": "DENVER VA CLINIC",
                    "npi": "1104217520",
                    "visn": "19",
                    "ien": 55431
                },
                {
                    "name": "JEWELL VA CLINIC",
                    "stationNumber": "554QB",
                    "npi": "1235520651",
                    "visn": "19",
                    "ien": 55432
                },
                {
                    "name": "SALIDA VA CLINIC",
                    "stationNumber": "554QC",
                    "npi": "1003207424",
                    "visn": "19",
                    "ien": 55433
                }
            ],
            "575": [
                {
                    "name": "GRAND JUNCTION VAMC",
                    "stationNumber": "575",
                    "npi": "1407806441",
                    "domain": "GRAND-JUNCT.MED.VA.GOV",
                    "visn": "19",
                    "ien": "575",
                    "mn": "GRA"
                },
                {
                    "name": "GRAND JUNCTION (NHCU)",
                    "stationNumber": "5759AA",
                    "visn": "19",
                    "ien": 57501
                },
                {
                    "name": "MONTROSE CBOC",
                    "stationNumber": "575GA",
                    "npi": "1144268731",
                    "visn": "19",
                    "ien": 57502
                },
                {
                    "name": "CRAIG TELEHEALTH OUTRCH CLINIC",
                    "stationNumber": "575GB",
                    "npi": "1043496169",
                    "visn": "19",
                    "ien": 57503
                },
                {
                    "name": "GLENWOOD SPRINGS VA CLINIC",
                    "stationNumber": "575QA",
                    "npi": "1902297328",
                    "visn": "19",
                    "ien": 57504
                },
                {
                    "name": "MOAB VA CLINIC",
                    "stationNumber": "575QB",
                    "npi": "1538550959",
                    "visn": "19",
                    "ien": 57505
                },
                {
                    "name": "GRAND JUNCTION VA MOBILE CLIN",
                    "stationNumber": "575QC",
                    "npi": "1386158897",
                    "visn": "19",
                    "ien": 57506
                }
            ],
            "623": [
                {
                    "name": "JACK C. MONTGOMERY VAMC",
                    "stationNumber": "623",
                    "officialName": "MUSKOGEE, OK VAMC",
                    "npi": "1598714925",
                    "domain": "MUSKOGEE.MED.VA.GOV",
                    "visn": "19",
                    "ien": "623",
                    "mn": "MUS"
                },
                {
                    "name": "MUSKOGEE NHC",
                    "stationNumber": "6239AA",
                    "visn": "19",
                    "ien": 62301
                },
                {
                    "name": "TULSA OUTPATIENT CLINIC",
                    "stationNumber": "623BY",
                    "officialName": "TULSA, OK CBOC",
                    "npi": "1750320776",
                    "visn": "19",
                    "ien": 62302
                },
                {
                    "name": "CLAREMORE VETERANS CENTER",
                    "stationNumber": "623DT",
                    "visn": "19",
                    "ien": 62303
                },
                {
                    "name": "TALIHINA VETERANS CENTER",
                    "stationNumber": "623DU",
                    "visn": "19",
                    "ien": 62304
                },
                {
                    "name": "MCALESTER VA CLINIC",
                    "stationNumber": "623GA",
                    "npi": "1316999410",
                    "visn": "19",
                    "ien": 62305
                },
                {
                    "name": "VINITA CBOC",
                    "stationNumber": "623GB",
                    "npi": "1740517275",
                    "visn": "19",
                    "ien": 62306
                },
                {
                    "name": "MCCURTAIN COUNTY VA CLINIC",
                    "stationNumber": "623GC",
                    "npi": "1851815294",
                    "visn": "19",
                    "ien": 62307
                },
                {
                    "name": "JACK C. MONTGOMERY EAST VA",
                    "stationNumber": "623QA",
                    "officialName": "JACK C. MONTGOMERY EAST VA CLINIC",
                    "npi": "1326436676",
                    "visn": "19",
                    "ien": 62308
                },
                {
                    "name": "TULSA ELEVENTH ST VA CLINIC",
                    "stationNumber": "623QB",
                    "npi": "1316335664",
                    "visn": "19",
                    "ien": 62309
                },
                {
                    "name": "YALE AVENUE VA CLINIC",
                    "stationNumber": "623QC",
                    "npi": "1184116857",
                    "ien": 62310,
                    "visn": "19"
                }
            ],
            "635": [
                {
                    "name": "OKLAHOMA CITY VAMC",
                    "stationNumber": "635",
                    "officialName": "OKLAHOMA CITY VA MEDICAL CENTER",
                    "npi": "1184673550",
                    "domain": "OKLAHOMA.MED.VA.GOV",
                    "visn": "19",
                    "ien": "635",
                    "mn": "OKL"
                },
                {
                    "name": "OKLAHOMA CITY (NHCU)",
                    "stationNumber": "6359AA",
                    "visn": "19",
                    "ien": 63501
                },
                {
                    "name": "ARDMORE (SVH)",
                    "stationNumber": "635DT",
                    "officialName": "ARDMORE CENTER",
                    "visn": "19",
                    "ien": 63502
                },
                {
                    "name": "NORMAN (SVH)",
                    "stationNumber": "635DU",
                    "officialName": "NORMAN CENTER",
                    "visn": "19",
                    "ien": 63503
                },
                {
                    "name": "SULPHUR (SVH)",
                    "stationNumber": "635DV",
                    "officialName": "SULPHUR CENTER",
                    "visn": "19",
                    "ien": 63504
                },
                {
                    "name": "CLINTON (SVH)",
                    "stationNumber": "635DW",
                    "officialName": "CLINTON CENTER",
                    "visn": "19",
                    "ien": 63505
                },
                {
                    "name": "LAWTON FT. SILL VET CENTER",
                    "stationNumber": "635DZ",
                    "visn": "19",
                    "ien": 63506
                },
                {
                    "name": "LAWTON (OPC)",
                    "stationNumber": "635GA",
                    "officialName": "LAWTON/FT. SILL VA OUTPATIENT CLINIC",
                    "npi": "1578506382",
                    "visn": "19",
                    "ien": 63507
                },
                {
                    "name": "WICHITA FALLS CBOC",
                    "stationNumber": "635GB",
                    "officialName": "WICHITA FALLS CBOC VETERANS CLINIC OF NORTH TEXAS",
                    "npi": "1275573016",
                    "visn": "19",
                    "ien": 63508
                },
                {
                    "name": "BLACKWELL CBOC",
                    "stationNumber": "635GC",
                    "npi": "1639118060",
                    "visn": "19",
                    "ien": 63509
                },
                {
                    "name": "ADA CBOC",
                    "stationNumber": "635GD",
                    "npi": "1982738852",
                    "visn": "19",
                    "ien": 63510
                },
                {
                    "name": "STILLWATER CBOC ",
                    "stationNumber": "635GE",
                    "officialName": "STILLWATER CBOC",
                    "npi": "1831329184",
                    "visn": "19",
                    "ien": 63511
                },
                {
                    "name": "ALTUS CBOC ",
                    "stationNumber": "635GF",
                    "officialName": "ALTUS CBOC",
                    "npi": "1285864546",
                    "visn": "19",
                    "ien": 63512
                },
                {
                    "name": "ENID CBOC ",
                    "stationNumber": "635GG",
                    "officialName": "ENID CBOC",
                    "npi": "1710117072",
                    "visn": "19",
                    "ien": 63513
                },
                {
                    "name": "CLINTON VA CLINIC (635GH)",
                    "stationNumber": "635GH",
                    "npi": "1215572656",
                    "visn": "19",
                    "ien": 63514
                },
                {
                    "name": "ARDMORE (ORC)",
                    "stationNumber": "635HB",
                    "officialName": "ARDMORE VETERANS CENTER CLINIC",
                    "npi": "1780635128",
                    "visn": "19",
                    "ien": 63515
                },
                {
                    "name": "IDES SHEPPARD AFB",
                    "stationNumber": "635MA",
                    "visn": "19",
                    "ien": 63516
                },
                {
                    "name": "IDES TINKER AFB",
                    "stationNumber": "635MB",
                    "visn": "19",
                    "ien": 63517
                },
                {
                    "name": "IDES VANCE AFB",
                    "stationNumber": "635MC",
                    "visn": "19",
                    "ien": 63518
                },
                {
                    "name": "IDES FORT SILL",
                    "stationNumber": "635MD",
                    "visn": "19",
                    "ien": 63519
                },
                {
                    "name": "IDES ALTUS AFB",
                    "stationNumber": "635ME",
                    "visn": "19",
                    "ien": 63520
                },
                {
                    "name": "OKLAHOMA CITY (PRRTP)",
                    "stationNumber": "635PA",
                    "officialName": "OKLAHOMA CITY PRRTP",
                    "visn": "19",
                    "ien": 63521
                },
                {
                    "name": "NORTH MAY VA CLINIC",
                    "stationNumber": "635QA",
                    "npi": "1811385172",
                    "visn": "19",
                    "ien": 63522
                },
                {
                    "name": "SOUTH OKLAHOMA CITY VA CLINIC",
                    "stationNumber": "635QB",
                    "npi": "1699135079",
                    "visn": "19",
                    "ien": 63523
                },
                {
                    "name": "FOURTEENTH STREET VA CLINIC",
                    "stationNumber": "635QC",
                    "npi": "1104308824",
                    "visn": "19",
                    "ien": 63524
                },
                {
                    "name": "LAWTON NORTH VA CLINIC",
                    "stationNumber": "635QD",
                    "npi": "1720615610",
                    "visn": "19",
                    "ien": 63525
                },
                {
                    "name": "TINKER VA CLINIC",
                    "stationNumber": "635QE",
                    "npi": "1598391716",
                    "visn": "19",
                    "ien": 63526
                }
            ],
            "660": [
                {
                    "name": "GEORGE E. WAHLEN VAMC",
                    "stationNumber": "660",
                    "npi": "1043260862",
                    "domain": "SALT-LAKE.MED.VA.GOV",
                    "visn": "19",
                    "ien": "660",
                    "mn": "SAL"
                },
                {
                    "name": "SALT LAKE CITY NHC",
                    "stationNumber": "6609AA",
                    "visn": "19",
                    "ien": 66001
                },
                {
                    "name": "SALT LAKE CITY VA FACILITY DOM",
                    "stationNumber": "660BU",
                    "officialName": "SALT LAKE CITY VA FACILITY DOMICILIARY",
                    "visn": "19",
                    "ien": 66002
                },
                {
                    "name": "SALT LAKE CITY",
                    "stationNumber": "660CZ",
                    "officialName": "HILL AIR FORCE BASE",
                    "visn": "19",
                    "ien": 66003
                },
                {
                    "name": "POCATELLO STNB",
                    "stationNumber": "660DT",
                    "officialName": "Idaho State Veterans Home ",
                    "visn": "19",
                    "ien": 66004
                },
                {
                    "name": "UTAH",
                    "stationNumber": "660DU",
                    "officialName": "UTAH STATE VETERANS NURSING HOME (660DU)",
                    "visn": "19",
                    "ien": 66005
                },
                {
                    "name": "OGDEN STATE VET NURSING FAC",
                    "stationNumber": "660DV",
                    "visn": "19",
                    "ien": 66006
                },
                {
                    "name": "POCATELLO VA CLINIC",
                    "stationNumber": "660GA",
                    "npi": "1487691002",
                    "visn": "19",
                    "ien": 66007
                },
                {
                    "name": "OGDEN VA CLINIC",
                    "stationNumber": "660GB",
                    "npi": "1407894868",
                    "visn": "19",
                    "ien": 66008
                },
                {
                    "name": "ROOSEVELT VA CLINIC",
                    "stationNumber": "660GD",
                    "npi": "1124062898",
                    "visn": "19",
                    "ien": 66009
                },
                {
                    "name": "OREM VA CLINIC",
                    "stationNumber": "660GE",
                    "npi": "1093753675",
                    "visn": "19",
                    "ien": 66010
                },
                {
                    "name": "ST. GEORGE VA CLINIC",
                    "stationNumber": "660GG",
                    "npi": "1184664559",
                    "visn": "19",
                    "ien": 66011
                },
                {
                    "name": "WESTERN SALT LAKE VA CLINIC",
                    "stationNumber": "660GJ",
                    "npi": "1619113115",
                    "visn": "19",
                    "ien": 66012
                },
                {
                    "name": "ELKO VA CLINIC",
                    "stationNumber": "660GK",
                    "npi": "1295126621",
                    "visn": "19",
                    "ien": 66013
                },
                {
                    "name": "IDES HILL AFB",
                    "stationNumber": "660MA",
                    "visn": "19",
                    "ien": 66014
                },
                {
                    "name": "SALT LAKE CITY PRRTP",
                    "stationNumber": "660PA",
                    "officialName": "VA SALT LAKE CITY PRRTP (660PA)",
                    "visn": "19",
                    "ien": 66015
                },
                {
                    "name": "IDAHO FALLS VA CLINIC",
                    "stationNumber": "660QA",
                    "npi": "1497146823",
                    "visn": "19",
                    "ien": 66016
                },
                {
                    "name": "PRICE VA CLINIC",
                    "stationNumber": "660QB",
                    "npi": "1124419551",
                    "visn": "19",
                    "ien": 66017
                },
                {
                    "name": "WEBER COUNTY VA CLINIC",
                    "stationNumber": "660QC",
                    "npi": "1356812655",
                    "visn": "19",
                    "ien": 66018
                }
            ],
            "666": [
                {
                    "name": "SHERIDAN VA MEDICAL CENTER",
                    "stationNumber": "666",
                    "npi": "1396795845",
                    "domain": "SHERIDAN.MED.VA.GOV",
                    "visn": "19",
                    "ien": "666",
                    "mn": "SHE"
                },
                {
                    "name": "SHERIDAN NHCU",
                    "stationNumber": "6669AA",
                    "visn": "19",
                    "ien": 66601
                },
                {
                    "name": "SUBSIDIARY VA DRRTP",
                    "stationNumber": "666BU",
                    "visn": "19",
                    "ien": 66602
                },
                {
                    "name": "BUFFALO, WY VET HOME",
                    "stationNumber": "666DT",
                    "officialName": "VETERAN HOME OF WYOMING",
                    "visn": "19",
                    "ien": 66603
                },
                {
                    "name": "CASPER, WY CBOC",
                    "stationNumber": "666GB",
                    "officialName": "CASPER CBOC",
                    "npi": "1538110226",
                    "visn": "19",
                    "ien": 66604
                },
                {
                    "name": "RIVERTON VA CLINIC",
                    "stationNumber": "666GC",
                    "npi": "1457399412",
                    "visn": "19",
                    "ien": 66605
                },
                {
                    "name": "CODY VA CLINIC",
                    "stationNumber": "666GD",
                    "npi": "1366489841",
                    "visn": "19",
                    "ien": 66606
                },
                {
                    "name": "GILLETTE VA CLINIC",
                    "stationNumber": "666GE",
                    "npi": "1518911940",
                    "visn": "19",
                    "ien": 66607
                },
                {
                    "name": "ROCK SPRINGS CBOC",
                    "stationNumber": "666GF",
                    "visn": "19",
                    "ien": 66608
                },
                {
                    "name": "SHER-PRRTP",
                    "stationNumber": "666PA",
                    "officialName": "SHERIDAN PRRTP",
                    "visn": "19",
                    "ien": 66609
                },
                {
                    "name": "AFTON VA CLINIC",
                    "stationNumber": "666QA",
                    "npi": "1205227428",
                    "visn": "19",
                    "ien": 66610
                },
                {
                    "name": "EVANSTON VA CLINIC",
                    "stationNumber": "666QB",
                    "npi": "1043601255",
                    "visn": "19",
                    "ien": 66611
                },
                {
                    "name": "WORLAND VA CLINIC",
                    "stationNumber": "666QC",
                    "npi": "1578954798",
                    "visn": "19",
                    "ien": 66612
                }
            ]
        },
        "23": {
            "437": [
                {
                    "name": "FARGO VA HCS",
                    "stationNumber": "437",
                    "npi": "1437100971",
                    "visn": "23",
                    "ien": "437"
                },
                {
                    "name": "FARGO (NHCU)",
                    "stationNumber": "4379AA",
                    "visn": "23",
                    "ien": 43701
                },
                {
                    "name": "MINOT AFB",
                    "stationNumber": "437CZ",
                    "visn": "23",
                    "ien": 43702
                },
                {
                    "name": "LISBON",
                    "stationNumber": "437DT",
                    "visn": "23",
                    "ien": 43703
                },
                {
                    "name": "FERGUS FALLS",
                    "stationNumber": "437DU",
                    "officialName": "MINESOTA VETERANS STATE HOME",
                    "visn": "23",
                    "ien": 43704
                },
                {
                    "name": "GRAFTON VA CLINIC",
                    "stationNumber": "437GA",
                    "npi": "1457306813",
                    "visn": "23",
                    "ien": 43705
                },
                {
                    "name": "BISMARCK",
                    "stationNumber": "437GB",
                    "npi": "1578514154",
                    "visn": "23",
                    "ien": 43706
                },
                {
                    "name": "FERGUS FALLS",
                    "stationNumber": "437GC",
                    "npi": "1407809478",
                    "visn": "23",
                    "ien": 43707
                },
                {
                    "name": "MINOT CBOC",
                    "stationNumber": "437GD",
                    "npi": "1528006574",
                    "visn": "23",
                    "ien": 43708
                },
                {
                    "name": "BEMIDJI CBOC",
                    "stationNumber": "437GE",
                    "npi": "1346437944",
                    "visn": "23",
                    "ien": 43709
                },
                {
                    "name": "WILLISTON CBOC",
                    "stationNumber": "437GF",
                    "npi": "1598952194",
                    "visn": "23",
                    "ien": 43710
                },
                {
                    "name": "GRAND FORKS CBOC",
                    "stationNumber": "437GI",
                    "npi": "1811220742",
                    "visn": "23",
                    "ien": 43711
                },
                {
                    "name": "DICKINSON VA CLINIC",
                    "stationNumber": "437GJ",
                    "npi": "1023426913",
                    "visn": "23",
                    "ien": 43712
                },
                {
                    "name": "JAMESTOWN VA CLINIC",
                    "stationNumber": "437GK",
                    "npi": "1225445463",
                    "visn": "23",
                    "ien": 43713
                },
                {
                    "name": "DEVILS LAKE VA CBOC",
                    "stationNumber": "437GL",
                    "npi": "1851796072",
                    "visn": "23",
                    "ien": 43714
                },
                {
                    "name": "IDES MINOT AFB",
                    "stationNumber": "437MA",
                    "visn": "23",
                    "ien": 43715
                },
                {
                    "name": "IDES GRAND FORKS AFB",
                    "stationNumber": "437MB",
                    "visn": "23",
                    "ien": 43716
                },
                {
                    "name": "FARGO PRRTP",
                    "stationNumber": "437PB",
                    "visn": "23",
                    "ien": 43717
                },
                {
                    "name": "NORTH FARGO VA CLINIC",
                    "stationNumber": "437QA",
                    "visn": "23",
                    "ien": 43718
                }
            ],
            "438": [
                {
                    "name": "SIOUX FALLS VA HCS",
                    "stationNumber": "438",
                    "npi": "1689624413",
                    "visn": "23",
                    "ien": "438"
                },
                {
                    "name": "SIOUX FALLS NHC",
                    "stationNumber": "4389AA",
                    "visn": "23",
                    "ien": 43801
                },
                {
                    "name": "LUVERNE",
                    "stationNumber": "438DT",
                    "officialName": "LUVERNE CBOC",
                    "visn": "23",
                    "ien": 43802
                },
                {
                    "name": "SPIRIT LAKE CBOC",
                    "stationNumber": "438GA",
                    "npi": "1881874337",
                    "visn": "23",
                    "ien": 43803
                },
                {
                    "name": "SIOUX CITY CBOC",
                    "stationNumber": "438GC",
                    "npi": "1043258577",
                    "visn": "23",
                    "ien": 43804
                },
                {
                    "name": "ABERDEEN CBOC",
                    "stationNumber": "438GD",
                    "npi": "1730139395",
                    "visn": "23",
                    "ien": 43805
                },
                {
                    "name": "WAGNER CBOC",
                    "stationNumber": "438GE",
                    "npi": "1982934691",
                    "visn": "23",
                    "ien": 43806
                },
                {
                    "name": "WATERTOWN VA CBOC",
                    "stationNumber": "438GF",
                    "officialName": "WATERTOWN CBOC",
                    "npi": "1679709265",
                    "visn": "23",
                    "ien": 43807
                },
                {
                    "name": "SIOUX FALLS VA CLINIC",
                    "stationNumber": "438QA",
                    "npi": "1821488032",
                    "visn": "23",
                    "ien": 43808
                }
            ],
            "568": [
                {
                    "name": "BLACK HILLS HCS",
                    "stationNumber": "568",
                    "officialName": "BLACK HILLS HEALTH CARE SYSTEM - FT. MEADE DIVISION",
                    "npi": "1851340087",
                    "domain": "BLACK-HILLS.MED.VA.GOV",
                    "visn": "23",
                    "ien": "568",
                    "mn": "BLA"
                },
                {
                    "name": "FORT MEADE NHC",
                    "stationNumber": "5689AA",
                    "visn": "23",
                    "ien": 56801
                },
                {
                    "name": "HOT SPRINGS NHC",
                    "stationNumber": "5689AB",
                    "visn": "23",
                    "ien": 56802
                },
                {
                    "name": "HOT SPRINGS, SD MC",
                    "stationNumber": "568A4",
                    "npi": "1790734440",
                    "visn": "23",
                    "ien": 56803
                },
                {
                    "name": "HOT SPRINGS DOM",
                    "stationNumber": "568BU",
                    "visn": "23",
                    "ien": 56804
                },
                {
                    "name": "FORT MEADE VA FACILITY DOM",
                    "stationNumber": "568BV",
                    "officialName": "FORT MEADE VA FACILITY DOMICILIARY",
                    "visn": "23",
                    "ien": 56805
                },
                {
                    "name": "HOT SPRINGS STATE DOM",
                    "stationNumber": "568DT",
                    "visn": "23",
                    "ien": 56806
                },
                {
                    "name": "RAPID CITY CBOC",
                    "stationNumber": "568GA",
                    "npi": "1346288438",
                    "visn": "23",
                    "ien": 56807
                },
                {
                    "name": "PIERRE CBOC",
                    "stationNumber": "568GB",
                    "npi": "1770520397",
                    "visn": "23",
                    "ien": 56808
                },
                {
                    "name": "NEWCASTLE",
                    "stationNumber": "568HA",
                    "npi": "1619914116",
                    "visn": "23",
                    "ien": 56809
                },
                {
                    "name": "GORDON CBOC",
                    "stationNumber": "568HB",
                    "npi": "1679517346",
                    "visn": "23",
                    "ien": 56810
                },
                {
                    "name": "PINE RIDGE",
                    "stationNumber": "568HF",
                    "npi": "1386688018",
                    "visn": "23",
                    "ien": 56811
                },
                {
                    "name": "SIOUX SAN",
                    "stationNumber": "568HG",
                    "visn": "23",
                    "ien": 56812
                },
                {
                    "name": "SCOTTSBLUFF CBOC",
                    "stationNumber": "568HH",
                    "npi": "1780623199",
                    "visn": "23",
                    "ien": 56813
                },
                {
                    "name": "MISSION CBOC",
                    "stationNumber": "568HJ",
                    "npi": "1033153705",
                    "visn": "23",
                    "ien": 56814
                },
                {
                    "name": "MCLAUGHLIN VA CLINIC",
                    "stationNumber": "568HK",
                    "npi": "1205870201",
                    "visn": "23",
                    "ien": 56815
                },
                {
                    "name": "WINNER",
                    "stationNumber": "568HP",
                    "npi": "1922049378",
                    "visn": "23",
                    "ien": 56816
                },
                {
                    "name": "IDES ELLSWORTH AFB",
                    "stationNumber": "568MA",
                    "visn": "23",
                    "ien": 56817
                },
                {
                    "name": "BHH-PRRTP",
                    "stationNumber": "568PA",
                    "visn": "23",
                    "ien": 56818
                }
            ],
            "618": [
                {
                    "name": "MINNEAPOLIS VA HCS",
                    "stationNumber": "618",
                    "npi": "1447209804",
                    "domain": "MINNEAPOLIS.MED.VA.GOV",
                    "visn": "23",
                    "ien": "618",
                    "mn": "MIN"
                },
                {
                    "name": "MINNEAPLS NHC",
                    "stationNumber": "6189AA",
                    "visn": "23",
                    "ien": 61801
                },
                {
                    "name": "MINNEAPLS STATE HOME",
                    "stationNumber": "6189AF",
                    "visn": "23",
                    "ien": 61802
                },
                {
                    "name": "CHIPPEWA FALLS STATE VET HOME",
                    "stationNumber": "6189AH",
                    "visn": "23",
                    "ien": 61803
                },
                {
                    "name": "TWIN PORTS CBOC",
                    "stationNumber": "618BY",
                    "npi": "1740229764",
                    "visn": "23",
                    "ien": 61804
                },
                {
                    "name": "MINNEAPOLIS (STATE VET HOME)",
                    "stationNumber": "618DT",
                    "visn": "23",
                    "ien": 61805
                },
                {
                    "name": "HASTINGS (STATE VET HOME)",
                    "stationNumber": "618DU",
                    "visn": "23",
                    "ien": 61806
                },
                {
                    "name": "SILVER BAY (STATE VET HOME)",
                    "stationNumber": "618DV",
                    "visn": "23",
                    "ien": 61807
                },
                {
                    "name": "ST. JAMES VA CLINIC",
                    "stationNumber": "618GA",
                    "officialName": "SOUTH CENTRAL CBOC",
                    "npi": "1518905074",
                    "visn": "23",
                    "ien": 61808
                },
                {
                    "name": "HIBBING (CBOC)",
                    "stationNumber": "618GB",
                    "npi": "1902851348",
                    "visn": "23",
                    "ien": 61809
                },
                {
                    "name": "MAPLEWOOD CBOC",
                    "stationNumber": "618GD",
                    "officialName": "ST. PAUL (CBOC)",
                    "npi": "1659328292",
                    "visn": "23",
                    "ien": 61810
                },
                {
                    "name": "CHIPPEWA VALLEY CBOC",
                    "stationNumber": "618GE",
                    "npi": "1417902941",
                    "visn": "23",
                    "ien": 61811
                },
                {
                    "name": "ROCHESTER (CBOC)",
                    "stationNumber": "618GG",
                    "npi": "1790723716",
                    "visn": "23",
                    "ien": 61812
                },
                {
                    "name": "HAYWARD CBOC",
                    "stationNumber": "618GH",
                    "npi": "1235320912",
                    "visn": "23",
                    "ien": 61813
                },
                {
                    "name": "NORTHWEST METRO VA CLINIC",
                    "stationNumber": "618GI",
                    "officialName": "NORTHWEST METRO VA OPC",
                    "npi": "1912206228",
                    "visn": "23",
                    "ien": 61814
                },
                {
                    "name": "SHAKOPEE CBOC",
                    "stationNumber": "618GJ",
                    "visn": "23",
                    "ien": 61815
                },
                {
                    "name": "ALBERT LEA CBOC",
                    "stationNumber": "618GK",
                    "visn": "23",
                    "ien": 61816
                },
                {
                    "name": "MINNEAPOLIS VA CBOC",
                    "stationNumber": "618GL",
                    "npi": "1609283845",
                    "visn": "23",
                    "ien": 61817
                },
                {
                    "name": "RICE LAKE VA CBOC",
                    "stationNumber": "618GM",
                    "npi": "1124436191",
                    "visn": "23",
                    "ien": 61818
                },
                {
                    "name": "LYLE C PEARSON CBOC",
                    "stationNumber": "618GN",
                    "npi": "1639615149",
                    "visn": "23",
                    "ien": 61819
                },
                {
                    "name": "FORT SNELLING VA CLINIC",
                    "stationNumber": "618QA",
                    "npi": "1659761864",
                    "visn": "23",
                    "ien": 61820
                },
                {
                    "name": "ELY VA CLINIC",
                    "stationNumber": "618QB",
                    "npi": "1073094884",
                    "visn": "23",
                    "ien": 61821
                },
                {
                    "name": "MINNEAPOLIS (SCI)",
                    "stationNumber": "618SCI",
                    "officialName": "MINNEAPOLIS SPINAL CORD INJ DISORDER CTR",
                    "visn": "23",
                    "ien": 61822
                }
            ],
            "636": [
                {
                    "name": "VA NWIHS, OMAHA DIVISION",
                    "stationNumber": "636",
                    "officialName": "VA CENTRAL PLAINS HEALTH NETWORK - OMAHA DIVISION",
                    "npi": "1679523054",
                    "domain": "CENTRAL-PLAINS.MED.VA.GOV",
                    "visn": "23",
                    "ien": "636",
                    "mn": "CEN"
                },
                {
                    "name": "GRAND ISLAND VA NHCU",
                    "stationNumber": "6369AB",
                    "visn": "23",
                    "ien": 63601
                },
                {
                    "name": "DES MOINES VA COMM. LIVING CTR",
                    "stationNumber": "6369AD",
                    "npi": "1801256110",
                    "visn": "23",
                    "ien": 63602
                },
                {
                    "name": "ILLINOIS NHC",
                    "stationNumber": "6369AJ",
                    "visn": "23",
                    "ien": 63603
                },
                {
                    "name": "VA NWIHS, GRAND ISLAND DIV",
                    "stationNumber": "636A4",
                    "npi": "1518912278",
                    "visn": "23",
                    "ien": 63604
                },
                {
                    "name": "VA NWIHS, LINCOLN DIVISION",
                    "stationNumber": "636A5",
                    "officialName": "VA CENTRAL PLAINS HEALTH NETWORK - LINCOLN DIVISION",
                    "npi": "1649211061",
                    "visn": "23",
                    "ien": 63605
                },
                {
                    "name": "VA CIHS, DES MOINES DIVISION",
                    "stationNumber": "636A6",
                    "npi": "1184673949",
                    "visn": "23",
                    "ien": 63606
                },
                {
                    "name": "IOWA CITY HCS",
                    "stationNumber": "636A8",
                    "npi": "1578512687",
                    "visn": "23",
                    "ien": 63607
                },
                {
                    "name": "VA DOM DES MOINES",
                    "stationNumber": "636BU",
                    "officialName": "DES MOINES VA DOMICILIARY",
                    "visn": "23",
                    "ien": 63608
                },
                {
                    "name": "VA DOM KNOXVILLE",
                    "stationNumber": "636BV",
                    "officialName": "KNOXVILLE VA DOMICILIARY",
                    "visn": "23",
                    "ien": 63609
                },
                {
                    "name": "GRAND ISLAND VA FACILITY DOM",
                    "stationNumber": "636BW",
                    "officialName": "GRAND ISLAND VA FACILITY DOMICILIARY",
                    "visn": "23",
                    "ien": 63610
                },
                {
                    "name": "OMAHA VA FACILITY DOM",
                    "stationNumber": "636BX",
                    "officialName": "OMAHA VA FACILITY DOMICILIARY",
                    "visn": "23",
                    "ien": 63611
                },
                {
                    "name": "OFFUTT AFB-BELLEVUE",
                    "stationNumber": "636CZ",
                    "officialName": "BELLEVUE (OFFUTT AFB)",
                    "visn": "23",
                    "ien": 63612
                },
                {
                    "name": "GRAND ISLAND STATE VET HOME",
                    "stationNumber": "636DU",
                    "officialName": "Grand Island Nebraska State Veterans Home",
                    "visn": "23",
                    "ien": 63613
                },
                {
                    "name": "NORFOLK STATE VET HOME",
                    "stationNumber": "636DV",
                    "officialName": "NORFOLK NEBRASKA STATE VETERANS HOME",
                    "visn": "23",
                    "ien": 63614
                },
                {
                    "name": "MARSHALLTOWN STATE DOM",
                    "stationNumber": "636DW",
                    "officialName": "MARSHALLTOWN STATE DOMICILIARY",
                    "visn": "23",
                    "ien": 63615
                },
                {
                    "name": "QUINCY STATE DOM",
                    "stationNumber": "636DX",
                    "officialName": "QUINCY VA DOMICILIARY",
                    "visn": "23",
                    "ien": 63616
                },
                {
                    "name": "EASTERN NEBRASKA SVH",
                    "stationNumber": "636DZ",
                    "officialName": "EASTERN NEBRASKA STATE VETERANS HOME",
                    "visn": "23",
                    "ien": 63617
                },
                {
                    "name": "MARSHALLTOWN STATE VET HOME",
                    "stationNumber": "636EL",
                    "officialName": "MARSHALLTOWN STATE HOME",
                    "visn": "23",
                    "ien": 63618
                },
                {
                    "name": "QUINCY STATE VET HOME",
                    "stationNumber": "636EM",
                    "officialName": "QUINCY STATE HOME",
                    "visn": "23",
                    "ien": 63619
                },
                {
                    "name": "DOMICILLIARY EAST NEBRASKA SVH",
                    "stationNumber": "636EN",
                    "visn": "23",
                    "ien": 63620
                },
                {
                    "name": "NORFOLK CBOC",
                    "stationNumber": "636GA",
                    "npi": "1184661639",
                    "visn": "23",
                    "ien": 63621
                },
                {
                    "name": "NORTH PLATTE CBOC",
                    "stationNumber": "636GB",
                    "npi": "1841238474",
                    "visn": "23",
                    "ien": 63622
                },
                {
                    "name": "MASON CITY CBOC",
                    "stationNumber": "636GC",
                    "officialName": "MASON CITY COMMINITY BASED OUTPATIENT CLINIC",
                    "npi": "1528005402",
                    "visn": "23",
                    "ien": 63623
                },
                {
                    "name": "MARSHALLTOWN CBOC",
                    "stationNumber": "636GD",
                    "officialName": "MARSHALLTOWN COMMUNITY BASED OUTPATIENT CLINIC",
                    "npi": "1861715021",
                    "visn": "23",
                    "ien": 63624
                },
                {
                    "name": "QUAD CITIES VA CLINIC",
                    "stationNumber": "636GF",
                    "npi": "1861443590",
                    "visn": "23",
                    "ien": 63625
                },
                {
                    "name": "QUINCY CBOC",
                    "stationNumber": "636GG",
                    "officialName": "QUINCY COMMUNITY BASED OUTPATIENT CLINIC",
                    "npi": "1629016845",
                    "visn": "23",
                    "ien": 63626
                },
                {
                    "name": "WATERLOO CBOC",
                    "stationNumber": "636GH",
                    "officialName": "WATERLOO COMMUNITY BASED OUTPATIENT CLINIC",
                    "npi": "1134169592",
                    "visn": "23",
                    "ien": 63627
                },
                {
                    "name": "GALESBURG CBOC",
                    "stationNumber": "636GI",
                    "officialName": "GALESBURG COMMUNITY BASED OUTPATIENT CLINIC",
                    "npi": "1629022108",
                    "visn": "23",
                    "ien": 63628
                },
                {
                    "name": "DUBUQUE CBOC",
                    "stationNumber": "636GJ",
                    "officialName": "DUBUQUE COMMUNITY BASED OUTPATIENT CLINIC",
                    "npi": "1346293818",
                    "visn": "23",
                    "ien": 63629
                },
                {
                    "name": "FORT DODGE CBOC",
                    "stationNumber": "636GK",
                    "npi": "1154374148",
                    "visn": "23",
                    "ien": 63630
                },
                {
                    "name": "BELLEVUE VA CBOC",
                    "stationNumber": "636GL",
                    "npi": "1023269818",
                    "visn": "23",
                    "ien": 63631
                },
                {
                    "name": "CARROLL CBOC",
                    "stationNumber": "636GM",
                    "npi": "1003111600",
                    "visn": "23",
                    "ien": 63632
                },
                {
                    "name": "CEDAR RAPIDS CBOC",
                    "stationNumber": "636GN",
                    "npi": "1417182395",
                    "visn": "23",
                    "ien": 63633
                },
                {
                    "name": "SHENANDOAH VA CBOC",
                    "stationNumber": "636GP",
                    "officialName": "SHENANDOAH CBOC",
                    "npi": "1811149529",
                    "visn": "23",
                    "ien": 63634
                },
                {
                    "name": "HOLDREGE, NE CBOC",
                    "stationNumber": "636GQ",
                    "visn": "23",
                    "ien": 63635
                },
                {
                    "name": "KNOXVILLE CBOC",
                    "stationNumber": "636GR",
                    "npi": "1821312109",
                    "visn": "23",
                    "ien": 63636
                },
                {
                    "name": "OTTUMWA CBOC",
                    "stationNumber": "636GS",
                    "npi": "1932418316",
                    "visn": "23",
                    "ien": 63637
                },
                {
                    "name": "STERLING CBOC",
                    "stationNumber": "636GT",
                    "npi": "1851660724",
                    "visn": "23",
                    "ien": 63638
                },
                {
                    "name": "DECORAH CBOC",
                    "stationNumber": "636GU",
                    "npi": "1376817148",
                    "visn": "23",
                    "ien": 63639
                },
                {
                    "name": "O'NEILL VA CBOC",
                    "stationNumber": "636GV",
                    "npi": "1386051514",
                    "visn": "23",
                    "ien": 63640
                },
                {
                    "name": "CORALVILLE VA CBOC",
                    "stationNumber": "636GW",
                    "npi": "1558778514",
                    "visn": "23",
                    "ien": 63641
                },
                {
                    "name": "FORT DODGE NORTH VA CLINIC",
                    "stationNumber": "636GX",
                    "npi": "1548772817",
                    "visn": "23",
                    "ien": 63642
                },
                {
                    "name": "IDES OFFUTT AFB",
                    "stationNumber": "636MA",
                    "visn": "23",
                    "ien": 63643
                },
                {
                    "name": "GRAND ISLAND SARRTP",
                    "stationNumber": "636PA",
                    "visn": "23",
                    "ien": 63644
                },
                {
                    "name": "OMAHA SARRTP",
                    "stationNumber": "636PB",
                    "visn": "23",
                    "ien": 63645
                },
                {
                    "name": "OMAHA PRRTP",
                    "stationNumber": "636PC",
                    "npi": "1922468230",
                    "visn": "23",
                    "ien": 63646
                },
                {
                    "name": "OMAHA VA CLINIC",
                    "stationNumber": "636QA",
                    "npi": "1093105207",
                    "visn": "23",
                    "ien": 63647
                },
                {
                    "name": "DES MOINES VA CLINIC",
                    "stationNumber": "636QB",
                    "npi": "1548650765",
                    "visn": "23",
                    "ien": 63648
                },
                {
                    "name": "LINN COUNTY VA CLINIC",
                    "stationNumber": "636QC",
                    "npi": "1700276920",
                    "visn": "23",
                    "ien": 63649
                },
                {
                    "name": "MACOMB VA CLINIC",
                    "stationNumber": "636QD",
                    "npi": "1578953725",
                    "visn": "23",
                    "ien": 63650
                },
                {
                    "name": "ROCK ISLAND VA CLINIC",
                    "stationNumber": "636QE",
                    "npi": "1851781033",
                    "visn": "23",
                    "ien": 63651
                },
                {
                    "name": "IOWA CITY VA MOBILE CLINIC",
                    "stationNumber": "636QG",
                    "npi": "1063887271",
                    "visn": "23",
                    "ien": 63652
                },
                {
                    "name": "DES MOINES VA MOBILE CLINIC",
                    "stationNumber": "636QH",
                    "npi": "1114304185",
                    "visn": "23",
                    "ien": 63653
                },
                {
                    "name": "DAVENPORT VA CLINIC",
                    "stationNumber": "636QI",
                    "npi": "1063961670",
                    "visn": "23",
                    "ien": 63654
                },
                {
                    "name": "IOWA CITY VA CLINIC",
                    "stationNumber": "636QJ",
                    "npi": "1346710696",
                    "visn": "23",
                    "ien": 63655
                }
            ],
            "656": [
                {
                    "name": "ST. CLOUD VA HCS",
                    "stationNumber": "656",
                    "officialName": "ST. CLOUD VA HEALTH CARE SYSTEM",
                    "npi": "1588614325",
                    "domain": "ST-CLOUD.MED.VA.GOV",
                    "visn": "23",
                    "ien": "656",
                    "mn": "STC"
                },
                {
                    "name": "ST. CLOUD (NHCU)",
                    "stationNumber": "6569AA",
                    "officialName": "ST CLOUD NHCU",
                    "visn": "23",
                    "ien": 65601
                },
                {
                    "name": "ST. CLOUD (DOM)",
                    "stationNumber": "656BU",
                    "officialName": "ST. CLOUD DOM",
                    "visn": "23",
                    "ien": 65602
                },
                {
                    "name": "BRAINERD CBOC",
                    "stationNumber": "656GA",
                    "officialName": "BRAINERD VA COMMUNITY OUTPATIENT CLINIC",
                    "npi": "1871640839",
                    "visn": "23",
                    "ien": 65603
                },
                {
                    "name": "MONTEVIDEO CBOC",
                    "stationNumber": "656GB",
                    "npi": "1124175179",
                    "visn": "23",
                    "ien": 65604
                },
                {
                    "name": "MAX J. BEILKE CBOC",
                    "stationNumber": "656GC",
                    "npi": "1942431242",
                    "visn": "23",
                    "ien": 65605
                },
                {
                    "name": "STC-PRRTP",
                    "stationNumber": "656PA",
                    "officialName": "ST CLOUD PRRTP",
                    "visn": "23",
                    "ien": 65606
                },
                {
                    "name": "ST. CLOUD VA MOBILE CLINIC",
                    "stationNumber": "656QA",
                    "npi": "1326438433",
                    "visn": "23",
                    "ien": 65607
                }
            ]
        },
        "4": {
            "460": [
                {
                    "name": "WILMINGTON VA MEDICAL CENTER",
                    "stationNumber": "460",
                    "npi": "1407805898",
                    "visn": "4",
                    "ien": "460"
                },
                {
                    "name": "WILMINGTON NHC",
                    "stationNumber": "4609AA",
                    "visn": "4",
                    "ien": 46001
                },
                {
                    "name": "VINELAND STNB",
                    "stationNumber": "4609AF",
                    "visn": "4",
                    "ien": 46002
                },
                {
                    "name": "NEW JERSEY MEMORIAL HOME",
                    "stationNumber": "460DU",
                    "visn": "4",
                    "ien": 46003
                },
                {
                    "name": "DELAWARE STATE VETERANS HOME",
                    "stationNumber": "460DV",
                    "visn": "4",
                    "ien": 46004
                },
                {
                    "name": "SUSSEX COUNTY VA CLINIC",
                    "stationNumber": "460GA",
                    "npi": "1033157946",
                    "visn": "4",
                    "ien": 46005
                },
                {
                    "name": "KENT COUNTY VA CLINIC",
                    "stationNumber": "460GC",
                    "npi": "1336395136",
                    "visn": "4",
                    "ien": 46006
                },
                {
                    "name": "CAPE MAY (WILMINGTON)",
                    "stationNumber": "460GD",
                    "npi": "1720131287",
                    "visn": "4",
                    "ien": 46007
                },
                {
                    "name": "ATLANTIC COUNTY VA CLINIC",
                    "stationNumber": "460HE",
                    "npi": "1295775682",
                    "visn": "4",
                    "ien": 46008
                },
                {
                    "name": "CUMBERLAND COUNTY VA CLINIC-NJ",
                    "stationNumber": "460HG",
                    "officialName": "CUMBERLAND COUNTY VA CLINIC",
                    "npi": "1053351866",
                    "visn": "4",
                    "ien": 46009
                },
                {
                    "name": "WILMINGTON VA MOBILE CLINIC",
                    "stationNumber": "460HK",
                    "visn": "4",
                    "ien": 46010
                },
                {
                    "name": "SMYRNA MORC",
                    "stationNumber": "460HL",
                    "visn": "4",
                    "ien": 46011
                },
                {
                    "name": "MILLINTON MORC",
                    "stationNumber": "460HM",
                    "visn": "4",
                    "ien": 46012
                },
                {
                    "name": "SALEM CITY MORC",
                    "stationNumber": "460HO",
                    "visn": "4",
                    "ien": 46013
                },
                {
                    "name": "IDES DOVER AFB",
                    "stationNumber": "460MA",
                    "visn": "4",
                    "ien": 46014
                }
            ],
            "503": [
                {
                    "name": "JAMES E VAN ZANDT VAMC",
                    "stationNumber": "503",
                    "npi": "1316996887",
                    "visn": "4",
                    "ien": "503"
                },
                {
                    "name": "ALTOONA (NHCU)",
                    "stationNumber": "5039AA",
                    "visn": "4",
                    "ien": 50301
                },
                {
                    "name": "HOLIDAYSBURG (STATE DOM)",
                    "stationNumber": "5039AF",
                    "visn": "4",
                    "ien": 50302
                },
                {
                    "name": "HOLLIDAYSBURG (STNB)",
                    "stationNumber": "503DT",
                    "officialName": "HOLLIDAYSBURG STATE VETERANS HOME",
                    "visn": "4",
                    "ien": 50303
                },
                {
                    "name": "JOHNSTOWN VA CLINIC",
                    "stationNumber": "503GA",
                    "npi": "1831145481",
                    "visn": "4",
                    "ien": 50304
                },
                {
                    "name": "DUBOIS CBOC",
                    "stationNumber": "503GB",
                    "npi": "1366495848",
                    "visn": "4",
                    "ien": 50305
                },
                {
                    "name": "STATE COLLEGE VA CLINIC",
                    "stationNumber": "503GC",
                    "npi": "1134168032",
                    "visn": "4",
                    "ien": 50306
                },
                {
                    "name": "HUNTINGDON COUNTY VA CLINIC",
                    "stationNumber": "503GD",
                    "visn": "4",
                    "ien": 50307
                },
                {
                    "name": "INDIANA COUNTY VA CLINIC",
                    "stationNumber": "503GE",
                    "visn": "4",
                    "ien": 50308
                }
            ],
            "529": [
                {
                    "name": "ABIE ABRAHAM VA CLINIC",
                    "stationNumber": "529",
                    "npi": "1851340178",
                    "domain": "BUTLER.MED.VA.GOV",
                    "visn": "4",
                    "ien": "529",
                    "mn": "BUT"
                },
                {
                    "name": "BUTLER (NHC)",
                    "stationNumber": "5299AA",
                    "visn": "4",
                    "ien": 52901
                },
                {
                    "name": "BUTLER VAMC",
                    "stationNumber": "529A4",
                    "npi": "1881176576",
                    "visn": "4",
                    "ien": 52902
                },
                {
                    "name": "BUTLER (D)",
                    "stationNumber": "529BU",
                    "officialName": "Butler Veterans Affairs Medical Center-Domiciliary",
                    "visn": "4",
                    "ien": 52903
                },
                {
                    "name": "MICHAEL A. MARZANO VA OPC",
                    "stationNumber": "529GA",
                    "npi": "1912952334",
                    "visn": "4",
                    "ien": 52904
                },
                {
                    "name": "LAWRENCE COUNTY VA CLINIC",
                    "stationNumber": "529GB",
                    "npi": "1518912690",
                    "visn": "4",
                    "ien": 52905
                },
                {
                    "name": "ARMSTRONG COUNTY VA CLINIC",
                    "stationNumber": "529GC",
                    "npi": "1376585588",
                    "visn": "4",
                    "ien": 52906
                },
                {
                    "name": "CLARION COUNTY VA CLINIC",
                    "stationNumber": "529GD",
                    "npi": "1346292760",
                    "visn": "4",
                    "ien": 52907
                },
                {
                    "name": "CRANBERRY TOWNSHIP VA CLINIC",
                    "stationNumber": "529GF",
                    "npi": "1801123385",
                    "visn": "4",
                    "ien": 52908
                },
                {
                    "name": "BUTLER CWT/TR",
                    "stationNumber": "529PB",
                    "officialName": "BUTLER COMP WORK THERAPY/TRANSITIONAL RES PRG",
                    "visn": "4",
                    "ien": 52909
                }
            ],
            "542": [
                {
                    "name": "COATESVILLE VA MEDICAL CENTER",
                    "stationNumber": "542",
                    "npi": "1669422192",
                    "domain": "COATESVILLE.MED.VA.GOV",
                    "visn": "4",
                    "ien": "542",
                    "mn": "COA"
                },
                {
                    "name": "COATESVILLE (NHC)",
                    "stationNumber": "5429AA",
                    "visn": "4",
                    "ien": 54201
                },
                {
                    "name": "COATESVILLE (DOM)",
                    "stationNumber": "542BU",
                    "npi": "1043670516",
                    "visn": "4",
                    "ien": 54202
                },
                {
                    "name": "DELAWARE COUNTY VA CLINIC",
                    "stationNumber": "542GA",
                    "npi": "1932146891",
                    "visn": "4",
                    "ien": 54203
                },
                {
                    "name": "SPRING CITY VA CLINIC",
                    "stationNumber": "542GE",
                    "npi": "1669410635",
                    "visn": "4",
                    "ien": 54204
                },
                {
                    "name": "COATESVILLE PRRTP",
                    "stationNumber": "542PA",
                    "visn": "4",
                    "ien": 54205
                }
            ],
            "562": [
                {
                    "name": "ERIE VA MEDICAL CENTER",
                    "stationNumber": "562",
                    "npi": "1396796827",
                    "domain": "ERIE.MED.VA.GOV",
                    "visn": "4",
                    "ien": "562",
                    "mn": "ERI"
                },
                {
                    "name": "ERI (NHC)",
                    "stationNumber": "5629AA",
                    "visn": "4",
                    "ien": 56201
                },
                {
                    "name": "ERI (STNB)",
                    "stationNumber": "5629AF",
                    "visn": "4",
                    "ien": 56202
                },
                {
                    "name": "ERIE VA DOMICILIARY",
                    "stationNumber": "562BU",
                    "visn": "4",
                    "ien": 56203
                },
                {
                    "name": "ERIE S",
                    "stationNumber": "562DT",
                    "officialName": "ERIE SOLDIERS & SAILORS HOME",
                    "visn": "4",
                    "ien": 56204
                },
                {
                    "name": "CRAWFORD COUNTY VA CLINIC",
                    "stationNumber": "562GA",
                    "npi": "1700839396",
                    "visn": "4",
                    "ien": 56205
                },
                {
                    "name": "ASHTABULA COUNTY VA CLINIC",
                    "stationNumber": "562GB",
                    "npi": "1871544064",
                    "visn": "4",
                    "ien": 56206
                },
                {
                    "name": "MCKEAN COUNTY VA CLINIC",
                    "stationNumber": "562GC",
                    "npi": "1598703118",
                    "visn": "4",
                    "ien": 56207
                },
                {
                    "name": "VENANGO COUNTY VA CLINIC",
                    "stationNumber": "562GD",
                    "npi": "1710037973",
                    "visn": "4",
                    "ien": 56208
                },
                {
                    "name": "WARREN COUNTY CLINIC",
                    "stationNumber": "562GE",
                    "npi": "1235275751",
                    "visn": "4",
                    "ien": 56209
                }
            ],
            "595": [
                {
                    "name": "LEBANON VA MEDICAL CENTER",
                    "stationNumber": "595",
                    "npi": "1124077672",
                    "domain": "LEBANON.MED.VA.GOV",
                    "visn": "4",
                    "ien": "595",
                    "mn": "LEB"
                },
                {
                    "name": "LEBANON NHC",
                    "stationNumber": "5959AA",
                    "visn": "4",
                    "ien": 59501
                },
                {
                    "name": "LEBANON VA FACILITY DOM",
                    "stationNumber": "595BU",
                    "officialName": "LEBANON VA FACILITY DOMICILIARY",
                    "visn": "4",
                    "ien": 59502
                },
                {
                    "name": "LEBANON",
                    "stationNumber": "595BY",
                    "officialName": "LEBANON VA MEDICAL CENTER",
                    "visn": "4",
                    "ien": 59503
                },
                {
                    "name": "CUMBERLAND COUNTY VA CLINIC",
                    "stationNumber": "595GA",
                    "npi": "1346291937",
                    "visn": "4",
                    "ien": 59504
                },
                {
                    "name": "LANCASTER VA CLINIC",
                    "stationNumber": "595GC",
                    "npi": "1265489165",
                    "visn": "4",
                    "ien": 59505
                },
                {
                    "name": "READING",
                    "stationNumber": "595GD",
                    "officialName": "BERKS VA OUTPATIENT CLINIC",
                    "npi": "1639117641",
                    "visn": "4",
                    "ien": 59506
                },
                {
                    "name": "YORK VA CLINIC",
                    "stationNumber": "595GE",
                    "npi": "1518901750",
                    "visn": "4",
                    "ien": 59507
                },
                {
                    "name": "SCHUYLKILL COUNTY VA CLINIC",
                    "stationNumber": "595GF",
                    "visn": "4",
                    "ien": 59508
                },
                {
                    "name": "LEBANON",
                    "stationNumber": "595PA",
                    "officialName": "LEBANON VA MEDICAL CENTER",
                    "visn": "4",
                    "ien": 59509
                },
                {
                    "name": "FORT INDIANTOWN GAP VA CLINIC",
                    "stationNumber": "595QA",
                    "npi": "1184012684",
                    "visn": "4",
                    "ien": 59510
                }
            ],
            "642": [
                {
                    "name": "PHILADELPHIA, PA VAMC",
                    "stationNumber": "642",
                    "npi": "1659321693",
                    "domain": "PHILADELPHIA.MED.VA.GOV",
                    "visn": "4",
                    "ien": "642",
                    "mn": "PHI"
                },
                {
                    "name": "PHILADELPHIA NURSING HOME",
                    "stationNumber": "6429AA",
                    "visn": "4",
                    "ien": 64201
                },
                {
                    "name": "PHILADELPHIA VA DOMICILIARY",
                    "stationNumber": "642BU",
                    "npi": "1679933048",
                    "visn": "4",
                    "ien": 64202
                },
                {
                    "name": "DELAWARE VALLEY STATE VET HOME",
                    "stationNumber": "642DT",
                    "visn": "4",
                    "ien": 64203
                },
                {
                    "name": "BURLINGTON COUNTY VA CLINIC",
                    "stationNumber": "642GA",
                    "npi": "1750324406",
                    "visn": "4",
                    "ien": 64204
                },
                {
                    "name": "VICTOR J. SARACINI VA OPC",
                    "stationNumber": "642GC",
                    "npi": "1689618795",
                    "visn": "4",
                    "ien": 64205
                },
                {
                    "name": "GLOUCESTER COUNTY VA CLINIC",
                    "stationNumber": "642GD",
                    "npi": "1669426086",
                    "visn": "4",
                    "ien": 64206
                },
                {
                    "name": "CAMDEN VA CLINIC",
                    "stationNumber": "642GF",
                    "npi": "1841608882",
                    "visn": "4",
                    "ien": 64207
                },
                {
                    "name": "WEST PHILADELPHIA VA CLINIC",
                    "stationNumber": "642GH",
                    "npi": "1407316219",
                    "visn": "4",
                    "ien": 64208
                },
                {
                    "name": "CHESTNUT STREET VA CLINIC",
                    "stationNumber": "642QA",
                    "npi": "1134517360",
                    "visn": "4",
                    "ien": 64209
                },
                {
                    "name": "FOURTH ST VA CLINIC",
                    "stationNumber": "642QB",
                    "npi": "1316324486",
                    "visn": "4",
                    "ien": 64210
                }
            ],
            "646": [
                {
                    "name": "PITTSBURGH VAMC UNIVERSITY DR.",
                    "stationNumber": "646",
                    "npi": "1811947815",
                    "domain": "PITTSBURGH.MED.VA.GOV",
                    "visn": "4",
                    "ien": "646",
                    "mn": "PIT"
                },
                {
                    "name": "H. JOHN HEINZ III VAMC",
                    "stationNumber": "646A4",
                    "npi": "1366486094",
                    "visn": "4",
                    "ien": 64601
                },
                {
                    "name": "PITTSBURGH-HEINZ VA FAC DOM",
                    "stationNumber": "646BU",
                    "visn": "4",
                    "ien": 64602
                },
                {
                    "name": "SOUTHWESTERN VET CTR, PA SVH",
                    "stationNumber": "646DT",
                    "officialName": "SOUTHWESTERN VETERANS CENTER",
                    "visn": "4",
                    "ien": 64603
                },
                {
                    "name": "BELMONT COUNTY VA CLINIC",
                    "stationNumber": "646GA",
                    "npi": "1821038290",
                    "visn": "4",
                    "ien": 64604
                },
                {
                    "name": "WESTMORELAND COUNTY VA CLINIC",
                    "stationNumber": "646GB",
                    "npi": "1730134040",
                    "visn": "4",
                    "ien": 64605
                },
                {
                    "name": "BEAVER COUNTY VA CLINIC",
                    "stationNumber": "646GC",
                    "npi": "1275584757",
                    "visn": "4",
                    "ien": 64606
                },
                {
                    "name": "WASHINGTON, PA CBOC",
                    "stationNumber": "646GD",
                    "officialName": "WASHINGTON VA PRIMARY CARE OUTPATIENT CLINIC",
                    "npi": "1518907328",
                    "visn": "4",
                    "ien": 64607
                },
                {
                    "name": "FAYETTE COUNTY VA CLINIC",
                    "stationNumber": "646GE",
                    "npi": "1184771123",
                    "visn": "4",
                    "ien": 64608
                },
                {
                    "name": "PITTSBURGH PRRTP",
                    "stationNumber": "646PA",
                    "npi": "1053771436",
                    "visn": "4",
                    "ien": 64609
                }
            ],
            "693": [
                {
                    "name": "WILKES-BARRE VAMC",
                    "stationNumber": "693",
                    "npi": "1952351967",
                    "domain": "WILKES-BARRE.MED.VA.GOV",
                    "visn": "4",
                    "ien": "693",
                    "mn": "WIL"
                },
                {
                    "name": "WILKES-BARRE (NHC)",
                    "stationNumber": "6939AA",
                    "visn": "4",
                    "ien": 69301
                },
                {
                    "name": "ALLENTOWN OPC",
                    "stationNumber": "693B4",
                    "npi": "1053362145",
                    "visn": "4",
                    "ien": 69302
                },
                {
                    "name": "WILKES-BARRE VA FACILITY DOM",
                    "stationNumber": "693BU",
                    "officialName": "WILKES-BARRE VA FACILITY DOMICILIARY",
                    "visn": "4",
                    "ien": 69303
                },
                {
                    "name": "WILKES-BARRE (DENT)",
                    "stationNumber": "693BY",
                    "visn": "4",
                    "ien": 69304
                },
                {
                    "name": "SCRANTON STNB",
                    "stationNumber": "693DT",
                    "visn": "4",
                    "ien": 69305
                },
                {
                    "name": "SAYRE OPC",
                    "stationNumber": "693GA",
                    "npi": "1669410262",
                    "visn": "4",
                    "ien": 69306
                },
                {
                    "name": "WILLIAMSPORT VA CLINIC",
                    "stationNumber": "693GB",
                    "npi": "1750321527",
                    "visn": "4",
                    "ien": 69307
                },
                {
                    "name": "TOBYHANNA VA CLINIC",
                    "stationNumber": "693GC",
                    "npi": "1740224971",
                    "visn": "4",
                    "ien": 69308
                },
                {
                    "name": "WILKES-BARRE (CBOC)",
                    "stationNumber": "693GD",
                    "visn": "4",
                    "ien": 69309
                },
                {
                    "name": "SCHUYLKILL COUNTY CBOC",
                    "stationNumber": "693GE",
                    "npi": "1558558163",
                    "visn": "4",
                    "ien": 69310
                },
                {
                    "name": "COLUMBIA COUNTY VA CLINIC",
                    "stationNumber": "693GF",
                    "npi": "1922059591",
                    "visn": "4",
                    "ien": 69311
                },
                {
                    "name": "NORTHAMPTON COUNTY CLINIC",
                    "stationNumber": "693GG",
                    "npi": "1366639973",
                    "visn": "4",
                    "ien": 69312
                },
                {
                    "name": "WIL-SARRTP",
                    "stationNumber": "693PA",
                    "visn": "4",
                    "ien": 69313
                },
                {
                    "name": "WAYNE COUNTY VA CLINIC",
                    "stationNumber": "693QA",
                    "visn": "4",
                    "ien": 69314
                }
            ]
        },
        "22": {
            "501": [
                {
                    "name": "RAYMOND G. MURPHY VAMC",
                    "stationNumber": "501",
                    "npi": "1013966118",
                    "visn": "22",
                    "ien": "501"
                },
                {
                    "name": "ALBUQUERQUE (NHCU)",
                    "stationNumber": "5019AA",
                    "visn": "22",
                    "ien": 50101
                },
                {
                    "name": "ALBUQUERQUE STATE DOMICILIARY",
                    "stationNumber": "5019AF",
                    "visn": "22",
                    "ien": 50102
                },
                {
                    "name": "SUBSIDIARY VA DRRTP",
                    "stationNumber": "501BU",
                    "visn": "22",
                    "ien": 50103
                },
                {
                    "name": "GALLUP VA DOMICILARY",
                    "stationNumber": "501BV",
                    "npi": "1376090316",
                    "visn": "22",
                    "ien": 50104
                },
                {
                    "name": "ALBUQUERQUE ARMY HOSPITAL",
                    "stationNumber": "501CN",
                    "visn": "22",
                    "ien": 50105
                },
                {
                    "name": "TRUTH OR CONSEQUENCES",
                    "stationNumber": "501DT",
                    "officialName": "TRUTH OR CONSEQUENCES (DOM)",
                    "visn": "22",
                    "ien": 50106
                },
                {
                    "name": "FORT BAYARD (STATE NURSING)",
                    "stationNumber": "501DU",
                    "visn": "22",
                    "ien": 50107
                },
                {
                    "name": "LAS VEGAS (CBOC)",
                    "stationNumber": "501G2",
                    "npi": "1053353102",
                    "visn": "22",
                    "ien": 50108
                },
                {
                    "name": "ARTESIA VA CLINIC",
                    "stationNumber": "501GA",
                    "npi": "1043261472",
                    "visn": "22",
                    "ien": 50109
                },
                {
                    "name": "FARMINGTON VA CLINIC",
                    "stationNumber": "501GB",
                    "npi": "1770536641",
                    "visn": "22",
                    "ien": 50110
                },
                {
                    "name": "SILVER CITY VA CLINIC",
                    "stationNumber": "501GC",
                    "npi": "1699713156",
                    "visn": "22",
                    "ien": 50111
                },
                {
                    "name": "GALLUP VA CLINIC",
                    "stationNumber": "501GD",
                    "npi": "1477509172",
                    "visn": "22",
                    "ien": 50112
                },
                {
                    "name": "ESPANOLA (CBOC)",
                    "stationNumber": "501GE",
                    "npi": "1972556322",
                    "visn": "22",
                    "ien": 50113
                },
                {
                    "name": "TRUTH OR CONSEQUENCES (CBOC)",
                    "stationNumber": "501GH",
                    "npi": "1619917044",
                    "visn": "22",
                    "ien": 50114
                },
                {
                    "name": "ALAMOGORDO VA CLINIC",
                    "stationNumber": "501GI",
                    "npi": "1053361634",
                    "visn": "22",
                    "ien": 50115
                },
                {
                    "name": "DURANGO VA CLINIC",
                    "stationNumber": "501GJ",
                    "npi": "1073566527",
                    "visn": "22",
                    "ien": 50116
                },
                {
                    "name": "SANTA FE VA CLINIC",
                    "stationNumber": "501GK",
                    "npi": "1295773331",
                    "visn": "22",
                    "ien": 50117
                },
                {
                    "name": "ORTHWEST METRO VA CLINIC",
                    "stationNumber": "501GM",
                    "officialName": "NORTHWEST METRO VA CLINIC",
                    "npi": "1740548494",
                    "visn": "22",
                    "ien": 50118
                },
                {
                    "name": "TAOS VA CLINIC",
                    "stationNumber": "501GN",
                    "npi": "1518374594",
                    "visn": "22",
                    "ien": 50119
                },
                {
                    "name": "RATON CBOC",
                    "stationNumber": "501HB",
                    "officialName": "RATON (CBOC)",
                    "npi": "1407894546",
                    "visn": "22",
                    "ien": 50120
                },
                {
                    "name": "IDES KIRTLAND AFB",
                    "stationNumber": "501MA",
                    "visn": "22",
                    "ien": 50121
                },
                {
                    "name": "COOPERATIVE STUDIES PROGRAM",
                    "stationNumber": "501MI",
                    "officialName": "COOPERATIVE STUDIES PROGRAM CLINICAL RESEARCH PHARM COORDINATING CENTER",
                    "visn": "22",
                    "ien": 50122
                },
                {
                    "name": "ALBUQUERQUE/GALLUP (PRRTP)",
                    "stationNumber": "501PA",
                    "visn": "22",
                    "ien": 50123
                },
                {
                    "name": "ALBUQUERQUE SARRTP",
                    "stationNumber": "501PB",
                    "officialName": "ALBUQUERQUE SUBSTANCE ABUSE REHAB TREATMENT PROG",
                    "visn": "22",
                    "ien": 50124
                },
                {
                    "name": "ALBUQUERQUE CWT/TR",
                    "stationNumber": "501PC",
                    "visn": "22",
                    "ien": 50125
                },
                {
                    "name": "NEW MEXICO (SCI)",
                    "stationNumber": "501SCI",
                    "officialName": "NEW MEXICO HCS SPINAL CORD OUTCOMES (SCI)",
                    "visn": "22",
                    "ien": 50126
                },
                {
                    "name": "KIRTLAND AFB",
                    "stationNumber": "501SG",
                    "visn": "22",
                    "ien": 50127
                }
            ],
            "600": [
                {
                    "name": "LONG BEACH VAMC",
                    "stationNumber": "600",
                    "officialName": "VA LONG BEACH HEALTHCARE SYSTEM",
                    "npi": "1003866468",
                    "visn": "22",
                    "ien": "600"
                },
                {
                    "name": "LONG BEACH NHCU",
                    "stationNumber": "6009AA",
                    "npi": "1730549049",
                    "visn": "22",
                    "ien": 60001
                },
                {
                    "name": "ANAHEIM,CA CBOC",
                    "stationNumber": "600GA",
                    "officialName": "ANAHEIM VETERANS HEALTH CLINIC",
                    "npi": "1699726851",
                    "visn": "22",
                    "ien": 60002
                },
                {
                    "name": "SANTA ANA,CA CBOC",
                    "stationNumber": "600GB",
                    "officialName": "SANTA ANA VETERANS HEALTH CLINIC",
                    "npi": "1598703563",
                    "visn": "22",
                    "ien": 60003
                },
                {
                    "name": "CABRILLO,CA CBOC",
                    "stationNumber": "600GC",
                    "officialName": "CABRILLO VETERANS HEALTH CLINIC",
                    "npi": "1053362756",
                    "visn": "22",
                    "ien": 60004
                },
                {
                    "name": "SANTA FE SPRINGS VA CLINIC",
                    "stationNumber": "600GD",
                    "npi": "1841239712",
                    "visn": "22",
                    "ien": 60005
                },
                {
                    "name": "LAGUNA HILLS VET HLTH CLINIC",
                    "stationNumber": "600GE",
                    "npi": "1700060845",
                    "visn": "22",
                    "ien": 60006
                },
                {
                    "name": "GARDENA VA CLINIC",
                    "stationNumber": "600GF",
                    "npi": "1396378030",
                    "visn": "22",
                    "ien": 60007
                },
                {
                    "name": "LONG BEACH VA CLINIC",
                    "stationNumber": "600QA",
                    "npi": "1538550694",
                    "visn": "22",
                    "ien": 60008
                },
                {
                    "name": "LONG BEACH VA MOBILE CLINIC",
                    "stationNumber": "600QB",
                    "npi": "1386041754",
                    "visn": "22",
                    "ien": 60009
                },
                {
                    "name": "LONG BEACH (SCI)",
                    "stationNumber": "600SCI",
                    "officialName": "LONG BEACH SPINAL CORD OUTCOMES (SCI)",
                    "visn": "22",
                    "ien": 60010
                }
            ],
            "605": [
                {
                    "name": "LOMA LINDA HCS",
                    "stationNumber": "605",
                    "npi": "1609827153",
                    "domain": "LOMA-LINDA.MED.VA.GOV",
                    "visn": "22",
                    "ien": "605",
                    "mn": "LOM"
                },
                {
                    "name": "LOMA LINDA NHC",
                    "stationNumber": "6059AA",
                    "officialName": "LOMA LINDA NURSING HOME CARE",
                    "visn": "22",
                    "ien": 60501
                },
                {
                    "name": "LOMA LINDA VA CLINIC",
                    "stationNumber": "605BZ",
                    "npi": "1225585870",
                    "visn": "22",
                    "ien": 60502
                },
                {
                    "name": "LOMA LINDA CIVH",
                    "stationNumber": "605DS",
                    "visn": "22",
                    "ien": 60503
                },
                {
                    "name": "BARSTOW STNB",
                    "stationNumber": "605DT",
                    "officialName": "VETERANS HOME OF CALIFORNIA - BARSTOW",
                    "visn": "22",
                    "ien": 60504
                },
                {
                    "name": "VICTORVILLE CBOC",
                    "stationNumber": "605GA",
                    "npi": "1689613507",
                    "visn": "22",
                    "ien": 60505
                },
                {
                    "name": "MURRIETA CBOC",
                    "stationNumber": "605GB",
                    "npi": "1871539981",
                    "visn": "22",
                    "ien": 60506
                },
                {
                    "name": "PALM DESERT CBOC",
                    "stationNumber": "605GC",
                    "npi": "1376581876",
                    "visn": "22",
                    "ien": 60507
                },
                {
                    "name": "CORONA CBOC",
                    "stationNumber": "605GD",
                    "npi": "1750333845",
                    "visn": "22",
                    "ien": 60508
                },
                {
                    "name": "RANCHO CUCAMONGA CBOC",
                    "stationNumber": "605GE",
                    "npi": "1982644373",
                    "visn": "22",
                    "ien": 60509
                },
                {
                    "name": "IDES 29 PALMS MARINE BASE",
                    "stationNumber": "605MA",
                    "visn": "22",
                    "ien": 60510
                },
                {
                    "name": "IDES FORT IRWIN ARMY BASE",
                    "stationNumber": "605MB",
                    "visn": "22",
                    "ien": 60511
                },
                {
                    "name": "BLYTHE VA CLINIC",
                    "stationNumber": "605QA",
                    "npi": "1548651516",
                    "visn": "22",
                    "ien": 60512
                }
            ],
            "644": [
                {
                    "name": "PHOENIX VAMC",
                    "stationNumber": "644",
                    "npi": "1437109832",
                    "domain": "PHOENIX.MED.VA.GOV",
                    "visn": "22",
                    "ien": "644",
                    "mn": "PHO"
                },
                {
                    "name": "PHOENIX NHC",
                    "stationNumber": "6449AA",
                    "visn": "22",
                    "ien": 64401
                },
                {
                    "name": "PHOENIX VA FACILITY DOM",
                    "stationNumber": "644BU",
                    "visn": "22",
                    "ien": 64402
                },
                {
                    "name": "SOUTHEAST CBOC",
                    "stationNumber": "644BY",
                    "npi": "1437197902",
                    "visn": "22",
                    "ien": 64403
                },
                {
                    "name": "PHOENIX ARMY HOSP",
                    "stationNumber": "644CN",
                    "visn": "22",
                    "ien": 64404
                },
                {
                    "name": "PHOENIX AFB HOSP",
                    "stationNumber": "644CZ",
                    "visn": "22",
                    "ien": 64405
                },
                {
                    "name": "PHOENIX STATE DOM",
                    "stationNumber": "644DT",
                    "visn": "22",
                    "ien": 64406
                },
                {
                    "name": "SUN CITY, AZ CBOC",
                    "stationNumber": "644GA",
                    "npi": "1598704512",
                    "visn": "22",
                    "ien": 64407
                },
                {
                    "name": "SHOW LOW CBOC",
                    "stationNumber": "644GB",
                    "npi": "1205874351",
                    "visn": "22",
                    "ien": 64408
                },
                {
                    "name": "SOUTHWEST CBOC",
                    "stationNumber": "644GC",
                    "npi": "1831138270",
                    "visn": "22",
                    "ien": 64409
                },
                {
                    "name": "PAYSON CBOC",
                    "stationNumber": "644GD",
                    "npi": "1215974696",
                    "visn": "22",
                    "ien": 64410
                },
                {
                    "name": "THUNDERBIRD VA CBOC",
                    "stationNumber": "644GE",
                    "npi": "1033345624",
                    "visn": "22",
                    "ien": 64411
                },
                {
                    "name": "GLOBE HEALTH CARE CLINIC",
                    "stationNumber": "644GF",
                    "npi": "1053586545",
                    "visn": "22",
                    "ien": 64412
                },
                {
                    "name": "N EAST PHOENIX VA CLINIC",
                    "stationNumber": "644GG",
                    "npi": "1265785364",
                    "ien": 64413,
                    "visn": "22"
                },
                {
                    "name": "PHOENIX MIDTOWN VA CLINIC",
                    "stationNumber": "644GH",
                    "npi": "1174065791",
                    "visn": "22",
                    "ien": 64414
                },
                {
                    "name": "IDES LUKE AFB",
                    "stationNumber": "644MA",
                    "visn": "22",
                    "ien": 64415
                },
                {
                    "name": "CARL T HAYDEN SARRTP",
                    "stationNumber": "644PB",
                    "npi": "1104286418",
                    "visn": "22",
                    "ien": 64416
                },
                {
                    "name": "PHOENIX VA CLINIC",
                    "stationNumber": "644QA",
                    "npi": "1619352077",
                    "visn": "22",
                    "ien": 64417
                },
                {
                    "name": "PHOENIX VA MOBILE CLINIC",
                    "stationNumber": "644QB",
                    "npi": "1629582184",
                    "visn": "22",
                    "ien": 64418
                }
            ],
            "649": [
                {
                    "name": "NORTHERN ARIZONA HCS",
                    "stationNumber": "649",
                    "officialName": "NORTHERN ARIZONA HEALTH CARE SYSTEM - PRESCOTT DIVISION",
                    "npi": "1275583684",
                    "domain": "PRESCOTT.MED.VA.GOV",
                    "visn": "22",
                    "ien": "649",
                    "mn": "PRE"
                },
                {
                    "name": "PRESCOTT NHC",
                    "stationNumber": "6499AA",
                    "visn": "22",
                    "ien": 64901
                },
                {
                    "name": "PRESCOTT DOM",
                    "stationNumber": "649BU",
                    "visn": "22",
                    "ien": 64902
                },
                {
                    "name": "KINGMAN CBOC",
                    "stationNumber": "649GA",
                    "npi": "1770539561",
                    "visn": "22",
                    "ien": 64903
                },
                {
                    "name": "FLAGSTAFF CBOC",
                    "stationNumber": "649GB",
                    "npi": "1952345076",
                    "visn": "22",
                    "ien": 64904
                },
                {
                    "name": "LAKE HAVASU CITY CBOC",
                    "stationNumber": "649GC",
                    "officialName": "Prescott CBOC Lake Havasu City",
                    "npi": "1720034697",
                    "visn": "22",
                    "ien": 64905
                },
                {
                    "name": "ANTHEM CBOC",
                    "stationNumber": "649GD",
                    "npi": "1477504975",
                    "visn": "22",
                    "ien": 64906
                },
                {
                    "name": "COTTONWOOD CBOC",
                    "stationNumber": "649GE",
                    "npi": "1477505493",
                    "visn": "22",
                    "ien": 64907
                },
                {
                    "name": "PRESCOTT MORC",
                    "stationNumber": "649HK",
                    "visn": "22",
                    "ien": 64908
                },
                {
                    "name": "CHINLE VA CLINIC",
                    "stationNumber": "649QA",
                    "npi": "1891183182",
                    "visn": "22",
                    "ien": 64909
                },
                {
                    "name": "HOLBROOK VA CLINIC",
                    "stationNumber": "649QB",
                    "npi": "1558759837",
                    "visn": "22",
                    "ien": 64910
                },
                {
                    "name": "PAGE VA CLINIC",
                    "stationNumber": "649QD",
                    "npi": "1356739650",
                    "visn": "22",
                    "ien": 64911
                },
                {
                    "name": "PRESCOTT VA CLINIC",
                    "stationNumber": "649QE",
                    "npi": "1265820567",
                    "visn": "22",
                    "ien": 64912
                },
                {
                    "name": "TUBA CITY VA CLINIC",
                    "stationNumber": "649QF",
                    "npi": "1437547734",
                    "visn": "22",
                    "ien": 64913
                },
                {
                    "name": "POLACCA VA CLINIC",
                    "stationNumber": "649QG",
                    "npi": "1669857975",
                    "visn": "22",
                    "ien": 64914
                },
                {
                    "name": "KAYENTA VA CLINIC",
                    "stationNumber": "649QH",
                    "npi": "1184009490",
                    "visn": "22",
                    "ien": 64915
                }
            ],
            "664": [
                {
                    "name": "SAN DIEGO HCS",
                    "stationNumber": "664",
                    "officialName": "VA SAN DIEGO HEALTHCARE SYSTEM (664)",
                    "npi": "1891747333",
                    "visn": "22",
                    "ien": "664"
                },
                {
                    "name": "SAN DIEGO NHC",
                    "stationNumber": "6649AA",
                    "visn": "22",
                    "ien": 66401
                },
                {
                    "name": "STNH CHULA VISTA",
                    "stationNumber": "6649AF",
                    "visn": "22",
                    "ien": 66402
                },
                {
                    "name": "SAN DIEGO DOMICILIARY",
                    "stationNumber": "664BU",
                    "visn": "22",
                    "ien": 66403
                },
                {
                    "name": "SAN DIEGO VA DOMICILIARY",
                    "stationNumber": "664BV",
                    "visn": "22",
                    "ien": 66404
                },
                {
                    "name": "MISSION VALLEY",
                    "stationNumber": "664BY",
                    "officialName": "MISSION VALLEY SOC",
                    "npi": "1629015060",
                    "visn": "22",
                    "ien": 66405
                },
                {
                    "name": "NAVAL MEDICAL CENTER SAN DIEGO",
                    "stationNumber": "664CU",
                    "visn": "22",
                    "ien": 66406
                },
                {
                    "name": "STHOME DOMICILLIARY CHULA VSTA",
                    "stationNumber": "664DT",
                    "visn": "22",
                    "ien": 66407
                },
                {
                    "name": "IMPERIAL VALLEY CBOC",
                    "stationNumber": "664GA",
                    "npi": "1366498313",
                    "visn": "22",
                    "ien": 66408
                },
                {
                    "name": "OCEANSIDE CBOC",
                    "stationNumber": "664GB",
                    "npi": "1568401404",
                    "visn": "22",
                    "ien": 66409
                },
                {
                    "name": "CHULA VISTA",
                    "stationNumber": "664GC",
                    "officialName": "CHULA VISTA CBOC",
                    "npi": "1821040965",
                    "visn": "22",
                    "ien": 66410
                },
                {
                    "name": "ESCONDIDO",
                    "stationNumber": "664GD",
                    "officialName": "ESCONDIDO CBOC",
                    "npi": "1538112552",
                    "visn": "22",
                    "ien": 66411
                },
                {
                    "name": "IDES BALBOA NAVAL HOSP",
                    "stationNumber": "664MA",
                    "visn": "22",
                    "ien": 66412
                },
                {
                    "name": "IDES CAMP PENDLETON",
                    "stationNumber": "664MB",
                    "visn": "22",
                    "ien": 66413
                },
                {
                    "name": "RIO VA CLINIC",
                    "stationNumber": "664QA",
                    "officialName": "MISSION GORGE VA CLINIC",
                    "npi": "1265823231",
                    "visn": "22",
                    "ien": 66414
                },
                {
                    "name": "SAN DIEGO SPINAL CORD OUTCOMES",
                    "stationNumber": "664SCI",
                    "officialName": "SAN DIEGO SPINAL CORD OUTCOMES (SCI)",
                    "visn": "22",
                    "ien": 66415
                }
            ],
            "678": [
                {
                    "name": "SOUTHERN ARIZONA VA HCS",
                    "stationNumber": "678",
                    "officialName": "SOUTHERN ARIZONA HEALTH CARE SYSTEM - TUCSON DIVISION",
                    "npi": "1568412880",
                    "domain": "TUCSON.MED.VA.GOV",
                    "visn": "22",
                    "ien": "678",
                    "mn": "TUC"
                },
                {
                    "name": "TUCSON NHC",
                    "stationNumber": "6789AA",
                    "visn": "22",
                    "ien": 67801
                },
                {
                    "name": "ARIZONA STATE VET NURSING HOME",
                    "stationNumber": "6789F",
                    "visn": "22",
                    "ien": 67802
                },
                {
                    "name": "TUCSON VA FACILITY DOM",
                    "stationNumber": "678BU",
                    "officialName": "TUCSON VA FACILITY DOMICILIARY",
                    "visn": "22",
                    "ien": 67803
                },
                {
                    "name": "SIERRA VISTA CBOC",
                    "stationNumber": "678GA",
                    "npi": "1235173931",
                    "visn": "22",
                    "ien": 67804
                },
                {
                    "name": "YUMA CBOC",
                    "stationNumber": "678GB",
                    "npi": "1649219528",
                    "visn": "22",
                    "ien": 67805
                },
                {
                    "name": "CASA GRANDE CBOC",
                    "stationNumber": "678GC",
                    "npi": "1083665871",
                    "visn": "22",
                    "ien": 67806
                },
                {
                    "name": "SAFFORD CBOC",
                    "stationNumber": "678GD",
                    "npi": "1902844632",
                    "visn": "22",
                    "ien": 67807
                },
                {
                    "name": "GREEN VALLEY CBOC",
                    "stationNumber": "678GE",
                    "npi": "1164477832",
                    "visn": "22",
                    "ien": 67808
                },
                {
                    "name": "NORTHWEST TUCSON URBAN CBOC",
                    "stationNumber": "678GF",
                    "npi": "1366648917",
                    "visn": "22",
                    "ien": 67809
                },
                {
                    "name": "SOUTHEAST TUCSON CBOC",
                    "stationNumber": "678GG",
                    "officialName": "SOUTHEAST TUCSON",
                    "npi": "1366606410",
                    "visn": "22",
                    "ien": 67810
                },
                {
                    "name": "IDES DAVIS-MONTHAN AFB",
                    "stationNumber": "678MA",
                    "visn": "22",
                    "ien": 67811
                },
                {
                    "name": "IDES FORT HUACHUCA ARMY BASE",
                    "stationNumber": "678MB",
                    "visn": "22",
                    "ien": 67812
                },
                {
                    "name": "TUC-PRRTP",
                    "stationNumber": "678PA",
                    "visn": "22",
                    "ien": 67813
                },
                {
                    "name": "COCHISE COUNTY VA CLINIC",
                    "stationNumber": "678QA",
                    "npi": "1104366897",
                    "visn": "22",
                    "ien": 67814
                },
                {
                    "name": "PINAL COUNTY VA CLINIC",
                    "stationNumber": "678QB",
                    "npi": "1013457704",
                    "visn": "22",
                    "ien": 67815
                }
            ],
            "691": [
                {
                    "name": "WEST LA VAMC",
                    "stationNumber": "691",
                    "officialName": "VA GREATER LOS ANGELES HEALTHCARE SYSTEM - WEST LOS ANGELES DIVISION",
                    "npi": "1326097775",
                    "visn": "22",
                    "ien": "691"
                },
                {
                    "name": "LOS ANGELES",
                    "stationNumber": "6919AA",
                    "visn": "22",
                    "ien": 69101
                },
                {
                    "name": "SEPULVEDA VAMC",
                    "stationNumber": "691A4",
                    "officialName": "SEPULVEDA OUTPATIENT & NURSING HOME CARE",
                    "npi": "1285676387",
                    "visn": "22",
                    "ien": 69102
                },
                {
                    "name": "NHCU-WLA",
                    "stationNumber": "691AA",
                    "visn": "22",
                    "ien": 69103
                },
                {
                    "name": "NHCU-SEP",
                    "stationNumber": "691AB",
                    "visn": "22",
                    "ien": 69104
                },
                {
                    "name": "WLA DOMICILIARY",
                    "stationNumber": "691BU",
                    "officialName": "WEST LOS ANGELES VA DOMICILIARY",
                    "visn": "22",
                    "ien": 69105
                },
                {
                    "name": "METHADONE/LAAM/SEP",
                    "stationNumber": "691CA",
                    "officialName": "SEPULVEDA DRUG TREATMENT PHARMACY",
                    "visn": "22",
                    "ien": 69106
                },
                {
                    "name": "RADIOPHARMACY/SEP",
                    "stationNumber": "691CD",
                    "officialName": "SEPULVEDA RADIOPHARMACY",
                    "visn": "22",
                    "ien": 69107
                },
                {
                    "name": "LOS ANGELES BASED PHARM",
                    "stationNumber": "691CE",
                    "officialName": "LOS ANGELES OUTPATIENT PHARMACY",
                    "visn": "22",
                    "ien": 69108
                },
                {
                    "name": "VET STATE HOME CALIF-VENTURA",
                    "stationNumber": "691DT",
                    "visn": "22",
                    "ien": 69109
                },
                {
                    "name": "WILLIAM J PETE KNIGHT VET HOME",
                    "stationNumber": "691DU",
                    "visn": "22",
                    "ien": 69110
                },
                {
                    "name": "VET STATE HOME CALIF-WEST LA",
                    "stationNumber": "691DV",
                    "visn": "22",
                    "ien": 69111
                },
                {
                    "name": "SANTA BARBARA CBOC",
                    "stationNumber": "691GB",
                    "npi": "1790727600",
                    "visn": "22",
                    "ien": 69112
                },
                {
                    "name": "GARDENA CBOC (691GC)",
                    "stationNumber": "691GC",
                    "officialName": "GARDENA CBOC",
                    "npi": "1780638254",
                    "visn": "22",
                    "ien": 69113
                },
                {
                    "name": "BAKERSFIELD",
                    "stationNumber": "691GD",
                    "officialName": "BAKERSFIELD CBOC",
                    "npi": "1821041633",
                    "visn": "22",
                    "ien": 69114
                },
                {
                    "name": "LAOPC",
                    "stationNumber": "691GE",
                    "officialName": "LOS ANGELES CBOC",
                    "npi": "1861432437",
                    "visn": "22",
                    "ien": 69115
                },
                {
                    "name": "EAST LOS ANGELES",
                    "stationNumber": "691GF",
                    "officialName": "EAST LOS ANGELES CBOC",
                    "npi": "1477508950",
                    "visn": "22",
                    "ien": 69116
                },
                {
                    "name": "ANTELOPE VALLEY VA CLINIC",
                    "stationNumber": "691GG",
                    "npi": "1295788925",
                    "visn": "22",
                    "ien": 69117
                },
                {
                    "name": "SAN LUIS OBISPO",
                    "stationNumber": "691GK",
                    "officialName": "SAN LUIS OBISPO CBOC",
                    "npi": "1174561799",
                    "visn": "22",
                    "ien": 69118
                },
                {
                    "name": "SANTA MARIA CBOC",
                    "stationNumber": "691GL",
                    "npi": "1124075254",
                    "visn": "22",
                    "ien": 69119
                },
                {
                    "name": "OXNARD CBOC",
                    "stationNumber": "691GM",
                    "npi": "1588606321",
                    "visn": "22",
                    "ien": 69120
                },
                {
                    "name": "LYNWOOD CBOC",
                    "stationNumber": "691GN",
                    "npi": "1770524530",
                    "visn": "22",
                    "ien": 69121
                },
                {
                    "name": "SAN GABRIEL VALLEY VA CLINIC",
                    "stationNumber": "691GP",
                    "npi": "1841831617",
                    "visn": "22",
                    "ien": 69122
                },
                {
                    "name": "IDES VANDENBERG AFB",
                    "stationNumber": "691MA",
                    "visn": "22",
                    "ien": 69123
                },
                {
                    "name": "IDES EDWARDS AFB",
                    "stationNumber": "691MB",
                    "visn": "22",
                    "ien": 69124
                },
                {
                    "name": "IDES LOS ANGELES AFB",
                    "stationNumber": "691MC",
                    "visn": "22",
                    "ien": 69125
                }
            ]
        },
        "16": {
            "502": [
                {
                    "name": "ALEXANDRIA VA MEDICAL CENTER",
                    "stationNumber": "502",
                    "npi": "1992754055",
                    "visn": "16",
                    "ien": "502"
                },
                {
                    "name": "ALEXANDRIA, LA (NHCU)",
                    "stationNumber": "5029AA",
                    "visn": "16",
                    "ien": 50201
                },
                {
                    "name": "SW LOUISIANA WAR VET HOME",
                    "stationNumber": "502DT",
                    "officialName": "SOUTHWEST LOUISIANA WAR VETERANS HOME",
                    "visn": "16",
                    "ien": 50202
                },
                {
                    "name": "JENNINGS CBOC",
                    "stationNumber": "502GA",
                    "npi": "1093761462",
                    "visn": "16",
                    "ien": 50203
                },
                {
                    "name": "LAFAYETTE VA CLINIC",
                    "stationNumber": "502GB",
                    "npi": "1275589137",
                    "visn": "16",
                    "ien": 50204
                },
                {
                    "name": "LAKE CHARLES VA CLINIC",
                    "stationNumber": "502GE",
                    "npi": "1811355779",
                    "visn": "16",
                    "ien": 50205
                },
                {
                    "name": "FORT POLK VA CLINIC",
                    "stationNumber": "502GF",
                    "npi": "1245513506",
                    "visn": "16",
                    "ien": 50206
                },
                {
                    "name": "NATCHITOCHES VA CLINIC",
                    "stationNumber": "502GG",
                    "visn": "16",
                    "ien": 50207
                },
                {
                    "name": "IDES FORT POLK",
                    "stationNumber": "502MA",
                    "visn": "16",
                    "ien": 50208
                },
                {
                    "name": "LAFAYETTE CAMPUS B VA CLINIC",
                    "stationNumber": "502QB",
                    "npi": "1366896714",
                    "visn": "16",
                    "ien": 50209
                }
            ],
            "520": [
                {
                    "name": "BILOXI VA MEDICAL CENTER",
                    "stationNumber": "520",
                    "npi": "1811946817",
                    "domain": "BILOXI.MED.VA.GOV",
                    "visn": "16",
                    "ien": "520",
                    "mn": "BIL"
                },
                {
                    "name": "BILOXI NURSING HOME",
                    "stationNumber": "5209AA",
                    "visn": "16",
                    "ien": 52001
                },
                {
                    "name": "GULFPORT DIV ",
                    "stationNumber": "520A0",
                    "officialName": "GULFPORT VAMC",
                    "npi": "1265481477",
                    "visn": "16",
                    "ien": 52002
                },
                {
                    "name": "BILOXI DOMICILIARY",
                    "stationNumber": "520BU",
                    "visn": "16",
                    "ien": 52003
                },
                {
                    "name": "BILOXI (OCS)",
                    "stationNumber": "520BY",
                    "officialName": "BILOXI OPC",
                    "visn": "16",
                    "ien": 52004
                },
                {
                    "name": "PENSACOLA VA CLINIC",
                    "stationNumber": "520BZ",
                    "npi": "1942249339",
                    "visn": "16",
                    "ien": 52005
                },
                {
                    "name": "GULFPORT (PHARM)",
                    "stationNumber": "520CA",
                    "visn": "16",
                    "ien": 52006
                },
                {
                    "name": "BILOXI (PHARM)",
                    "stationNumber": "520CB",
                    "visn": "16",
                    "ien": 52007
                },
                {
                    "name": "CLIFFORD SIMS SVH",
                    "stationNumber": "520DT",
                    "visn": "16",
                    "ien": 52008
                },
                {
                    "name": "MOBILE OPC",
                    "stationNumber": "520GA",
                    "npi": "1003857764",
                    "visn": "16",
                    "ien": 52009
                },
                {
                    "name": "PANAMA CITY BEACH VA CLINIC",
                    "stationNumber": "520GB",
                    "npi": "1710925219",
                    "visn": "16",
                    "ien": 52010
                },
                {
                    "name": "EGLIN AIR FORCE BASE VA CLINIC",
                    "stationNumber": "520GC",
                    "npi": "1760657084",
                    "visn": "16",
                    "ien": 52011
                },
                {
                    "name": "IDES BILOXI",
                    "stationNumber": "520MA",
                    "visn": "16",
                    "ien": 52012
                },
                {
                    "name": "IDES PENSACOLA",
                    "stationNumber": "520MB",
                    "visn": "16",
                    "ien": 52013
                },
                {
                    "name": "IDES KEESLER AFB",
                    "stationNumber": "520MC",
                    "visn": "16",
                    "ien": 52014
                },
                {
                    "name": "IDES GULFPORT (NAVAL)",
                    "stationNumber": "520MD",
                    "visn": "16",
                    "ien": 52015
                },
                {
                    "name": "IDES NAVAL HOSPITAL PENSACOLA",
                    "stationNumber": "520ME",
                    "visn": "16",
                    "ien": 52016
                },
                {
                    "name": "IDES HURLBURT FIELD",
                    "stationNumber": "520MF",
                    "visn": "16",
                    "ien": 52017
                },
                {
                    "name": "IDES EGLIN AFB",
                    "stationNumber": "520MG",
                    "visn": "16",
                    "ien": 52018
                },
                {
                    "name": "IDES PANAMA CITY (NAVAL)",
                    "stationNumber": "520MH",
                    "visn": "16",
                    "ien": 52019
                },
                {
                    "name": "IDES TYNDALL AFB",
                    "stationNumber": "520MI",
                    "visn": "16",
                    "ien": 52020
                },
                {
                    "name": "PRRTP",
                    "stationNumber": "520PA",
                    "officialName": "GULFPORT PRRTP",
                    "visn": "16",
                    "ien": 52021
                },
                {
                    "name": "PANAMA CITY BEACH WEST VA CLIN",
                    "stationNumber": "520QA",
                    "npi": "1326436601",
                    "visn": "16",
                    "ien": 52022
                }
            ],
            "564": [
                {
                    "name": "FAYETTEVILLE VA MEDICAL",
                    "stationNumber": "564",
                    "npi": "1003867326",
                    "domain": "FAYETTVL-AR.MED.VA.GOV",
                    "visn": "16",
                    "ien": "564",
                    "mn": "FAY"
                },
                {
                    "name": "MISSOURI STATE VET HOME-MT VER",
                    "stationNumber": "5649AF",
                    "visn": "16",
                    "ien": 56401
                },
                {
                    "name": "LEROY R. POND MH RRTP",
                    "stationNumber": "564BU",
                    "ien": 56402,
                    "visn": "16"
                },
                {
                    "name": "GENE TAYLOR VA OPC",
                    "stationNumber": "564BY",
                    "npi": "1417900671",
                    "visn": "16",
                    "ien": 56403
                },
                {
                    "name": "FAYETTEVILLE STATE VET HOME",
                    "stationNumber": "564DT",
                    "visn": "16",
                    "ien": 56404
                },
                {
                    "name": "HARRISON VA CLINIC",
                    "stationNumber": "564GA",
                    "npi": "1912952235",
                    "visn": "16",
                    "ien": 56405
                },
                {
                    "name": "FORT SMITH VA CLINIC",
                    "stationNumber": "564GB",
                    "npi": "1003869991",
                    "visn": "16",
                    "ien": 56406
                },
                {
                    "name": "BRANSON CBOC",
                    "stationNumber": "564GC",
                    "npi": "1609024058",
                    "visn": "16",
                    "ien": 56407
                },
                {
                    "name": "OZARK CBOC",
                    "stationNumber": "564GD",
                    "npi": "1831415751",
                    "visn": "16",
                    "ien": 56408
                },
                {
                    "name": "JAY CBOC",
                    "stationNumber": "564GE",
                    "npi": "1659607570",
                    "visn": "16",
                    "ien": 56409
                },
                {
                    "name": "JOPLIN VA CLINIC",
                    "stationNumber": "564GF",
                    "npi": "1992264824",
                    "visn": "16",
                    "ien": 56410
                },
                {
                    "name": "TOWNSHIP VA CLINIC",
                    "stationNumber": "564QA",
                    "npi": "1528456811",
                    "visn": "16",
                    "ien": 56411
                },
                {
                    "name": "SUNBRIDGE VA CLINIC",
                    "stationNumber": "564QB",
                    "npi": "1437547726",
                    "visn": "16",
                    "ien": 56412
                }
            ],
            "580": [
                {
                    "name": "HOUSTON VAMC",
                    "stationNumber": "580",
                    "officialName": "MICHAEL E. DEBAKEY VA MEDICAL CENTER",
                    "npi": "1033168786",
                    "domain": "HOUSTON.MED.VA.GOV",
                    "visn": "16",
                    "ien": "580",
                    "mn": "HOU"
                },
                {
                    "name": "HOUSTON (NHCU)",
                    "stationNumber": "5809AA",
                    "visn": "16",
                    "ien": 58001
                },
                {
                    "name": "DOM CARE FOR HOMELESS VET",
                    "stationNumber": "580BU",
                    "npi": "1114587094",
                    "visn": "16",
                    "ien": 58002
                },
                {
                    "name": "BEAUMONT OPC",
                    "stationNumber": "580BY",
                    "officialName": "BEAUMONT",
                    "npi": "1528019593",
                    "visn": "16",
                    "ien": 58003
                },
                {
                    "name": "LUFKIN",
                    "stationNumber": "580BZ",
                    "officialName": "LUFKIN OUTPATIENT CLINIC",
                    "npi": "1003863861",
                    "visn": "16",
                    "ien": 58004
                },
                {
                    "name": "HOUSTON (CBOC)",
                    "stationNumber": "580GA",
                    "officialName": "HOUSTON CBOC",
                    "npi": "1235309170",
                    "visn": "16",
                    "ien": 58005
                },
                {
                    "name": "GALVESTON COUNTY VA CLINIC",
                    "stationNumber": "580GC",
                    "npi": "1689865552",
                    "visn": "16",
                    "ien": 58006
                },
                {
                    "name": "CONROE CBOC",
                    "stationNumber": "580GD",
                    "npi": "1073637112",
                    "visn": "16",
                    "ien": 58007
                },
                {
                    "name": "KATY CBOC",
                    "stationNumber": "580GE",
                    "npi": "1952640286",
                    "visn": "16",
                    "ien": 58008
                },
                {
                    "name": "LAKE JACKSON CBOC",
                    "stationNumber": "580GF",
                    "npi": "1902186000",
                    "visn": "16",
                    "ien": 58009
                },
                {
                    "name": "RICHMOND CBOC",
                    "stationNumber": "580GG",
                    "npi": "1649596560",
                    "visn": "16",
                    "ien": 58010
                },
                {
                    "name": "TOMBALL CBOC",
                    "stationNumber": "580GH",
                    "npi": "1790117398",
                    "visn": "16",
                    "ien": 58011
                },
                {
                    "name": "TEXAS CITY VA CLINIC",
                    "stationNumber": "580GJ",
                    "npi": "1871900811",
                    "visn": "16",
                    "ien": 58012
                },
                {
                    "name": "HOUSTON VA CLINIC",
                    "stationNumber": "580QA",
                    "visn": "16",
                    "ien": 58013
                },
                {
                    "name": "HOUSTON VA MOBILE CLINIC",
                    "stationNumber": "580QB",
                    "npi": "1063079366",
                    "visn": "16",
                    "ien": 58014
                },
                {
                    "name": "HOUSTON SPINAL CORD OUTCOMES",
                    "stationNumber": "580SCI",
                    "officialName": "HOUSTON SPINAL CORD OUTCOMES (SCI)",
                    "visn": "16",
                    "ien": 58015
                }
            ],
            "598": [
                {
                    "name": "CENTRAL ARKANSAS HCS",
                    "stationNumber": "598",
                    "officialName": "CENTRAL ARKANSAS HEALTH CARE SYSTEM - LITTLE ROCK ",
                    "npi": "1619926268",
                    "domain": "LITTLE-ROCK.MED.VA.GOV",
                    "visn": "16",
                    "ien": "598",
                    "mn": "LIT"
                },
                {
                    "name": "LITTLE ROCK NURSING HOME ",
                    "stationNumber": "5989AA",
                    "officialName": "LITTLE ROCK, AR NURSING HOME CARE UNIT",
                    "visn": "16",
                    "ien": 59801
                },
                {
                    "name": "ARKANSAS STATE VETERANS HOME",
                    "stationNumber": "5989AF",
                    "visn": "16",
                    "ien": 59802
                },
                {
                    "name": "NORTH LITTLE ROCK, AR VANPH",
                    "stationNumber": "598A0",
                    "npi": "1467402826",
                    "visn": "16",
                    "ien": 59803
                },
                {
                    "name": "LITTLE ROCK, AR DOMICILIARY",
                    "stationNumber": "598BU",
                    "visn": "16",
                    "ien": 59804
                },
                {
                    "name": "LITTLE ROCK, AR STATE HOME",
                    "stationNumber": "598DT",
                    "visn": "16",
                    "ien": 59805
                },
                {
                    "name": "MOUNTAIN HOME VA CLINIC",
                    "stationNumber": "598GA",
                    "officialName": "MOUNTAIN HOME, AR CBOC",
                    "npi": "1063450617",
                    "visn": "16",
                    "ien": 59806
                },
                {
                    "name": "EL DORADO, AR (CBOC)",
                    "stationNumber": "598GB",
                    "officialName": "EL DORADO, AR CBOC",
                    "npi": "1215980164",
                    "visn": "16",
                    "ien": 59807
                },
                {
                    "name": "HOT SPRINGS, AR (CBOC)",
                    "stationNumber": "598GC",
                    "officialName": "HOT SPRINGS, AR CBOC",
                    "npi": "1215982582",
                    "visn": "16",
                    "ien": 59808
                },
                {
                    "name": "MENA AR (CBOC)",
                    "stationNumber": "598GD",
                    "officialName": "MENA, AR (CBOC)",
                    "npi": "1861439697",
                    "visn": "16",
                    "ien": 59809
                },
                {
                    "name": "PINE BLUFF CBOC",
                    "stationNumber": "598GE",
                    "npi": "1548457146",
                    "visn": "16",
                    "ien": 59810
                },
                {
                    "name": "SEARCY CBOC",
                    "stationNumber": "598GF",
                    "npi": "1063745891",
                    "visn": "16",
                    "ien": 59811
                },
                {
                    "name": "CONWAY CBOC ",
                    "stationNumber": "598GG",
                    "officialName": "CONWAY CBOC",
                    "visn": "16",
                    "ien": 59812
                },
                {
                    "name": "RUSSELLVILLE CBOC ",
                    "stationNumber": "598GH",
                    "officialName": "RUSSELLVILLE CBOC",
                    "visn": "16",
                    "ien": 59813
                },
                {
                    "name": "IDES LITTLE ROCK AFB",
                    "stationNumber": "598MA",
                    "visn": "16",
                    "ien": 59814
                },
                {
                    "name": "LITTLE ROCK, AR PRRTP",
                    "stationNumber": "598PA",
                    "visn": "16",
                    "ien": 59815
                },
                {
                    "name": "LITTLE ROCK VA CLINIC",
                    "stationNumber": "598QA",
                    "npi": "1528456894",
                    "visn": "16",
                    "ien": 59816
                },
                {
                    "name": "LITTLE ROCK VA MOBILE CLINIC",
                    "stationNumber": "598QB",
                    "npi": "1295120152",
                    "visn": "16",
                    "ien": 59817
                }
            ],
            "629": [
                {
                    "name": "SE LOUISIANA VETERANS HCS",
                    "stationNumber": "629",
                    "officialName": "NEW ORLEANS VAMC",
                    "npi": "1952351389",
                    "domain": "NEW-ORLEANS.MED.VA.GOV",
                    "visn": "16",
                    "ien": "629",
                    "mn": "NEW"
                },
                {
                    "name": "NEW ORL-NHC",
                    "stationNumber": "6299AA",
                    "officialName": "Lindy Boggs Transitional Care Unit",
                    "visn": "16",
                    "ien": 62901
                },
                {
                    "name": "New Orleans-BROPC",
                    "stationNumber": "629BY",
                    "officialName": "VA Baton Rouge Outpatient Clinic",
                    "npi": "1437101235",
                    "visn": "16",
                    "ien": 62902
                },
                {
                    "name": "New Orleans-Jackson",
                    "stationNumber": "629DT",
                    "officialName": "Louisiana War Veterans Home",
                    "visn": "16",
                    "ien": 62903
                },
                {
                    "name": "SOUTHEAST LOUISIANA VET HOME",
                    "stationNumber": "629DU",
                    "visn": "16",
                    "ien": 62904
                },
                {
                    "name": "HOUMA CBOC",
                    "stationNumber": "629GA",
                    "npi": "1245286277",
                    "visn": "16",
                    "ien": 62905
                },
                {
                    "name": "HAMMOND CBOC",
                    "stationNumber": "629GB",
                    "npi": "1003969239",
                    "visn": "16",
                    "ien": 62906
                },
                {
                    "name": "SLIDELL CBOC",
                    "stationNumber": "629GC",
                    "npi": "1033266077",
                    "visn": "16",
                    "ien": 62907
                },
                {
                    "name": "ST. JOHN VA OPC",
                    "stationNumber": "629GD",
                    "officialName": "ST. JOHN VA OUTPATIENT CLINIC",
                    "npi": "1760535058",
                    "visn": "16",
                    "ien": 62908
                },
                {
                    "name": "FRANKLIN CBOC",
                    "stationNumber": "629GE",
                    "npi": "1942530746",
                    "visn": "16",
                    "ien": 62909
                },
                {
                    "name": "BOGALUSA CBOC",
                    "stationNumber": "629GF",
                    "npi": "1023348828",
                    "visn": "16",
                    "ien": 62910
                },
                {
                    "name": "IDES BELLE CHASSE NAVAL HC",
                    "stationNumber": "629MA",
                    "visn": "16",
                    "ien": 62911
                },
                {
                    "name": "NEW ORL-PRRTP",
                    "stationNumber": "629PA",
                    "officialName": "NEW ORLEANS PSYCHOLOGICAL RESIDENTIAL REHABILITATION TREATMENT PROGRAM",
                    "visn": "16",
                    "ien": 62912
                },
                {
                    "name": "BATON ROUGE SOUTH VA CLINIC",
                    "stationNumber": "629QA",
                    "npi": "1134517493",
                    "visn": "16",
                    "ien": 62913
                }
            ],
            "667": [
                {
                    "name": "SHREVEPORT VAMC",
                    "stationNumber": "667",
                    "officialName": "OVERTON BROOKS VA MEDICAL CENTER",
                    "npi": "1891745626",
                    "domain": "SHREVEPORT.MED.VA.GOV",
                    "visn": "16",
                    "ien": "667",
                    "mn": "SHR"
                },
                {
                    "name": "SHREVEPORT ARMY HOSP",
                    "stationNumber": "667CN",
                    "visn": "16",
                    "ien": 66701
                },
                {
                    "name": "NORTH MONRO",
                    "stationNumber": "667DT",
                    "officialName": "NORTHEAST LOUISIANA WAR VETERANS HOME",
                    "visn": "16",
                    "ien": 66702
                },
                {
                    "name": "NW LOUISIANA WAR VETERANS HOME",
                    "stationNumber": "667DU",
                    "officialName": "NORTHWEST LOUISIANA WAR VETERANS HOME",
                    "visn": "16",
                    "ien": 66703
                },
                {
                    "name": "TEXARKANA",
                    "stationNumber": "667GA",
                    "officialName": "TEXARKANA CBOC",
                    "npi": "1417996489",
                    "visn": "16",
                    "ien": 66704
                },
                {
                    "name": "MONROE VA CLINIC",
                    "stationNumber": "667GB",
                    "npi": "1649218090",
                    "visn": "16",
                    "ien": 66705
                },
                {
                    "name": "LONGVIEW",
                    "stationNumber": "667GC",
                    "officialName": "LONGVIEW CBOC",
                    "npi": "1457307522",
                    "visn": "16",
                    "ien": 66706
                },
                {
                    "name": "IDES BARKSDALE AFB",
                    "stationNumber": "667MA",
                    "visn": "16",
                    "ien": 66707
                },
                {
                    "name": "KNIGHT STREET VA CLINIC",
                    "stationNumber": "667QA",
                    "npi": "1639567993",
                    "visn": "16",
                    "ien": 66708
                }
            ]
        },
        "17": {
            "504": [
                {
                    "name": "AMARILLO HCS",
                    "stationNumber": "504",
                    "npi": "1225087794",
                    "domain": "AMARILLO.MED.VA.GOV",
                    "visn": "17",
                    "ien": "504",
                    "mn": "AMA"
                },
                {
                    "name": "AMARILLO (NHCU)",
                    "stationNumber": "5049AA",
                    "officialName": "AMARILLO VA HEALTH CARE CENTER",
                    "visn": "17",
                    "ien": 50401
                },
                {
                    "name": "LUBBOCK",
                    "stationNumber": "504BY",
                    "officialName": "LUBBOCK OUTPATIENT CLINIC",
                    "npi": "1790746188",
                    "visn": "17",
                    "ien": 50402
                },
                {
                    "name": "CLOVIS",
                    "stationNumber": "504BZ",
                    "officialName": "CLOVIS OUTPATIENT CLINIC",
                    "npi": "1821040957",
                    "visn": "17",
                    "ien": 50403
                },
                {
                    "name": "REESE",
                    "stationNumber": "504CZ",
                    "officialName": "REESE AFB",
                    "visn": "17",
                    "ien": 50404
                },
                {
                    "name": "USSCRY-ROAN TEXAS SVH",
                    "stationNumber": "504DT",
                    "officialName": "USSCRY-ROAN TEXAS STATE VETERANS HOME",
                    "visn": "17",
                    "ien": 50405
                },
                {
                    "name": "CHILDRESS VA CLINIC",
                    "stationNumber": "504GA",
                    "npi": "1629022520",
                    "visn": "17",
                    "ien": 50406
                },
                {
                    "name": "CLAYTON",
                    "stationNumber": "504HA",
                    "officialName": "CLAYTON OUTPATIENT CLINIC",
                    "visn": "17",
                    "ien": 50407
                },
                {
                    "name": "DALHART CBOC",
                    "stationNumber": "504HB",
                    "npi": "1225072424",
                    "visn": "17",
                    "ien": 50408
                },
                {
                    "name": "IDES CANNON AFB",
                    "stationNumber": "504MA",
                    "visn": "17",
                    "ien": 50409
                },
                {
                    "name": "AMARILLO (PRRTP)",
                    "stationNumber": "504PA",
                    "officialName": "AMARILLO VA HEALTHCARE SYSTEM",
                    "visn": "17",
                    "ien": 50410
                }
            ],
            "519": [
                {
                    "name": "WEST TEXAS HCS",
                    "stationNumber": "519",
                    "officialName": "WEST TEXAS VA HEALTH CARE SYSTEM - BIG SPRING DIVISION",
                    "npi": "1902855901",
                    "domain": "BIG-SPRING.MED.VA.GOV",
                    "visn": "17",
                    "ien": "519",
                    "mn": "BIG"
                },
                {
                    "name": "BIG SPRING",
                    "stationNumber": "5199AA",
                    "officialName": "BIG SPRING NHC",
                    "visn": "17",
                    "ien": 51901
                },
                {
                    "name": "WEST TEXAS DOMICILIARY",
                    "stationNumber": "519BU",
                    "npi": "1922466713",
                    "visn": "17",
                    "ien": 51902
                },
                {
                    "name": "LAMUN LUSK SANCHEZ ST VET HOME",
                    "stationNumber": "519DT",
                    "officialName": "LLS-SVH",
                    "visn": "17",
                    "ien": 51903
                },
                {
                    "name": "PERMIAN BASIN CBOC",
                    "stationNumber": "519GA",
                    "npi": "1861430217",
                    "visn": "17",
                    "ien": 51904
                },
                {
                    "name": "HOBBS",
                    "stationNumber": "519GB",
                    "officialName": "HOBBS CBOC",
                    "npi": "1437104882",
                    "visn": "17",
                    "ien": 51905
                },
                {
                    "name": "FT STOCKTON CBOC",
                    "stationNumber": "519GD",
                    "officialName": "FT. STOCKTON CBOC",
                    "npi": "1891740098",
                    "visn": "17",
                    "ien": 51906
                },
                {
                    "name": "ABILENE",
                    "stationNumber": "519HC",
                    "officialName": "ABILENE CBOC",
                    "npi": "1699725267",
                    "visn": "17",
                    "ien": 51907
                },
                {
                    "name": "Stamford CBOC ",
                    "stationNumber": "519HD",
                    "officialName": "STAMFORD CBOC",
                    "npi": "1689613630",
                    "visn": "17",
                    "ien": 51908
                },
                {
                    "name": "SAN ANGELO",
                    "stationNumber": "519HF",
                    "officialName": "SAN ANGELO CBOC",
                    "npi": "1346288891",
                    "visn": "17",
                    "ien": 51909
                },
                {
                    "name": "IDES GOODFELLOW AFB",
                    "stationNumber": "519MA",
                    "visn": "17",
                    "ien": 51910
                },
                {
                    "name": "IDES DYESS AFB",
                    "stationNumber": "519MB",
                    "visn": "17",
                    "ien": 51911
                }
            ],
            "549": [
                {
                    "name": "DALLAS VA MEDICAL CENTER",
                    "stationNumber": "549",
                    "npi": "1174572416",
                    "domain": "DALLAS.MED.VA.GOV",
                    "visn": "17",
                    "ien": "549",
                    "mn": "DAL"
                },
                {
                    "name": "DALLAS (NHC)",
                    "stationNumber": "5499AA",
                    "officialName": "NORTH TEXAS HEALTH CARE SYSTEM - DALLAS NHC",
                    "visn": "17",
                    "ien": 54901
                },
                {
                    "name": "BONHAM VA CLC",
                    "stationNumber": "5499AB",
                    "npi": "1437519741",
                    "ien": 54902,
                    "visn": "17"
                },
                {
                    "name": "BONHAM NHCU",
                    "stationNumber": "5499AC",
                    "officialName": "BONHAM NURSING HOME CARE UNIT",
                    "visn": "17",
                    "ien": 54903
                },
                {
                    "name": "SAM RAYBURN MEM VET CENTER",
                    "stationNumber": "549A4",
                    "npi": "1558314948",
                    "visn": "17",
                    "ien": 54904
                },
                {
                    "name": "DALLAS",
                    "stationNumber": "549BU",
                    "officialName": "NORTH TEXAS HEALTH CARE SYSTEM - DALLAS DOM",
                    "visn": "17",
                    "ien": 54905
                },
                {
                    "name": "BONHAM DOMICILIARY (549BV)",
                    "stationNumber": "549BV",
                    "officialName": "NORTH TEXAS HEALTH CARE SYSTEM - BONHAM DOM",
                    "visn": "17",
                    "ien": 54906
                },
                {
                    "name": "FORT WORTH VA CLINIC",
                    "stationNumber": "549BY",
                    "npi": "1285687319",
                    "visn": "17",
                    "ien": 54907
                },
                {
                    "name": "CLYDE W COSPER SVH",
                    "stationNumber": "549DT",
                    "officialName": "NORTH TEXAS HEALTH CARE SYSTEM - COSPER SVH",
                    "visn": "17",
                    "ien": 54908
                },
                {
                    "name": "WATKINS-LOGAN TX STATE VETHOME",
                    "stationNumber": "549DU",
                    "visn": "17",
                    "ien": 54909
                },
                {
                    "name": "TYLER VA CLINIC",
                    "stationNumber": "549GA",
                    "npi": "1306997200",
                    "visn": "17",
                    "ien": 54910
                },
                {
                    "name": "DALLAS CBOC",
                    "stationNumber": "549GB",
                    "officialName": "NORTH TEXAS HEALTH CARE SYSTEM - DALLAS CBOC",
                    "npi": "1750324224",
                    "visn": "17",
                    "ien": 54911
                },
                {
                    "name": "BONHAM VA CLINIC",
                    "stationNumber": "549GC",
                    "npi": "1912958653",
                    "visn": "17",
                    "ien": 54912
                },
                {
                    "name": "DENTON VA CLINIC",
                    "stationNumber": "549GD",
                    "npi": "1972558443",
                    "visn": "17",
                    "ien": 54913
                },
                {
                    "name": "BRIDGEPORT VA CLINIC",
                    "stationNumber": "549GE",
                    "npi": "1568417087",
                    "visn": "17",
                    "ien": 54914
                },
                {
                    "name": "GRANBURY VA CLINIC",
                    "stationNumber": "549GF",
                    "npi": "1972558450",
                    "visn": "17",
                    "ien": 54915
                },
                {
                    "name": "GREENVILLE VA CLINIC TX",
                    "stationNumber": "549GH",
                    "npi": "1811942121",
                    "visn": "17",
                    "ien": 54916
                },
                {
                    "name": "CLEBURNE CBOC",
                    "stationNumber": "549GI",
                    "officialName": "NORTH TEXAS HEALTH CARE SYSTEM - CLEBURNE CBOC",
                    "npi": "1215989355",
                    "visn": "17",
                    "ien": 54917
                },
                {
                    "name": "SHERMAN VA CLINIC",
                    "stationNumber": "549GJ",
                    "npi": "1679731525",
                    "visn": "17",
                    "ien": 54918
                },
                {
                    "name": "POLK STREET VA CLINIC",
                    "stationNumber": "549GK",
                    "npi": "1215344254",
                    "visn": "17",
                    "ien": 54919
                },
                {
                    "name": "PLANO VA CLINIC",
                    "stationNumber": "549GL",
                    "npi": "1508313149",
                    "visn": "17",
                    "ien": 54920
                },
                {
                    "name": "GRAND PRAIRIE VA CLINIC",
                    "stationNumber": "549GM",
                    "npi": "1386129575",
                    "visn": "17",
                    "ien": 54921
                },
                {
                    "name": "NORTH TEXAS VA MOBILE CLINIC",
                    "stationNumber": "549HK",
                    "npi": "1144695305",
                    "visn": "17",
                    "ien": 54922
                },
                {
                    "name": "IDES BRANCH HC FORT WORTH",
                    "stationNumber": "549MA",
                    "visn": "17",
                    "ien": 54923
                },
                {
                    "name": "DALLAS",
                    "stationNumber": "549PA",
                    "officialName": "NORTH TEXAS HEALTH CARE SYSTEM - DALLAS PRRTP",
                    "visn": "17",
                    "ien": 54924
                },
                {
                    "name": "BONHAM CWTRR",
                    "stationNumber": "549PB",
                    "officialName": "BONHAM COMPENSATED WORK THERAPY RESIDENTAL REHAB",
                    "visn": "17",
                    "ien": 54925
                },
                {
                    "name": "DALLAS VA CLINIC",
                    "stationNumber": "549QA",
                    "npi": "1073998431",
                    "visn": "17",
                    "ien": 54926
                },
                {
                    "name": "EAST LANCASTER VA CLINIC",
                    "stationNumber": "549QB",
                    "npi": "1649645672",
                    "visn": "17",
                    "ien": 54927
                },
                {
                    "name": "BROADWAY VA CLINIC",
                    "stationNumber": "549QC",
                    "npi": "1629459730",
                    "visn": "17",
                    "ien": 54928
                },
                {
                    "name": "NORTH TX SPINAL CORD OUTCOMES",
                    "stationNumber": "549SCI",
                    "officialName": "NORTH TEXAS HCS SPINAL CORD OUTCOMES (SCI)",
                    "visn": "17",
                    "ien": 54929
                }
            ],
            "671": [
                {
                    "name": "AUDIE L. MURPHY MEMORIAL HOSP",
                    "stationNumber": "671",
                    "npi": "1922058569",
                    "domain": "SAN-ANTONIO.MED.VA.GOV",
                    "visn": "17",
                    "ien": "671",
                    "mn": "SAN"
                },
                {
                    "name": "SAN ANTONIO NHC",
                    "stationNumber": "6719AA",
                    "visn": "17",
                    "ien": 67101
                },
                {
                    "name": "KERRVILLE NHC",
                    "stationNumber": "6719AB",
                    "visn": "17",
                    "ien": 67102
                },
                {
                    "name": "KERRVILLE VA MEDICAL CENTER",
                    "stationNumber": "671A4",
                    "npi": "1154370708",
                    "visn": "17",
                    "ien": 67103
                },
                {
                    "name": "MC ALLEN OUTPATIENT CLINIC",
                    "stationNumber": "671B0",
                    "npi": "1720021983",
                    "visn": "17",
                    "ien": 67104
                },
                {
                    "name": "SAN ANTONIO VA FACILITY DOM",
                    "stationNumber": "671BU",
                    "officialName": "SAN ANTONIO VA FACILITY DOMICILIARY",
                    "visn": "17",
                    "ien": 67105
                },
                {
                    "name": "FRANK M. TEJEDA VA OPC",
                    "stationNumber": "671BY",
                    "npi": "1720034473",
                    "visn": "17",
                    "ien": 67106
                },
                {
                    "name": "CORPUS CHRISTI",
                    "stationNumber": "671BZ",
                    "officialName": "SOUTH TEXAS HEALTH CARE SYSTEM - CORPUS CHRISTI OUTPATIENT CLINIC",
                    "npi": "1255383253",
                    "visn": "17",
                    "ien": 67107
                },
                {
                    "name": "BROOKE ARMY MEDICAL CENTER",
                    "stationNumber": "671CN",
                    "visn": "17",
                    "ien": 67108
                },
                {
                    "name": "CORPUS CHRISTI NAVAL HOSPITAL",
                    "stationNumber": "671CT",
                    "visn": "17",
                    "ien": 67109
                },
                {
                    "name": "WILFORD HALL MEDICAL CENTER",
                    "stationNumber": "671CZ",
                    "officialName": "WILFORD HALL USAF MEDICAL CENTER",
                    "visn": "17",
                    "ien": 67110
                },
                {
                    "name": "FRANK TEJEDA TX STATE VET HOME",
                    "stationNumber": "671DT",
                    "officialName": "FRANK M TEJEDA TEXAS STATE VETERANS HOME",
                    "visn": "17",
                    "ien": 67111
                },
                {
                    "name": "ALFREDO GONZALEZ STVH",
                    "stationNumber": "671DU",
                    "officialName": "ALFREDO GONZALEZ STATE VETERANS HOME",
                    "visn": "17",
                    "ien": 67112
                },
                {
                    "name": "HARLINGEN CBOC",
                    "stationNumber": "671GA",
                    "officialName": "VA Harlingen Outpatient Clinic ",
                    "npi": "1093868333",
                    "visn": "17",
                    "ien": 67113
                },
                {
                    "name": "VICTORIA VA CLINIC",
                    "stationNumber": "671GB",
                    "npi": "1407896863",
                    "visn": "17",
                    "ien": 67114
                },
                {
                    "name": "DEL RIO VA CLINIC",
                    "stationNumber": "671GC",
                    "npi": "1962731133",
                    "visn": "17",
                    "ien": 67115
                },
                {
                    "name": "EAGLE PASS",
                    "stationNumber": "671GD",
                    "officialName": "UNITED MEDICAL CENTERS",
                    "npi": "1326090093",
                    "visn": "17",
                    "ien": 67116
                },
                {
                    "name": "LAREDO",
                    "stationNumber": "671GE",
                    "officialName": "SOUTH TEXAS HEALTH CARE SYSTEM - LAREDO OUTPATIENT CLINIC",
                    "npi": "1578519328",
                    "visn": "17",
                    "ien": 67117
                },
                {
                    "name": "SOUTH BEXAR COUNTY VA CLINIC",
                    "stationNumber": "671GF",
                    "npi": "1073552006",
                    "visn": "17",
                    "ien": 67118
                },
                {
                    "name": "ALICE",
                    "stationNumber": "671GG",
                    "officialName": "CHRISTUS SPOHN SAN DIEGO FAMILY HEALTH CLINIC",
                    "npi": "1164473385",
                    "visn": "17",
                    "ien": 67119
                },
                {
                    "name": "BEEVILLE VA CLINIC",
                    "stationNumber": "671GH",
                    "npi": "1720031099",
                    "visn": "17",
                    "ien": 67120
                },
                {
                    "name": "KINGSVILLE",
                    "stationNumber": "671GI",
                    "officialName": "CHRISTUS SPOHN BISHOP FAMILY HEALTH CLINIC",
                    "npi": "1306892773",
                    "visn": "17",
                    "ien": 67121
                },
                {
                    "name": "UVALDE VA CLINIC",
                    "stationNumber": "671GJ",
                    "npi": "1154361541",
                    "visn": "17",
                    "ien": 67122
                },
                {
                    "name": "SAN ANTONIO VA CLINIC",
                    "stationNumber": "671GK",
                    "npi": "1447294103",
                    "visn": "17",
                    "ien": 67123
                },
                {
                    "name": "NEW BRAUNFELS VA CLINIC",
                    "stationNumber": "671GL",
                    "npi": "1548207962",
                    "visn": "17",
                    "ien": 67124
                },
                {
                    "name": "SEGUIN VA CLINIC",
                    "stationNumber": "671GN",
                    "npi": "1194180406",
                    "visn": "17",
                    "ien": 67125
                },
                {
                    "name": "NORTH CENTRAL FED VA CLINIC",
                    "stationNumber": "671GO",
                    "npi": "1801083142",
                    "visn": "17",
                    "ien": 67126
                },
                {
                    "name": "BALCONES HEIGHTS VA CLINIC",
                    "stationNumber": "671GP",
                    "npi": "1003224049",
                    "visn": "17",
                    "ien": 67127
                },
                {
                    "name": "SHAVANO PARK VA CLINIC",
                    "stationNumber": "671GQ",
                    "npi": "1962819045",
                    "visn": "17",
                    "ien": 67128
                },
                {
                    "name": "IDES FORT SAM HOUSTON",
                    "stationNumber": "671MA",
                    "visn": "17",
                    "ien": 67129
                },
                {
                    "name": "IDES 59TH MEDWING LACKLAND AFB",
                    "stationNumber": "671MB",
                    "visn": "17",
                    "ien": 67130
                },
                {
                    "name": "SAN ANTONIO PRRTP",
                    "stationNumber": "671PA",
                    "visn": "17",
                    "ien": 67131
                },
                {
                    "name": "SOUTH TEXAS VA MOBILE CLINIC",
                    "stationNumber": "671QA",
                    "npi": "1023406311",
                    "visn": "17",
                    "ien": 67132
                },
                {
                    "name": "DATA POINT VA CLINIC",
                    "stationNumber": "671QB",
                    "npi": "1801284161",
                    "visn": "17",
                    "ien": 67133
                },
                {
                    "name": "SOUTH TX SPINAL CORD OUTCOMES",
                    "stationNumber": "671SCI",
                    "officialName": "SOUTH TEXAS SPINAL CORD OUTCOMES",
                    "visn": "17",
                    "ien": 67134
                }
            ],
            "674": [
                {
                    "name": "OLIN E. TEAGUE VET CENTER",
                    "stationNumber": "674",
                    "npi": "1558311936",
                    "domain": "CENTRAL-TEXAS.MED.VA.GOV",
                    "visn": "17",
                    "ien": "674",
                    "mn": "CEN"
                },
                {
                    "name": "TEMPLE NHC",
                    "stationNumber": "6749AA",
                    "visn": "17",
                    "ien": 67401
                },
                {
                    "name": "WACO NURSING HOME CARE",
                    "stationNumber": "6749AB",
                    "visn": "17",
                    "ien": 67402
                },
                {
                    "name": "DORIS MILLER VAMC",
                    "stationNumber": "674A4",
                    "officialName": "Waco, Tx VAMC",
                    "npi": "1124078563",
                    "visn": "17",
                    "ien": 67403
                },
                {
                    "name": "TEMPLE",
                    "stationNumber": "674BU",
                    "officialName": "Temple, Tx  Domiciliary",
                    "visn": "17",
                    "ien": 67404
                },
                {
                    "name": "WACO VA FACILITY DOM",
                    "stationNumber": "674BV",
                    "visn": "17",
                    "ien": 67405
                },
                {
                    "name": "AUSTIN VA CLINIC",
                    "stationNumber": "674BY",
                    "npi": "1538110739",
                    "visn": "17",
                    "ien": 67406
                },
                {
                    "name": "STATE HOME TX",
                    "stationNumber": "674DT",
                    "officialName": "Temple, Tx  NHC",
                    "visn": "17",
                    "ien": 67407
                },
                {
                    "name": "PALESTINE VA CLINIC",
                    "stationNumber": "674GA",
                    "npi": "1184662736",
                    "visn": "17",
                    "ien": 67408
                },
                {
                    "name": "BROWNWOOD VA CLINIC",
                    "stationNumber": "674GB",
                    "npi": "1548211154",
                    "visn": "17",
                    "ien": 67409
                },
                {
                    "name": "BRYAN VA CLINIC",
                    "stationNumber": "674GC",
                    "npi": "1609828771",
                    "visn": "17",
                    "ien": 67410
                },
                {
                    "name": "CEDAR PARK VA CLINIC",
                    "stationNumber": "674GD",
                    "npi": "1518918309",
                    "visn": "17",
                    "ien": 67411
                },
                {
                    "name": "TEMPLE VA CLINIC",
                    "stationNumber": "674GF",
                    "npi": "1508273673",
                    "visn": "17",
                    "ien": 67412
                },
                {
                    "name": "LAGRANGE VA CLINIC",
                    "stationNumber": "674HB",
                    "visn": "17",
                    "ien": 67413
                },
                {
                    "name": "IDES FORT HOOD",
                    "stationNumber": "674MA",
                    "visn": "17",
                    "ien": 67414
                },
                {
                    "name": "WAC-PRRTP",
                    "stationNumber": "674PA",
                    "officialName": "Waco, Tx  PRRTP",
                    "visn": "17",
                    "ien": 67415
                },
                {
                    "name": "TEMPLE CWT/TR",
                    "stationNumber": "674PB",
                    "visn": "17",
                    "ien": 67416
                }
            ],
            "756": [
                {
                    "name": "EL PASO VA HCS",
                    "stationNumber": "756",
                    "officialName": "EL PASO HCS",
                    "npi": "1275633604",
                    "domain": "EL-PASO.MED.VA.GOV",
                    "visn": "17",
                    "ien": "756",
                    "mn": "ELP"
                },
                {
                    "name": "EL PASO USAH",
                    "stationNumber": "756CN",
                    "visn": "17",
                    "ien": 75601
                },
                {
                    "name": "AMBROSIO GUILLEN TEXAS SVH",
                    "stationNumber": "756DT",
                    "officialName": "AMBROSIO GUILLEN TEXAS STATE VET HOME",
                    "visn": "17",
                    "ien": 75602
                },
                {
                    "name": "LAS CRUCES",
                    "stationNumber": "756GA",
                    "officialName": "LAS CRUCES CBOC",
                    "npi": "1972559730",
                    "visn": "17",
                    "ien": 75603
                },
                {
                    "name": "EL PASO EASTSIDE CBOC",
                    "stationNumber": "756GB",
                    "npi": "1154652287",
                    "visn": "17",
                    "ien": 75604
                },
                {
                    "name": "EL PASO WESTSIDE VA CLINIC",
                    "stationNumber": "756GC",
                    "npi": "1568097491",
                    "visn": "17",
                    "ien": 75605
                },
                {
                    "name": "IDES FT. BLISS",
                    "stationNumber": "756MA",
                    "visn": "17",
                    "ien": 75606
                },
                {
                    "name": "IDES HOLLOMAN AFB",
                    "stationNumber": "756MB",
                    "visn": "17",
                    "ien": 75607
                },
                {
                    "name": "EL PASO SOUTH CENTRAL VA CLIN",
                    "stationNumber": "756QA",
                    "officialName": "EL PASO VA CLINIC",
                    "visn": "17",
                    "ien": 75608
                },
                {
                    "name": "EL PASO CENTRAL VA CLINIC",
                    "stationNumber": "756QB",
                    "npi": "1982178190",
                    "visn": "17",
                    "ien": 75609
                }
            ]
        },
        "10": {
            "506": [
                {
                    "name": "ANN ARBOR VA MEDICAL CENTER",
                    "stationNumber": "506",
                    "npi": "1659321701",
                    "domain": "ANN-ARBOR.MED.VA.GOV",
                    "visn": "10",
                    "ien": "506",
                    "mn": "ANN"
                },
                {
                    "name": "ANN ARBOR NURSING HOME",
                    "stationNumber": "5069AA",
                    "visn": "10",
                    "ien": 50601
                },
                {
                    "name": "TOLEDO VA CLINIC",
                    "stationNumber": "506GA",
                    "npi": "1497795868",
                    "visn": "10",
                    "ien": 50602
                },
                {
                    "name": "FLINT VA CLINIC",
                    "stationNumber": "506GB",
                    "npi": "1245283126",
                    "visn": "10",
                    "ien": 50603
                },
                {
                    "name": "JACKSON MI CBOC",
                    "stationNumber": "506GC",
                    "officialName": "JACKSON CBOC",
                    "npi": "1912952094",
                    "visn": "10",
                    "ien": 50604
                },
                {
                    "name": "PACKARD ROAD VA CLINIC",
                    "stationNumber": "506QA",
                    "npi": "1417345398",
                    "visn": "10",
                    "ien": 50605
                },
                {
                    "name": "GREEN ROAD VA CLINIC",
                    "stationNumber": "506QB",
                    "npi": "1043722325",
                    "visn": "10",
                    "ien": 50606
                }
            ],
            "515": [
                {
                    "name": "BATTLE CREEK VA MEDICAL CENTER",
                    "stationNumber": "515",
                    "npi": "1417906512",
                    "domain": "BATTLE-CREEK.MED.VA.GOV",
                    "visn": "10",
                    "ien": "515",
                    "mn": "BAT"
                },
                {
                    "name": "BATTLE CREEK NURSING HOME",
                    "stationNumber": "5159AA",
                    "officialName": "BATTLE CREEK NH",
                    "visn": "10",
                    "ien": 51501
                },
                {
                    "name": "BATTLE CREEK VA FACILITY DOM",
                    "stationNumber": "515BU",
                    "officialName": "BATTLE CREEK VA FACILITY DOMICILIARY",
                    "visn": "10",
                    "ien": 51502
                },
                {
                    "name": "WYOMING HEALTH CARE CENTER",
                    "stationNumber": "515BY",
                    "npi": "1114966256",
                    "visn": "10",
                    "ien": 51503
                },
                {
                    "name": "GRAND RAPIDS STATE HOME",
                    "stationNumber": "515DT",
                    "visn": "10",
                    "ien": 51504
                },
                {
                    "name": "MUSKEGON VA CLINIC",
                    "stationNumber": "515GA",
                    "npi": "1649218058",
                    "visn": "10",
                    "ien": 51505
                },
                {
                    "name": "LANSING SOUTH VA CLINIC",
                    "stationNumber": "515GB",
                    "officialName": "LANSING CBOC",
                    "npi": "1952357865",
                    "visn": "10",
                    "ien": 51506
                },
                {
                    "name": "BENTON HARBOR VA CLINIC",
                    "stationNumber": "515GC",
                    "npi": "1174576458",
                    "visn": "10",
                    "ien": 51507
                },
                {
                    "name": "BATTLE CREEK-PRRTP",
                    "stationNumber": "515PA",
                    "officialName": "BATTLE CREEK PRRTP",
                    "visn": "10",
                    "ien": 51508
                },
                {
                    "name": "LANSING NORTH VA CLINIC",
                    "stationNumber": "515QA",
                    "npi": "1528456415",
                    "visn": "10",
                    "ien": 51509
                },
                {
                    "name": "CENTURY AVENUE VA CLINIC",
                    "stationNumber": "515QB",
                    "npi": "1316335292",
                    "visn": "10",
                    "ien": 51510
                }
            ],
            "538": [
                {
                    "name": "CHILLICOTHE VA MEDICAL CENTER",
                    "stationNumber": "538",
                    "npi": "1568411395",
                    "domain": "CHILLICOTHE.MED.VA.GOV",
                    "visn": "10",
                    "ien": "538",
                    "mn": "CHI"
                },
                {
                    "name": "CHILLICOTHE NHCU",
                    "stationNumber": "5389AA",
                    "visn": "10",
                    "ien": 53801
                },
                {
                    "name": "CHILLICOTHE, OH DOM",
                    "stationNumber": "538BU",
                    "visn": "10",
                    "ien": 53802
                },
                {
                    "name": "ATHENS VA CLINIC",
                    "stationNumber": "538GA",
                    "npi": "1245281542",
                    "visn": "10",
                    "ien": 53803
                },
                {
                    "name": "PORTSMOUTH VA CLINIC",
                    "stationNumber": "538GB",
                    "npi": "1992742241",
                    "visn": "10",
                    "ien": 53804
                },
                {
                    "name": "MARIETTA VA CLINIC",
                    "stationNumber": "538GC",
                    "npi": "1730136375",
                    "visn": "10",
                    "ien": 53805
                },
                {
                    "name": "LANCASTER VA CLINIC",
                    "stationNumber": "538GD",
                    "npi": "1336195189",
                    "visn": "10",
                    "ien": 53806
                },
                {
                    "name": "CAMBRIDGE VA CLINIC",
                    "stationNumber": "538GE",
                    "npi": "1245387356",
                    "visn": "10",
                    "ien": 53807
                },
                {
                    "name": "WILMINGTON VA CLINIC (OH)",
                    "stationNumber": "538GF",
                    "npi": "1861809824",
                    "visn": "10",
                    "ien": 53808
                },
                {
                    "name": "CHILLICOTHE, OH PRRTP",
                    "stationNumber": "538PA",
                    "visn": "10",
                    "ien": 53809
                },
                {
                    "name": "CHILLICOTHE VA MOBILE CLINIC",
                    "stationNumber": "538QA",
                    "npi": "1063800332",
                    "visn": "10",
                    "ien": 53810
                }
            ],
            "539": [
                {
                    "name": "CINCINNATI",
                    "stationNumber": "539",
                    "officialName": "CINCINNATI VAMC",
                    "npi": "1093764821",
                    "domain": "CINCINNATI.MED.VA.GOV",
                    "visn": "10",
                    "ien": "539",
                    "mn": "CIN"
                },
                {
                    "name": "CINCINNATI COM LVNG CTR",
                    "stationNumber": "5399AA",
                    "npi": "1477911261",
                    "visn": "10",
                    "ien": 53901
                },
                {
                    "name": "FORT THOMAS COMMUNITY LVNG CTR",
                    "stationNumber": "5399AB",
                    "npi": "1437519733",
                    "ien": 53902,
                    "visn": "10"
                },
                {
                    "name": "CINCINNATI VAMC FORT THOMAS",
                    "stationNumber": "539A4",
                    "npi": "1932516283",
                    "visn": "10",
                    "ien": 53903
                },
                {
                    "name": "CINCINNATI VA FACILITY DOM",
                    "stationNumber": "539BU",
                    "visn": "10",
                    "ien": 53904
                },
                {
                    "name": "FORT THOMAS PRRTP",
                    "stationNumber": "539BV",
                    "npi": "1568822955",
                    "visn": "10",
                    "ien": 53905
                },
                {
                    "name": "OHIO VETERANS HOME GEORGETOWN",
                    "stationNumber": "539DT",
                    "visn": "10",
                    "ien": 53906
                },
                {
                    "name": "BELLEVUE",
                    "stationNumber": "539GA",
                    "officialName": "BELLEVUE CBOC",
                    "npi": "1144271016",
                    "visn": "10",
                    "ien": 53907
                },
                {
                    "name": "CLERMONT",
                    "stationNumber": "539GB",
                    "officialName": "CINCINNATI (CLERMONT COUNTY)",
                    "npi": "1639121775",
                    "visn": "10",
                    "ien": 53908
                },
                {
                    "name": "LAWRENCEBURG",
                    "stationNumber": "539GC",
                    "officialName": "LAWRENCEBURG (DEARBORN COUNTY)",
                    "npi": "1609829209",
                    "visn": "10",
                    "ien": 53909
                },
                {
                    "name": "FLORENCE CBOC",
                    "stationNumber": "539GD",
                    "npi": "1952458960",
                    "visn": "10",
                    "ien": 53910
                },
                {
                    "name": "HAMILTON OH CBOC",
                    "stationNumber": "539GE",
                    "officialName": "HAMILTON OHIO CBOC",
                    "npi": "1164601258",
                    "visn": "10",
                    "ien": 53911
                },
                {
                    "name": "GEORGETOWN CBOC",
                    "stationNumber": "539GF",
                    "visn": "10",
                    "ien": 53912
                },
                {
                    "name": "CIN-PRRTP",
                    "stationNumber": "539PA",
                    "visn": "10",
                    "ien": 53913
                },
                {
                    "name": "CINCINNATI VA MOBILE CLINIC",
                    "stationNumber": "539QA",
                    "npi": "1821486911",
                    "visn": "10",
                    "ien": 53914
                },
                {
                    "name": "HIGHLAND AVENUE VA CLINIC",
                    "stationNumber": "539QB",
                    "npi": "1780072876",
                    "visn": "10",
                    "ien": 53915
                },
                {
                    "name": "VINE STREET VA CLINIC",
                    "stationNumber": "539QC",
                    "npi": "1629559539",
                    "visn": "10",
                    "ien": 53916
                },
                {
                    "name": "NORWOOD VA CLINIC",
                    "stationNumber": "539QD",
                    "npi": "1548734759",
                    "visn": "10",
                    "ien": 53917
                }
            ],
            "541": [
                {
                    "name": "CLEVELAND VAMC",
                    "stationNumber": "541",
                    "npi": "1093764219",
                    "domain": "CLEVELAND.MED.VA.GOV",
                    "visn": "10",
                    "ien": "541",
                    "mn": "CLE"
                },
                {
                    "name": "CLEVELAND COMM LIVING CENTER",
                    "stationNumber": "5419AA",
                    "visn": "10",
                    "ien": 54101
                },
                {
                    "name": "SANDUSKY",
                    "stationNumber": "5419AF",
                    "officialName": "OHIO VETERANS HOME",
                    "visn": "10",
                    "ien": 54102
                },
                {
                    "name": "CLEVELAND DOMICILIARY",
                    "stationNumber": "541BU",
                    "visn": "10",
                    "ien": 54103
                },
                {
                    "name": "CANTON CBOC",
                    "stationNumber": "541BY",
                    "npi": "1487606018",
                    "visn": "10",
                    "ien": 54104
                },
                {
                    "name": "YOUNGSTOWN CBOC",
                    "stationNumber": "541BZ",
                    "npi": "1831139203",
                    "visn": "10",
                    "ien": 54105
                },
                {
                    "name": "CLEVELAND PHARM",
                    "stationNumber": "541CA",
                    "officialName": "CLEVELAND PHARMACY",
                    "visn": "10",
                    "ien": 54106
                },
                {
                    "name": "SANDUSKY STATE DOMICILIARY",
                    "stationNumber": "541DT",
                    "officialName": "OHIO VETERANS HOME",
                    "visn": "10",
                    "ien": 54107
                },
                {
                    "name": "LORAIN CBOC",
                    "stationNumber": "541GB",
                    "npi": "1013964790",
                    "visn": "10",
                    "ien": 54108
                },
                {
                    "name": "SANDUSKY CBOC",
                    "stationNumber": "541GC",
                    "npi": "1124066105",
                    "visn": "10",
                    "ien": 54109
                },
                {
                    "name": "DAVID F WINDER CLINIC",
                    "stationNumber": "541GD",
                    "officialName": "MANSFIELD CBOC",
                    "npi": "1972550507",
                    "visn": "10",
                    "ien": 54110
                },
                {
                    "name": "MCCAFERTY CBOC",
                    "stationNumber": "541GE",
                    "npi": "1992743934",
                    "visn": "10",
                    "ien": 54111
                },
                {
                    "name": "LAKE COUNTY VA CLINIC",
                    "stationNumber": "541GF",
                    "officialName": "PAINSVILLE CBOC",
                    "npi": "1902844558",
                    "visn": "10",
                    "ien": 54112
                },
                {
                    "name": "AKRON CBOC",
                    "stationNumber": "541GG",
                    "npi": "1134179799",
                    "visn": "10",
                    "ien": 54113
                },
                {
                    "name": "EAST LIVERPOOL CBOC",
                    "stationNumber": "541GH",
                    "npi": "1598718090",
                    "visn": "10",
                    "ien": 54114
                },
                {
                    "name": "WARREN CBOC",
                    "stationNumber": "541GI",
                    "npi": "1104865971",
                    "visn": "10",
                    "ien": 54115
                },
                {
                    "name": "NEW PHILADELPHIA CBOC",
                    "stationNumber": "541GJ",
                    "npi": "1205983228",
                    "visn": "10",
                    "ien": 54116
                },
                {
                    "name": "RAVENNA CBOC",
                    "stationNumber": "541GK",
                    "npi": "1124175153",
                    "visn": "10",
                    "ien": 54117
                },
                {
                    "name": "PARMA CBOC",
                    "stationNumber": "541GL",
                    "npi": "1487926853",
                    "visn": "10",
                    "ien": 54118
                },
                {
                    "name": "CLEVELAND AMBULATORY CTR CBOC",
                    "stationNumber": "541GM",
                    "npi": "1508267394",
                    "visn": "10",
                    "ien": 54119
                },
                {
                    "name": "CLEVELAND PRRTP",
                    "stationNumber": "541PA",
                    "visn": "10",
                    "ien": 54120
                },
                {
                    "name": "AKRON VA CLINIC",
                    "stationNumber": "541QA",
                    "npi": "1598153686",
                    "visn": "10",
                    "ien": 54121
                },
                {
                    "name": "CLEVELAND VA CLINIC",
                    "stationNumber": "541QB",
                    "npi": "1962867275",
                    "visn": "10",
                    "ien": 54122
                },
                {
                    "name": "CLEVELAND I VA MOBILE CLINIC",
                    "stationNumber": "541QC",
                    "npi": "1700277720",
                    "visn": "10",
                    "ien": 54123
                },
                {
                    "name": "CLEVELAND II VA MOBILE CLINIC",
                    "stationNumber": "541QD",
                    "npi": "1457742470",
                    "visn": "10",
                    "ien": 54124
                },
                {
                    "name": "CLEVELAND EAST BLVD VA MOBILE",
                    "stationNumber": "541QE",
                    "npi": "1144611179",
                    "visn": "10",
                    "ien": 54125
                },
                {
                    "name": "CUYAHOGA COUNTY VA MOBILE",
                    "stationNumber": "541QF",
                    "npi": "1053702084",
                    "visn": "10",
                    "ien": 54126
                },
                {
                    "name": "CLEVELAND SPINAL CORD OUTCOMES",
                    "stationNumber": "541SCI",
                    "officialName": "CLEVELAND SPINAL CORD OUTCOMES (SCI)",
                    "visn": "10",
                    "ien": 54127
                }
            ],
            "552": [
                {
                    "name": "DAYTON",
                    "stationNumber": "552",
                    "officialName": "DAYTON, OH VAMC",
                    "npi": "1518916733",
                    "domain": "DAYTON.MED.VA.GOV",
                    "visn": "10",
                    "ien": "552",
                    "mn": "DAY"
                },
                {
                    "name": "DAYTON (NHCU)",
                    "stationNumber": "5529AA",
                    "visn": "10",
                    "ien": 55201
                },
                {
                    "name": "DAYTON VA FACILITY DOM",
                    "stationNumber": "552BU",
                    "visn": "10",
                    "ien": 55202
                },
                {
                    "name": "MIDDLETOWN",
                    "stationNumber": "552GA",
                    "officialName": "MIDDLETOWN CBOC",
                    "npi": "1023056504",
                    "visn": "10",
                    "ien": 55203
                },
                {
                    "name": "LIMA",
                    "stationNumber": "552GB",
                    "officialName": "LIMA CBOC",
                    "npi": "1437106259",
                    "visn": "10",
                    "ien": 55204
                },
                {
                    "name": "RICHMOND, OH CBOC",
                    "stationNumber": "552GC",
                    "officialName": "RICHMOND CBOC",
                    "npi": "1003854183",
                    "visn": "10",
                    "ien": 55205
                },
                {
                    "name": "SPRINGFIELD CBOC-OH",
                    "stationNumber": "552GD",
                    "officialName": "SPRINGFIELD CBOC",
                    "npi": "1902845316",
                    "visn": "10",
                    "ien": 55206
                },
                {
                    "name": "IDES WRIGHT PATTERSON AFB",
                    "stationNumber": "552MA",
                    "visn": "10",
                    "ien": 55207
                },
                {
                    "name": "DAYTON VA MOBILE CLINIC",
                    "stationNumber": "552QA",
                    "npi": "1649668732",
                    "visn": "10",
                    "ien": 55208
                },
                {
                    "name": "DAYTON THIRD STREET VA MOBILE",
                    "stationNumber": "552QB",
                    "npi": "1114318144",
                    "visn": "10",
                    "ien": 55209
                },
                {
                    "name": "MONTGOMERY COUNTY VA MOBILE",
                    "stationNumber": "552QC",
                    "npi": "1437540515",
                    "visn": "10",
                    "ien": 55210
                }
            ],
            "553": [
                {
                    "name": "DETROIT, MI VAMC",
                    "stationNumber": "553",
                    "officialName": "JOHN D. DINGELL VAMC",
                    "npi": "1366491136",
                    "domain": "DETROIT.MED.VA.GOV",
                    "visn": "10",
                    "ien": "553",
                    "mn": "DET"
                },
                {
                    "name": "ALLEN PARK VA NURSING HOME",
                    "stationNumber": "5539AA",
                    "visn": "10",
                    "ien": 55301
                },
                {
                    "name": "GRAND RAPIDS, MI NHC",
                    "stationNumber": "5539AF",
                    "visn": "10",
                    "ien": 55302
                },
                {
                    "name": "DETROIT VA CARE SITE",
                    "stationNumber": "553BU",
                    "officialName": "DETROIT HOMELESS DRRTP",
                    "npi": "1154781615",
                    "visn": "10",
                    "ien": 55303
                },
                {
                    "name": "DETROIT VA DOMICILIARY",
                    "stationNumber": "553BV",
                    "npi": "1598125965",
                    "visn": "10",
                    "ien": 55304
                },
                {
                    "name": "OSCODA AFB",
                    "stationNumber": "553CZ",
                    "visn": "10",
                    "ien": 55305
                },
                {
                    "name": "YALE VA CLINIC",
                    "stationNumber": "553GA",
                    "npi": "1922048750",
                    "visn": "10",
                    "ien": 55306
                },
                {
                    "name": "PONTIAC VA CLINIC",
                    "stationNumber": "553GB",
                    "npi": "1609815034",
                    "visn": "10",
                    "ien": 55307
                },
                {
                    "name": "PIQUETTE STREET VA CLINIC",
                    "stationNumber": "553QA",
                    "npi": "1497143374",
                    "visn": "10",
                    "ien": 55308
                }
            ],
            "583": [
                {
                    "name": "RICHARD L. ROUDEBUSH VAMC",
                    "stationNumber": "583",
                    "npi": "1811946361",
                    "domain": "INDIANAPOLIS.MED.VA.GOV",
                    "visn": "10",
                    "ien": "583",
                    "mn": "IND"
                },
                {
                    "name": "INDIANAPOLIS NHC",
                    "stationNumber": "5839AA",
                    "visn": "10",
                    "ien": 58301
                },
                {
                    "name": "INDIANAPOLIS STATE HOME",
                    "stationNumber": "5839AF",
                    "visn": "10",
                    "ien": 58302
                },
                {
                    "name": "INDIANAPOLIS VA FACILITY DOM",
                    "stationNumber": "583BU",
                    "officialName": "INDIANAPOLIS VA FACILITY DOMICILIARY",
                    "npi": "1831697598",
                    "visn": "10",
                    "ien": 58303
                },
                {
                    "name": "LAFAYETTE STATE HOME",
                    "stationNumber": "583DT",
                    "visn": "10",
                    "ien": 58304
                },
                {
                    "name": "TERRE HAUTE VA CLINIC",
                    "stationNumber": "583GA",
                    "npi": "1326087271",
                    "visn": "10",
                    "ien": 58305
                },
                {
                    "name": "BLOOMINGTON VA CLINIC",
                    "stationNumber": "583GB",
                    "npi": "1831140417",
                    "visn": "10",
                    "ien": 58306
                },
                {
                    "name": "MARTINSVILLE VA CLINIC",
                    "stationNumber": "583GC",
                    "npi": "1578849832",
                    "visn": "10",
                    "ien": 58307
                },
                {
                    "name": "INDIANAPOLIS WEST VA CLINIC",
                    "stationNumber": "583GD",
                    "npi": "1124435185",
                    "visn": "10",
                    "ien": 58308
                },
                {
                    "name": "WEST LAFAYETTE VA CLINIC",
                    "stationNumber": "583GE",
                    "npi": "1639549470",
                    "visn": "10",
                    "ien": 58309
                },
                {
                    "name": "WAKEMAN VA CLINIC",
                    "stationNumber": "583GF",
                    "npi": "1487008934",
                    "visn": "10",
                    "ien": 58310
                },
                {
                    "name": "MONROE COUNTY VA CLINIC",
                    "stationNumber": "583QA",
                    "npi": "1790173078",
                    "visn": "10",
                    "ien": 58311
                },
                {
                    "name": "INDIANAPOLIS VA CLINIC",
                    "stationNumber": "583QB",
                    "npi": "1154719045",
                    "visn": "10",
                    "ien": 58312
                },
                {
                    "name": "VIGO COUNTY VA CLINIC",
                    "stationNumber": "583QC",
                    "npi": "1982092870",
                    "visn": "10",
                    "ien": 58313
                },
                {
                    "name": "INDIANAPOLIS YMCA VA CLINIC",
                    "stationNumber": "583QD",
                    "npi": "1164992897",
                    "visn": "10",
                    "ien": 58314
                },
                {
                    "name": "CRANE VA CLINIC",
                    "stationNumber": "583QF",
                    "npi": "1063042869",
                    "visn": "10",
                    "ien": 58315
                }
            ],
            "610": [
                {
                    "name": "MARION VA MEDICAL CENTER",
                    "stationNumber": "610",
                    "npi": "1225087786",
                    "domain": "NORTHERN-INDIANA.MED.VA.GOV",
                    "visn": "10",
                    "ien": "610",
                    "mn": "NOR"
                },
                {
                    "name": "MARION",
                    "stationNumber": "6109AA",
                    "visn": "10",
                    "ien": 61001
                },
                {
                    "name": "FORT WAYNE VA MEDICAL CENTER",
                    "stationNumber": "610A4",
                    "npi": "1568411791",
                    "visn": "10",
                    "ien": 61002
                },
                {
                    "name": "MARION VA FACILITY DOM",
                    "stationNumber": "610BU",
                    "officialName": "MARION VA FACILITY DOMICILIARY",
                    "visn": "10",
                    "ien": 61003
                },
                {
                    "name": "ST. JOSEPH COUNTY VA CLINIC",
                    "stationNumber": "610BY",
                    "npi": "1427597707",
                    "visn": "10",
                    "ien": 61004
                },
                {
                    "name": "MUNCIE VA CLINIC",
                    "stationNumber": "610GB",
                    "npi": "1649217027",
                    "visn": "10",
                    "ien": 61005
                },
                {
                    "name": "GOSHEN VA CLINIC",
                    "stationNumber": "610GC",
                    "npi": "1528217361",
                    "visn": "10",
                    "ien": 61006
                },
                {
                    "name": "PERU VA CLINIC",
                    "stationNumber": "610GD",
                    "npi": "1558684829",
                    "visn": "10",
                    "ien": 61007
                },
                {
                    "name": "MARION PRRTP",
                    "stationNumber": "610PB",
                    "visn": "10",
                    "ien": 61008
                },
                {
                    "name": "FORT WAYNE VA CLINIC",
                    "stationNumber": "610QA",
                    "npi": "1346629367",
                    "visn": "10",
                    "ien": 61009
                },
                {
                    "name": "COLUMBIA PLACE VA CLINIC",
                    "stationNumber": "610QB",
                    "npi": "1508219858",
                    "visn": "10",
                    "ien": 61010
                }
            ],
            "655": [
                {
                    "name": "ALEDA E. LUTZ VAMC",
                    "stationNumber": "655",
                    "npi": "1164471066",
                    "domain": "SAGINAW.MED.VA.GOV",
                    "visn": "10",
                    "ien": "655",
                    "mn": "SAG"
                },
                {
                    "name": "SAGINAW (NHCU)",
                    "stationNumber": "6559AA",
                    "visn": "10",
                    "ien": 65501
                },
                {
                    "name": "SAGINAW (DENT)",
                    "stationNumber": "655BY",
                    "visn": "10",
                    "ien": 65502
                },
                {
                    "name": "GAYLORD OUTPATIENT CLINIC",
                    "stationNumber": "655GA",
                    "officialName": "Gaylord VA Outpatient Clinic",
                    "npi": "1003869843",
                    "visn": "10",
                    "ien": 65503
                },
                {
                    "name": "TRAVERSE CITY VA CLINIC",
                    "stationNumber": "655GB",
                    "npi": "1811931918",
                    "visn": "10",
                    "ien": 65504
                },
                {
                    "name": "OSCODA VA CLINIC",
                    "stationNumber": "655GC",
                    "npi": "1265470843",
                    "visn": "10",
                    "ien": 65505
                },
                {
                    "name": "LT. COL CLEMENT VAN WAGONER VA",
                    "stationNumber": "655GD",
                    "npi": "1952550931",
                    "visn": "10",
                    "ien": 65506
                },
                {
                    "name": "CLARE VA CLINIC",
                    "stationNumber": "655GE",
                    "npi": "1649411265",
                    "visn": "10",
                    "ien": 65507
                },
                {
                    "name": "BAD AXE VA CLINIC",
                    "stationNumber": "655GF",
                    "npi": "1588979801",
                    "visn": "10",
                    "ien": 65508
                },
                {
                    "name": "CADILLAC VA CLINIC",
                    "stationNumber": "655GG",
                    "npi": "1194010173",
                    "visn": "10",
                    "ien": 65509
                },
                {
                    "name": "CHEBOYGAN COUNTY VA CLINIC",
                    "stationNumber": "655GH",
                    "npi": "1326329962",
                    "visn": "10",
                    "ien": 65510
                },
                {
                    "name": "GRAYLING VA CLINIC",
                    "stationNumber": "655GI",
                    "npi": "1780931733",
                    "visn": "10",
                    "ien": 65511
                },
                {
                    "name": "SAGINAW VA CLINIC",
                    "stationNumber": "655QA",
                    "npi": "1831587716",
                    "visn": "10",
                    "ien": 65512
                }
            ],
            "757": [
                {
                    "name": "COLUMBUS VAMC",
                    "stationNumber": "757",
                    "officialName": "CHALMERS P. WYLIE VA AMBULATORY CARE CENTER (757)",
                    "npi": "1881648558",
                    "domain": "COLUMBUS.MED.VA.GOV",
                    "visn": "10",
                    "ien": "757",
                    "mn": "COS"
                },
                {
                    "name": "ZANESVILLE",
                    "stationNumber": "757GA",
                    "officialName": "ZANESVILLE CBOC",
                    "npi": "1164462529",
                    "visn": "10",
                    "ien": 75701
                },
                {
                    "name": "GROVE CITY VA CLINIC (757GB)",
                    "stationNumber": "757GB",
                    "officialName": "GROVE CITY CBOC",
                    "npi": "1306891304",
                    "visn": "10",
                    "ien": 75702
                },
                {
                    "name": "MARION CBOC",
                    "stationNumber": "757GC",
                    "npi": "1033302518",
                    "visn": "10",
                    "ien": 75703
                },
                {
                    "name": "NEWARK CBOC",
                    "stationNumber": "757GD",
                    "npi": "1720139371",
                    "visn": "10",
                    "ien": 75704
                },
                {
                    "name": "COLUMBUS VA MOBILE CLINIC",
                    "stationNumber": "757QA",
                    "npi": "1902294895",
                    "visn": "10",
                    "ien": 75705
                },
                {
                    "name": "NORTH JAMES ROAD VA MOBILE",
                    "stationNumber": "757QB",
                    "npi": "1952792905",
                    "visn": "10",
                    "ien": 75706
                },
                {
                    "name": "COLUMBUS VA CLINIC",
                    "stationNumber": "757QC",
                    "npi": "1962851626",
                    "visn": "10",
                    "ien": 75707
                }
            ]
        },
        "7": {
            "508": [
                {
                    "name": "ATLANTA VAMC",
                    "stationNumber": "508",
                    "npi": "1902855216",
                    "domain": "ATLANTA.MED.VA.GOV",
                    "visn": "7",
                    "ien": "508",
                    "mn": "ATL"
                },
                {
                    "name": "ATLANTA VA COMM LIVING CTR",
                    "stationNumber": "5089AA",
                    "officialName": "ATLANTA NURSING HOME",
                    "npi": "1558729343",
                    "visn": "7",
                    "ien": 50801
                },
                {
                    "name": "CARROLLTON NURSING HOME",
                    "stationNumber": "5089AB",
                    "officialName": "CARROLLTON NURSING HOME CARE",
                    "visn": "7",
                    "ien": 50802
                },
                {
                    "name": "ATLANTA DOM",
                    "stationNumber": "508BU",
                    "npi": "1659739449",
                    "visn": "7",
                    "ien": 50803
                },
                {
                    "name": "ATLANTA VA CBOC",
                    "stationNumber": "508GA",
                    "npi": "1922164300",
                    "visn": "7",
                    "ien": 50804
                },
                {
                    "name": "OAKWOOD VA CLINIC",
                    "stationNumber": "508GE",
                    "npi": "1407890007",
                    "visn": "7",
                    "ien": 50805
                },
                {
                    "name": "AUSTELL CBOC",
                    "stationNumber": "508GF",
                    "npi": "1336183979",
                    "visn": "7",
                    "ien": 50806
                },
                {
                    "name": "STOCKBRIDGE CBOC",
                    "stationNumber": "508GG",
                    "npi": "1073769873",
                    "visn": "7",
                    "ien": 50807
                },
                {
                    "name": "LAWRENCEVILLE CBOC",
                    "stationNumber": "508GH",
                    "npi": "1780631689",
                    "visn": "7",
                    "ien": 50808
                },
                {
                    "name": "NEWNAN CBOC",
                    "stationNumber": "508GI",
                    "npi": "1780814368",
                    "visn": "7",
                    "ien": 50809
                },
                {
                    "name": "BLAIRSVILLE CBOC",
                    "stationNumber": "508GJ",
                    "visn": "7",
                    "ien": 50810
                },
                {
                    "name": "CARROLLTON CBOC",
                    "stationNumber": "508GK",
                    "npi": "1669729570",
                    "visn": "7",
                    "ien": 50811
                },
                {
                    "name": "ROME VA CLINIC",
                    "stationNumber": "508GL",
                    "npi": "1275940355",
                    "visn": "7",
                    "ien": 50812
                },
                {
                    "name": "PICKENS COUNTY VA CLINIC",
                    "stationNumber": "508GM",
                    "npi": "1962050286",
                    "visn": "7",
                    "ien": 50813
                },
                {
                    "name": "COVINGTON VA CLINIC (GA)",
                    "stationNumber": "508GN",
                    "npi": "1295383511",
                    "visn": "7",
                    "ien": 50814
                },
                {
                    "name": "NE COBB COUNTY VA CLINIC",
                    "stationNumber": "508GO",
                    "npi": "1811551633",
                    "visn": "7",
                    "ien": 50815
                },
                {
                    "name": "SOUTH COBB COUNTY VA CLINIC",
                    "stationNumber": "508GP",
                    "npi": "1831747153",
                    "visn": "7",
                    "ien": 50816
                },
                {
                    "name": "ATL-PRRTP",
                    "stationNumber": "508PA",
                    "visn": "7",
                    "ien": 50817
                },
                {
                    "name": "EAST POINT VA CLINIC",
                    "stationNumber": "508QB",
                    "npi": "1205224425",
                    "visn": "7",
                    "ien": 50818
                },
                {
                    "name": "HENDERSON MILLS VA CLINIC",
                    "stationNumber": "508QC",
                    "npi": "1669860888",
                    "visn": "7",
                    "ien": 50819
                },
                {
                    "name": "OLD EAST POINT VA CLINIC",
                    "stationNumber": "508QD",
                    "npi": "1174911200",
                    "visn": "7",
                    "ien": 50820
                },
                {
                    "name": "OLD LAWRENCEVILLE VA CLINIC",
                    "stationNumber": "508QE",
                    "npi": "1528456654",
                    "visn": "7",
                    "ien": 50821
                },
                {
                    "name": "ATLANTA VA CLINIC",
                    "stationNumber": "508QF",
                    "npi": "1013395102",
                    "visn": "7",
                    "ien": 50822
                },
                {
                    "name": "CARROLL COUNTY VA MOBIL CLINIC",
                    "stationNumber": "508QG",
                    "visn": "7",
                    "ien": 50823
                },
                {
                    "name": "COLLEGE PARK VA CLINIC",
                    "stationNumber": "508QH",
                    "npi": "1811545130",
                    "visn": "7",
                    "ien": 50824
                },
                {
                    "name": "NORTH DEKALB COUNTY VA CLINIC",
                    "stationNumber": "508QI",
                    "npi": "1821628629",
                    "visn": "7",
                    "ien": 50825
                },
                {
                    "name": "NORTH FULTON COUNTY VA CLINIC",
                    "stationNumber": "508QJ",
                    "npi": "1790333649",
                    "visn": "7",
                    "ien": 50826
                }
            ],
            "509": [
                {
                    "name": "AUGUSTA VAMC",
                    "stationNumber": "509",
                    "npi": "1801844469",
                    "domain": "AUGUSTA.MED.VA.GOV",
                    "visn": "7",
                    "ien": "509",
                    "mn": "AUG"
                },
                {
                    "name": "AUGUSTA NURSING HOME",
                    "stationNumber": "5099AA",
                    "visn": "7",
                    "ien": 50901
                },
                {
                    "name": "AUGUSTA (STATE HOME)",
                    "stationNumber": "5099AF",
                    "visn": "7",
                    "ien": 50902
                },
                {
                    "name": "AUGUSTA UPTOWN",
                    "stationNumber": "509A0",
                    "officialName": "AUGUSTA VAMC - UPTOWN",
                    "visn": "7",
                    "ien": 50903
                },
                {
                    "name": "AUGUSTA VA FACILITY DOM",
                    "stationNumber": "509BU",
                    "visn": "7",
                    "ien": 50904
                },
                {
                    "name": "AUGUSTA (ARMY HOSP)",
                    "stationNumber": "509CS",
                    "visn": "7",
                    "ien": 50905
                },
                {
                    "name": "AUGUSTA (PUBLIC HOSP)",
                    "stationNumber": "509DM",
                    "visn": "7",
                    "ien": 50906
                },
                {
                    "name": "AUGUSTA (STATE HOME)",
                    "stationNumber": "509DT",
                    "visn": "7",
                    "ien": 50907
                },
                {
                    "name": "ATHENS (GA) CBOC",
                    "stationNumber": "509GA",
                    "officialName": "ATHENS CBOC",
                    "npi": "1407909799",
                    "visn": "7",
                    "ien": 50908
                },
                {
                    "name": "AIKEN CBOC",
                    "stationNumber": "509GB",
                    "npi": "1306014113",
                    "visn": "7",
                    "ien": 50909
                },
                {
                    "name": "IDES FORT GORDON",
                    "stationNumber": "509MA",
                    "visn": "7",
                    "ien": 50910
                },
                {
                    "name": "STATESBORO VA CLINIC",
                    "stationNumber": "509QA",
                    "npi": "1437547569",
                    "visn": "7",
                    "ien": 50911
                },
                {
                    "name": "AUGUSTA SPINAL CORD OUTCOMES",
                    "stationNumber": "509SCI",
                    "officialName": "AUGUSTA SPINAL CORD OUTCOMES (SCI)",
                    "visn": "7",
                    "ien": 50912
                }
            ],
            "521": [
                {
                    "name": "BIRMINGHAM VAMC",
                    "stationNumber": "521",
                    "npi": "1043269046",
                    "domain": "BIRMINGHAM.MED.VA.GOV",
                    "visn": "7",
                    "ien": "521",
                    "mn": "BIR"
                },
                {
                    "name": "BIRMINGHAM NURSING HOME",
                    "stationNumber": "5219AA",
                    "visn": "7",
                    "ien": 52101
                },
                {
                    "name": "BIRMINGHAM DENTAL",
                    "stationNumber": "521BY",
                    "visn": "7",
                    "ien": 52102
                },
                {
                    "name": "COLONEL ROBERT L HOWARD SVH",
                    "stationNumber": "521DT",
                    "officialName": "HUNTSVILLE (STNB)",
                    "visn": "7",
                    "ien": 52103
                },
                {
                    "name": "ALEXANDER CITY",
                    "stationNumber": "521DU",
                    "visn": "7",
                    "ien": 52104
                },
                {
                    "name": "BAY MINETTE (521DV)",
                    "stationNumber": "521DV",
                    "officialName": "BAY MINETTE",
                    "visn": "7",
                    "ien": 52105
                },
                {
                    "name": "HUNTSVILLE (CBOC)",
                    "stationNumber": "521GA",
                    "officialName": "HUNTSVILLE CBOC",
                    "npi": "1043250632",
                    "visn": "7",
                    "ien": 52106
                },
                {
                    "name": "DECATUR",
                    "stationNumber": "521GB",
                    "officialName": "DECATUR CBOC",
                    "npi": "1073566741",
                    "visn": "7",
                    "ien": 52107
                },
                {
                    "name": "FLORENCE",
                    "stationNumber": "521GC",
                    "officialName": "FLORENCE CBOC",
                    "npi": "1689618373",
                    "visn": "7",
                    "ien": 52108
                },
                {
                    "name": "RAINBOW CITY CBOC",
                    "stationNumber": "521GD",
                    "npi": "1063450195",
                    "visn": "7",
                    "ien": 52109
                },
                {
                    "name": "ANNISTON",
                    "stationNumber": "521GE",
                    "officialName": "ANNISTON CBOC",
                    "npi": "1306897400",
                    "visn": "7",
                    "ien": 52110
                },
                {
                    "name": "JASPER VA CLINIC",
                    "stationNumber": "521GF",
                    "npi": "1891741294",
                    "visn": "7",
                    "ien": 52111
                },
                {
                    "name": "BESSEMER CBOC",
                    "stationNumber": "521GG",
                    "npi": "1053599480",
                    "visn": "7",
                    "ien": 52112
                },
                {
                    "name": "CHILDERSBURG CBOC",
                    "stationNumber": "521GH",
                    "npi": "1255574729",
                    "visn": "7",
                    "ien": 52113
                },
                {
                    "name": "GUNTERSVILLE CBOC",
                    "stationNumber": "521GI",
                    "npi": "1992024947",
                    "visn": "7",
                    "ien": 52114
                },
                {
                    "name": "BIRMINGHAM VA CLINIC",
                    "stationNumber": "521GJ",
                    "npi": "1518319052",
                    "visn": "7",
                    "ien": 52115
                },
                {
                    "name": "IDES REDSTONE ARMY ARSENAL",
                    "stationNumber": "521MA",
                    "visn": "7",
                    "ien": 52116
                },
                {
                    "name": "BIRMINGHAM PRRTP",
                    "stationNumber": "521PA",
                    "visn": "7",
                    "ien": 52117
                },
                {
                    "name": "CALLAHAN VA CLINIC",
                    "stationNumber": "521QA",
                    "npi": "1346638475",
                    "visn": "7",
                    "ien": 52118
                }
            ],
            "534": [
                {
                    "name": "CHARLESTON VAMC",
                    "stationNumber": "534",
                    "officialName": "RALPH H. JOHNSON VA MEDICAL CENTER (534)",
                    "npi": "1659320620",
                    "domain": "CHARLESTON.MED.VA.GOV",
                    "visn": "7",
                    "ien": "534",
                    "mn": "CHA"
                },
                {
                    "name": "CHARLESTON NURSING HOME CARE",
                    "stationNumber": "5349AA",
                    "visn": "7",
                    "ien": 53401
                },
                {
                    "name": "SAVANNAH OPC-CHARLESTON",
                    "stationNumber": "534BY",
                    "officialName": "SAVANNAH PRIMARY CARE CLINIC",
                    "npi": "1851339451",
                    "visn": "7",
                    "ien": 53402
                },
                {
                    "name": "VETERANS VICTORY HOUSE SVH",
                    "stationNumber": "534DT",
                    "officialName": "VETERANS VICTORY HOUSE STATE VETERANS HOME",
                    "visn": "7",
                    "ien": 53403
                },
                {
                    "name": "MYRTLE BEACH",
                    "stationNumber": "534GB",
                    "officialName": "MYRTLE  BEACH PRIMARY CARE CLINIC",
                    "npi": "1821036286",
                    "visn": "7",
                    "ien": 53404
                },
                {
                    "name": "BEAUFORT",
                    "stationNumber": "534GC",
                    "officialName": "BEAUFORT PRIMARY CARE CLINIC",
                    "npi": "1194776807",
                    "visn": "7",
                    "ien": 53405
                },
                {
                    "name": "GOOSE CREEK VA CLINIC",
                    "stationNumber": "534GD",
                    "npi": "1235286741",
                    "visn": "7",
                    "ien": 53406
                },
                {
                    "name": "HINESVILLE CBOC",
                    "stationNumber": "534GE",
                    "npi": "1427344308",
                    "visn": "7",
                    "ien": 53407
                },
                {
                    "name": "NORTH CHARLESTON VA CBOC",
                    "stationNumber": "534GF",
                    "npi": "1437566528",
                    "visn": "7",
                    "ien": 53408
                },
                {
                    "name": "IDES NAVAL HSPTAL BEAUFORT",
                    "stationNumber": "534MA",
                    "visn": "7",
                    "ien": 53409
                },
                {
                    "name": "IDES NAVL HLTH CLNC CHARLESTON",
                    "stationNumber": "534MB",
                    "visn": "7",
                    "ien": 53410
                },
                {
                    "name": "IDES CHARLESTON AFB",
                    "stationNumber": "534MC",
                    "visn": "7",
                    "ien": 53411
                },
                {
                    "name": "MARKET COMMONS VA CLINIC",
                    "stationNumber": "534QA",
                    "npi": "1164810297",
                    "visn": "7",
                    "ien": 53412
                },
                {
                    "name": "TRIDENT 2 VA CLINIC",
                    "stationNumber": "534QB",
                    "npi": "1255729398",
                    "visn": "7",
                    "ien": 53413
                },
                {
                    "name": "CHARLESTON VA CLINIC",
                    "stationNumber": "534QC",
                    "npi": "1083099444",
                    "visn": "7",
                    "ien": 53414
                }
            ],
            "544": [
                {
                    "name": "COLUMBIA, SC VAMC",
                    "stationNumber": "544",
                    "officialName": "WM JENNINGS BRYAN DORN VETERANS AFFAIRS MEDICAL CENTER",
                    "npi": "1578514030",
                    "domain": "COLUMBIA-SC.MED.VA.GOV",
                    "visn": "7",
                    "ien": "544",
                    "mn": "COL"
                },
                {
                    "name": "COLUMBIA, SC NURSING HOME",
                    "stationNumber": "5449AA",
                    "visn": "7",
                    "ien": 54401
                },
                {
                    "name": "COLUMBIA, SC NURSING HOME",
                    "stationNumber": "5449AF",
                    "visn": "7",
                    "ien": 54402
                },
                {
                    "name": "ANDERSON (STATE HOME) ",
                    "stationNumber": "5449AG",
                    "visn": "7",
                    "ien": 54403
                },
                {
                    "name": "GREENVILLE VA CLINIC SC",
                    "stationNumber": "544BZ",
                    "npi": "1104865211",
                    "visn": "7",
                    "ien": 54404
                },
                {
                    "name": "COLUMBIA, SC ARMY HOSP",
                    "stationNumber": "544CN",
                    "visn": "7",
                    "ien": 54405
                },
                {
                    "name": "COLUMBIA, SC NAVY HOSP",
                    "stationNumber": "544CT",
                    "visn": "7",
                    "ien": 54406
                },
                {
                    "name": "COLUMBIA, SC ST NURSING HOME",
                    "stationNumber": "544DT",
                    "ien": 54407,
                    "visn": "7"
                },
                {
                    "name": "FLORENCE, SC CBOC",
                    "stationNumber": "544GB",
                    "officialName": "FLORENCE CBOC",
                    "npi": "1588617815",
                    "visn": "7",
                    "ien": 54408
                },
                {
                    "name": "ROCK HILL, SC CBOC",
                    "stationNumber": "544GC",
                    "officialName": "ROCK HILL CBOC",
                    "npi": "1649218579",
                    "visn": "7",
                    "ien": 54409
                },
                {
                    "name": "ANDERSON, SC CBOC",
                    "stationNumber": "544GD",
                    "officialName": "ANDERSON CBOC",
                    "npi": "1215988480",
                    "visn": "7",
                    "ien": 54410
                },
                {
                    "name": "ORANGEBURG, SC CBOC",
                    "stationNumber": "544GE",
                    "officialName": "ORANGEBURG CBOC",
                    "npi": "1912945593",
                    "visn": "7",
                    "ien": 54411
                },
                {
                    "name": "SUMTER CBOC",
                    "stationNumber": "544GF",
                    "npi": "1972542983",
                    "visn": "7",
                    "ien": 54412
                },
                {
                    "name": "SPARTANBURG CBOC",
                    "stationNumber": "544GG",
                    "npi": "1861643181",
                    "visn": "7",
                    "ien": 54413
                },
                {
                    "name": "COLUMBIA VA MOBILE CLINIC",
                    "stationNumber": "544HK",
                    "npi": "1053786228",
                    "visn": "7",
                    "ien": 54414
                },
                {
                    "name": "IDES SHAW AFB",
                    "stationNumber": "544MA",
                    "visn": "7",
                    "ien": 54415
                },
                {
                    "name": "IDES FT. JACKSON",
                    "stationNumber": "544MB",
                    "visn": "7",
                    "ien": 54416
                }
            ],
            "557": [
                {
                    "name": "DUBLIN VAMC",
                    "stationNumber": "557",
                    "officialName": "DUBLIN",
                    "npi": "1033168257",
                    "domain": "DUBLIN.MED.VA.GOV",
                    "visn": "7",
                    "ien": "557",
                    "mn": "DUB"
                },
                {
                    "name": "DUBLIN VANB",
                    "stationNumber": "5579AA",
                    "visn": "7",
                    "ien": 55701
                },
                {
                    "name": "MILLEDGEVILLE STATE HOME",
                    "stationNumber": "5579AF",
                    "visn": "7",
                    "ien": 55702
                },
                {
                    "name": "DUBLIN DOM",
                    "stationNumber": "557BU",
                    "officialName": "DUBLIN DOMICILIARY",
                    "visn": "7",
                    "ien": 55703
                },
                {
                    "name": "MILLEDGEVILLE",
                    "stationNumber": "557DT",
                    "officialName": "GWVH DOMICILIARY MILLEDGEVILLE",
                    "visn": "7",
                    "ien": 55704
                },
                {
                    "name": "MACON CBOC",
                    "stationNumber": "557GA",
                    "officialName": "MACON VA CBOC",
                    "npi": "1023065638",
                    "visn": "7",
                    "ien": 55705
                },
                {
                    "name": "ALBANY, GA (CBOC)",
                    "stationNumber": "557GB",
                    "officialName": "ALBANY VA CBOC",
                    "npi": "1316997026",
                    "visn": "7",
                    "ien": 55706
                },
                {
                    "name": "MILLEDGEVILE CBOC",
                    "stationNumber": "557GC",
                    "npi": "1114261211",
                    "visn": "7",
                    "ien": 55707
                },
                {
                    "name": "BRUNSWICK CBOC",
                    "stationNumber": "557GE",
                    "visn": "7",
                    "ien": 55708
                },
                {
                    "name": "TIFTON VA CLINIC",
                    "stationNumber": "557GF",
                    "npi": "1770836884",
                    "visn": "7",
                    "ien": 55709
                },
                {
                    "name": "PERRY OUTREACH CLINIC",
                    "stationNumber": "557HA",
                    "npi": "1578948105",
                    "visn": "7",
                    "ien": 55710
                },
                {
                    "name": "IDES ROBINS AFB",
                    "stationNumber": "557MA",
                    "visn": "7",
                    "ien": 55711
                }
            ],
            "619": [
                {
                    "name": "CENTRAL ALABAMA HCS",
                    "stationNumber": "619",
                    "officialName": "CENTRAL ALABAMA HEALTH CARE SYSTEM - MONTGOMERY DIVISION",
                    "npi": "1174573547",
                    "domain": "CENTRAL-ALABAMA.MED.VA.GOV",
                    "visn": "7",
                    "ien": "619",
                    "mn": "CEN"
                },
                {
                    "name": "TUSKEGEE-NHC",
                    "stationNumber": "6199AB",
                    "visn": "7",
                    "ien": 61901
                },
                {
                    "name": "TUSKEGEE VAMC",
                    "stationNumber": "619A4",
                    "officialName": " CENTRAL ALABAMA HEALTH CARE SYSTEM - TUSKEGEE DIVISION",
                    "npi": "1659321131",
                    "visn": "7",
                    "ien": 61902
                },
                {
                    "name": "TUSKEGEE-DOM",
                    "stationNumber": "619BU",
                    "officialName": "CENTRAL ALABAMA HCS - DOMICILIARY",
                    "visn": "7",
                    "ien": 61903
                },
                {
                    "name": "COLUMBUS CBOC",
                    "stationNumber": "619GA",
                    "officialName": "CENTRAL ALABAMA HCS - COLUMBUS CBOC",
                    "npi": "1255383311",
                    "visn": "7",
                    "ien": 61904
                },
                {
                    "name": "DOTHAN",
                    "stationNumber": "619GB",
                    "officialName": "CENTRAL ALABAMA HCS - DOTHAN CBOC",
                    "npi": "1679526131",
                    "visn": "7",
                    "ien": 61905
                },
                {
                    "name": "WIREGRASS CBOC",
                    "stationNumber": "619GD",
                    "npi": "1790015402",
                    "visn": "7",
                    "ien": 61906
                },
                {
                    "name": "MONROE COUNTY CBOC",
                    "stationNumber": "619GE",
                    "npi": "1063755668",
                    "visn": "7",
                    "ien": 61907
                },
                {
                    "name": "MONTGOMERY VA CLINIC",
                    "stationNumber": "619GF",
                    "npi": "1942665278",
                    "visn": "7",
                    "ien": 61908
                },
                {
                    "name": "TUSKEGEE PRRTP",
                    "stationNumber": "619PA",
                    "officialName": "CENTRAL ALABAMA HCS - PRRTP",
                    "visn": "7",
                    "ien": 61909
                },
                {
                    "name": "HOUSTON COUNTY VA CLINIC",
                    "stationNumber": "619QA",
                    "officialName": "DOTHAN 2 VA CLINIC",
                    "npi": "1518355650",
                    "visn": "7",
                    "ien": 61910
                },
                {
                    "name": "FORT BENNING VA CLINIC",
                    "stationNumber": "619QB",
                    "npi": "1982074720",
                    "visn": "7",
                    "ien": 61911
                }
            ],
            "679": [
                {
                    "name": "TUSCALOOSA",
                    "stationNumber": "679",
                    "officialName": "TUSCALOOSA VA MEDICAL CENTER",
                    "npi": "1801846126",
                    "domain": "TUSCALOOSA.MED.VA.GOV",
                    "visn": "7",
                    "ien": "679",
                    "mn": "TUS"
                },
                {
                    "name": "TUSCALOOSA NHC",
                    "stationNumber": "6799AA",
                    "officialName": "TUSCALOOSA VA MEDICAL CENTER",
                    "visn": "7",
                    "ien": 67901
                },
                {
                    "name": "TUSCALOOSA HOMELESS DOM",
                    "stationNumber": "679BU",
                    "visn": "7",
                    "ien": 67902
                },
                {
                    "name": "SELMA VA CLINIC",
                    "stationNumber": "679GA",
                    "npi": "1871900829",
                    "visn": "7",
                    "ien": 67903
                },
                {
                    "name": "MOBILE CLINIC TUSCALOOSA",
                    "stationNumber": "679HK",
                    "npi": "1285019810",
                    "visn": "7",
                    "ien": 67904
                },
                {
                    "name": "IDES COLUMBUS",
                    "stationNumber": "679MA",
                    "visn": "7",
                    "ien": 67905
                },
                {
                    "name": "TUSCALOOSA PRRTP",
                    "stationNumber": "679PA",
                    "npi": "1023478336",
                    "visn": "7",
                    "ien": 67906
                }
            ]
        },
        "5": {
            "512": [
                {
                    "name": "VA MARYLAND HEALTH CARE SYS",
                    "stationNumber": "512",
                    "npi": "1083663157",
                    "domain": "BALTIMORE.MED.VA.GOV",
                    "visn": "5",
                    "ien": "512",
                    "mn": "BAL"
                },
                {
                    "name": "BALTIMORE NURSING HOME",
                    "stationNumber": "5129AA",
                    "visn": "5",
                    "ien": 51201
                },
                {
                    "name": "PERRY POINT",
                    "stationNumber": "5129AC",
                    "officialName": "PERRY POINT NURSING HOME",
                    "visn": "5",
                    "ien": 51202
                },
                {
                    "name": "PERRY POINT",
                    "stationNumber": "512A5",
                    "officialName": "PERRY POINT VAMC",
                    "npi": "1750331831",
                    "visn": "5",
                    "ien": 51203
                },
                {
                    "name": "PERRY POINT DOM",
                    "stationNumber": "512BU",
                    "visn": "5",
                    "ien": 51204
                },
                {
                    "name": "BALTIMORE VA FACILITY DOM",
                    "stationNumber": "512BV",
                    "visn": "5",
                    "ien": 51205
                },
                {
                    "name": "FORT MEADE",
                    "stationNumber": "512CN",
                    "visn": "5",
                    "ien": 51206
                },
                {
                    "name": "PERRY POINT",
                    "stationNumber": "512DM",
                    "officialName": "PERRY POINT PUBH",
                    "visn": "5",
                    "ien": 51207
                },
                {
                    "name": "PERRY POINT",
                    "stationNumber": "512DS",
                    "officialName": "PERRY POINT CIVH",
                    "visn": "5",
                    "ien": 51208
                },
                {
                    "name": "CAMBRIDGE VA CLINIC",
                    "stationNumber": "512GA",
                    "npi": "1073564688",
                    "visn": "5",
                    "ien": 51209
                },
                {
                    "name": "GLEN BURNIE VA CLINIC",
                    "stationNumber": "512GC",
                    "npi": "1972557304",
                    "visn": "5",
                    "ien": 51210
                },
                {
                    "name": "LOCH RAVEN VAMC",
                    "stationNumber": "512GD",
                    "npi": "1447206560",
                    "visn": "5",
                    "ien": 51211
                },
                {
                    "name": "POCOMOKE CITY VA CLINIC",
                    "stationNumber": "512GE",
                    "npi": "1326087719",
                    "visn": "5",
                    "ien": 51212
                },
                {
                    "name": "E. BALTIMORE COUNTY VA CLINIC",
                    "stationNumber": "512GF",
                    "npi": "1245380872",
                    "visn": "5",
                    "ien": 51213
                },
                {
                    "name": "FORT MEADE VA CLINIC",
                    "stationNumber": "512GG",
                    "npi": "1811236060",
                    "visn": "5",
                    "ien": 51214
                },
                {
                    "name": "IDES FORT MEADE",
                    "stationNumber": "512MA",
                    "visn": "5",
                    "ien": 51215
                },
                {
                    "name": "IDES NAVAL HC ANNAPOLIS",
                    "stationNumber": "512MB",
                    "visn": "5",
                    "ien": 51216
                },
                {
                    "name": "BALTIMORE PRRTP",
                    "stationNumber": "512PA",
                    "visn": "5",
                    "ien": 51217
                },
                {
                    "name": "PERRY POINT PRRTP",
                    "stationNumber": "512PC",
                    "visn": "5",
                    "ien": 51218
                },
                {
                    "name": "BALTIMORE VA CLINIC ",
                    "stationNumber": "512QA",
                    "npi": "1376054429",
                    "visn": "5",
                    "ien": 51219
                }
            ],
            "517": [
                {
                    "name": "BECKLEY VA MEDICAL CENTER",
                    "stationNumber": "517",
                    "npi": "1750330858",
                    "domain": "BECKLEY.MED.VA.GOV",
                    "visn": "5",
                    "ien": "517",
                    "mn": "BEC"
                },
                {
                    "name": "BECKLEY NURSING HOME",
                    "stationNumber": "5179AA",
                    "visn": "5",
                    "ien": 51701
                },
                {
                    "name": "GREENBRIER COUNTY VA CLINIC",
                    "stationNumber": "517GB",
                    "npi": "1861794067",
                    "visn": "5",
                    "ien": 51702
                },
                {
                    "name": "BECKLEY VA MOBILE CLINIC",
                    "stationNumber": "517HK",
                    "visn": "5",
                    "ien": 51703
                },
                {
                    "name": "BECKLEY PRRTP",
                    "stationNumber": "517PA",
                    "visn": "5",
                    "ien": 51704
                },
                {
                    "name": "PRINCETON VA CLINIC",
                    "stationNumber": "517QA",
                    "npi": "1477947331",
                    "visn": "5",
                    "ien": 51705
                }
            ],
            "540": [
                {
                    "name": "LOUIS A JOHNSON VAMC",
                    "stationNumber": "540",
                    "npi": "1295784056",
                    "domain": "CLARKSBURG.MED.VA.GOV",
                    "visn": "5",
                    "ien": "540",
                    "mn": "CLA"
                },
                {
                    "name": "CLARKSBURG NHCU",
                    "stationNumber": "5409AA",
                    "officialName": "LOUIS A JOHNSON VA MEDICAL CENTER",
                    "visn": "5",
                    "ien": 54001
                },
                {
                    "name": "CLARKSBURG VA FACILITY DOM",
                    "stationNumber": "540BU",
                    "officialName": "CLARKSBURG VA FACILITY DOMICILIARY",
                    "visn": "5",
                    "ien": 54002
                },
                {
                    "name": "WEST VIRGINIA VA NURSING HOME",
                    "stationNumber": "540DT",
                    "visn": "5",
                    "ien": 54003
                },
                {
                    "name": "TUCKER COUNTY VA CLINIC",
                    "stationNumber": "540GA",
                    "npi": "1669412037",
                    "visn": "5",
                    "ien": 54004
                },
                {
                    "name": "CLARKSBURG/WOOD",
                    "stationNumber": "540GB",
                    "npi": "1508804964",
                    "visn": "5",
                    "ien": 54005
                },
                {
                    "name": "BRAXTON COUNTY VA CLINIC",
                    "stationNumber": "540GC",
                    "npi": "1790739118",
                    "visn": "5",
                    "ien": 54006
                },
                {
                    "name": "MONONGALIA COUNTY VA CLINIC",
                    "stationNumber": "540GD",
                    "npi": "1972753655",
                    "visn": "5",
                    "ien": 54007
                },
                {
                    "name": "CLARKSBURG VA MOBILE CLINIC",
                    "stationNumber": "540HK",
                    "npi": "1619342870",
                    "visn": "5",
                    "ien": 54008
                }
            ],
            "581": [
                {
                    "name": "HUNTINGTON VAMC",
                    "stationNumber": "581",
                    "npi": "1669421319",
                    "domain": "HUNTINGTON.MED.VA.GOV",
                    "visn": "5",
                    "ien": "581",
                    "mn": "HUN"
                },
                {
                    "name": "BARBOURSVILLE VETERANS HOME",
                    "stationNumber": "581DT",
                    "officialName": "BARBOURSVILLE STATE HOME",
                    "visn": "5",
                    "ien": 58101
                },
                {
                    "name": "EK OUTPATIENT CLINIC",
                    "stationNumber": "581GA",
                    "officialName": "PRESTONSBURG CBOC",
                    "npi": "1366480436",
                    "visn": "5",
                    "ien": 58102
                },
                {
                    "name": "CHARLESTON VA CLINIC",
                    "stationNumber": "581GB",
                    "npi": "1144271883",
                    "visn": "5",
                    "ien": 58103
                },
                {
                    "name": "GALLIPOLIS VA CLINIC",
                    "stationNumber": "581GG",
                    "npi": "1952718900",
                    "visn": "5",
                    "ien": 58104
                },
                {
                    "name": "LENORE VA CLINIC",
                    "stationNumber": "581GH",
                    "npi": "1861809816",
                    "visn": "5",
                    "ien": 58105
                },
                {
                    "name": "HUNTINGTON 9TH ST. VA CLINIC",
                    "stationNumber": "581QA",
                    "officialName": "HUNTINGTON 9TH STREET VA CLINIC",
                    "npi": "1063800191",
                    "visn": "5",
                    "ien": 58106
                },
                {
                    "name": "HUNTINGTON VA MOBILE CLINIC",
                    "stationNumber": "581QB",
                    "npi": "1619475027",
                    "visn": "5",
                    "ien": 58107
                }
            ],
            "613": [
                {
                    "name": "MARTINSBURG VA MEDICAL CENTER",
                    "stationNumber": "613",
                    "npi": "1366492738",
                    "domain": "MARTINSBURG.MED.VA.GOV",
                    "visn": "5",
                    "ien": "613",
                    "mn": "MAR"
                },
                {
                    "name": "MARTINSBURG NHC",
                    "stationNumber": "6139AA",
                    "visn": "5",
                    "ien": 61301
                },
                {
                    "name": "MARTINSBURG DOM",
                    "stationNumber": "613BU",
                    "visn": "5",
                    "ien": 61302
                },
                {
                    "name": "CUMBERLAND",
                    "stationNumber": "613GA",
                    "officialName": "CUMBERLAND CBOC",
                    "npi": "1932152642",
                    "visn": "5",
                    "ien": 61303
                },
                {
                    "name": "HAGERSTOWN VA CLINIC",
                    "stationNumber": "613GB",
                    "npi": "1649226689",
                    "visn": "5",
                    "ien": 61304
                },
                {
                    "name": "STEPHENS CITY VA CLINIC",
                    "stationNumber": "613GC",
                    "npi": "1003856527",
                    "visn": "5",
                    "ien": 61305
                },
                {
                    "name": "FRANKLIN VA CLINIC",
                    "stationNumber": "613GD",
                    "npi": "1194771071",
                    "visn": "5",
                    "ien": 61306
                },
                {
                    "name": "PETERSBURG VA CLINIC",
                    "stationNumber": "613GE",
                    "npi": "1912941501",
                    "visn": "5",
                    "ien": 61307
                },
                {
                    "name": "HARRISONBURG VA CLINIC",
                    "stationNumber": "613GF",
                    "npi": "1790730190",
                    "visn": "5",
                    "ien": 61308
                },
                {
                    "name": "FORT DETRICK VA CLINIC",
                    "stationNumber": "613GG",
                    "npi": "1215224399",
                    "visn": "5",
                    "ien": 61309
                },
                {
                    "name": "MOBILE VAN CLINIC",
                    "stationNumber": "613HK",
                    "officialName": "MOBILE VAN CLINIC MORC",
                    "visn": "5",
                    "ien": 61310
                },
                {
                    "name": "MARTINSBURG CWT/TR",
                    "stationNumber": "613PB",
                    "visn": "5",
                    "ien": 61311
                }
            ],
            "688": [
                {
                    "name": "WASHINGTON VA MEDICAL CENTER",
                    "stationNumber": "688",
                    "npi": "1336198704",
                    "domain": "WASHINGTON.MED.VA.GOV",
                    "visn": "5",
                    "ien": "688",
                    "mn": "WAS"
                },
                {
                    "name": "WASHINGTON NHC",
                    "stationNumber": "6889AA",
                    "visn": "5",
                    "ien": 68801
                },
                {
                    "name": "WASHINGTON STNB",
                    "stationNumber": "6889AF",
                    "visn": "5",
                    "ien": 68802
                },
                {
                    "name": "WASHINGTON DC VA FACILITY DOM",
                    "stationNumber": "688BU",
                    "officialName": "WASHINGTON DC VA FACILITY DOMICILIARY",
                    "visn": "5",
                    "ien": 68803
                },
                {
                    "name": "WALTER REED",
                    "stationNumber": "688CN",
                    "officialName": "WALTER REED ARMY HOSPITAL",
                    "visn": "5",
                    "ien": 68804
                },
                {
                    "name": "BETHESDA NAVEL",
                    "stationNumber": "688CT",
                    "visn": "5",
                    "ien": 68805
                },
                {
                    "name": "ST. ELIZABETHS",
                    "stationNumber": "688DB",
                    "officialName": "ST ELIZABETHS",
                    "visn": "5",
                    "ien": 68806
                },
                {
                    "name": "D.C. STATE HOME",
                    "stationNumber": "688DU",
                    "officialName": "STATE HOME",
                    "visn": "5",
                    "ien": 68807
                },
                {
                    "name": "CHARLOTTE HALL STATE VET HOME",
                    "stationNumber": "688DV",
                    "visn": "5",
                    "ien": 68808
                },
                {
                    "name": "FORT BELVOIR VA CLINIC",
                    "stationNumber": "688GA",
                    "npi": "1770533358",
                    "visn": "5",
                    "ien": 68809
                },
                {
                    "name": "SOUTHEAST WASHINGTON VA CLINIC",
                    "stationNumber": "688GB",
                    "npi": "1386682805",
                    "visn": "5",
                    "ien": 68810
                },
                {
                    "name": "CHARLOTTE HALL VA CLINIC",
                    "stationNumber": "688GD",
                    "npi": "1649218165",
                    "visn": "5",
                    "ien": 68811
                },
                {
                    "name": "S PRINCE GEORGES CITY VA CLINC",
                    "stationNumber": "688GE",
                    "officialName": "S PRINCE GEORGES CITY VA CLINIC",
                    "npi": "1932483443",
                    "visn": "5",
                    "ien": 68812
                },
                {
                    "name": "MONTGOMERY COUNTY VA CLINIC",
                    "stationNumber": "688GF",
                    "npi": "1003398728",
                    "visn": "5",
                    "ien": 68813
                },
                {
                    "name": "LEXINGTON PARK VA CLINIC",
                    "stationNumber": "688GG",
                    "npi": "1285272385",
                    "visn": "5",
                    "ien": 68814
                },
                {
                    "name": "IDES ANDREWS AFB",
                    "stationNumber": "688MA",
                    "visn": "5",
                    "ien": 68815
                },
                {
                    "name": "IDES WALTER REED ARMY MEDCTR",
                    "stationNumber": "688MB",
                    "visn": "5",
                    "ien": 68816
                },
                {
                    "name": "IDES NATL NAVAL MEDCTR BETHSDA",
                    "stationNumber": "688MC",
                    "visn": "5",
                    "ien": 68817
                },
                {
                    "name": "IDES FORT BELVOIR",
                    "stationNumber": "688MD",
                    "visn": "5",
                    "ien": 68818
                },
                {
                    "name": "IDES QUANTICO MARINE BASE",
                    "stationNumber": "688ME",
                    "visn": "5",
                    "ien": 68819
                },
                {
                    "name": "IDES BOLLING AFB",
                    "stationNumber": "688MF",
                    "visn": "5",
                    "ien": 68820
                },
                {
                    "name": "IDES NAVAL AIR STN PATXNT RIVR",
                    "stationNumber": "688MG",
                    "visn": "5",
                    "ien": 68821
                },
                {
                    "name": "WASHINGTON CWT/TR",
                    "stationNumber": "688PA",
                    "visn": "5",
                    "ien": 68822
                },
                {
                    "name": "FRANKLIN STREET VA CLINIC",
                    "stationNumber": "688QA",
                    "npi": "1790183036",
                    "visn": "5",
                    "ien": 68823
                }
            ]
        },
        "8": {
            "516": [
                {
                    "name": "C.W. BILL YOUNG DEPT OF VAMC",
                    "stationNumber": "516",
                    "npi": "1881643609",
                    "domain": "BAY-PINES.MED.VA.GOV",
                    "visn": "8",
                    "ien": "516",
                    "mn": "BAY"
                },
                {
                    "name": "BAY PINES NURSING HOME",
                    "stationNumber": "5169AA",
                    "visn": "8",
                    "ien": 51601
                },
                {
                    "name": "BAY PINES DOMICILIARY",
                    "stationNumber": "516BU",
                    "visn": "8",
                    "ien": 51602
                },
                {
                    "name": "LEE COUNTY VA CLINIC",
                    "stationNumber": "516BZ",
                    "npi": "1710933452",
                    "visn": "8",
                    "ien": 51603
                },
                {
                    "name": "DOUGLAS JACOBSON SVH",
                    "stationNumber": "516DT",
                    "visn": "8",
                    "ien": 51604
                },
                {
                    "name": "SARASOTA COUNTY CBOC",
                    "stationNumber": "516GA",
                    "npi": "1134167737",
                    "visn": "8",
                    "ien": 51605
                },
                {
                    "name": "ST. PETERSBURG VA CLINIC",
                    "stationNumber": "516GB",
                    "npi": "1801835764",
                    "visn": "8",
                    "ien": 51606
                },
                {
                    "name": "PALM HARBOR VA CLINIC",
                    "stationNumber": "516GC",
                    "npi": "1245282334",
                    "visn": "8",
                    "ien": 51607
                },
                {
                    "name": "BRADENTON VA CLINIC",
                    "stationNumber": "516GD",
                    "npi": "1073560603",
                    "visn": "8",
                    "ien": 51608
                },
                {
                    "name": "PORT CHARLOTTE VA CLINIC",
                    "stationNumber": "516GE",
                    "npi": "1649219031",
                    "visn": "8",
                    "ien": 51609
                },
                {
                    "name": "NAPLES VA CLINIC",
                    "stationNumber": "516GF",
                    "npi": "1639117294",
                    "visn": "8",
                    "ien": 51610
                },
                {
                    "name": "SEBRING VA CLINIC",
                    "stationNumber": "516GH",
                    "npi": "1427009604",
                    "visn": "8",
                    "ien": 51611
                },
                {
                    "name": "BAY PINES-PRRTP",
                    "stationNumber": "516PA",
                    "visn": "8",
                    "ien": 51612
                }
            ],
            "546": [
                {
                    "name": "MIAMI VAMC",
                    "stationNumber": "546",
                    "officialName": "BRUCE W. CARTER DEPT OF VAMC",
                    "npi": "1467402263",
                    "domain": "MIAMI.MED.VA.GOV",
                    "visn": "8",
                    "ien": "546",
                    "mn": "MIA"
                },
                {
                    "name": "MIAMI (NHCU)",
                    "stationNumber": "5469AA",
                    "visn": "8",
                    "ien": 54601
                },
                {
                    "name": "MIAMI (OCS)",
                    "stationNumber": "546B0",
                    "visn": "8",
                    "ien": 54602
                },
                {
                    "name": "MIAMI VA FACILITY DOM",
                    "stationNumber": "546BU",
                    "officialName": "MIAMI VA FACILITY DOMICILIARY",
                    "visn": "8",
                    "ien": 54603
                },
                {
                    "name": "BROWARD COUNTY VA OPC",
                    "stationNumber": "546BZ",
                    "officialName": "WILLIAM BILL KLING VAOC",
                    "npi": "1538106992",
                    "visn": "8",
                    "ien": 54604
                },
                {
                    "name": "KEY WEST NAVY HOSP",
                    "stationNumber": "546CU",
                    "visn": "8",
                    "ien": 54605
                },
                {
                    "name": "MIAMI (ALCOHOL/DRUG)",
                    "stationNumber": "546GA",
                    "npi": "1043258940",
                    "visn": "8",
                    "ien": 54606
                },
                {
                    "name": "KEY WEST CBOC",
                    "stationNumber": "546GB",
                    "npi": "1588610406",
                    "visn": "8",
                    "ien": 54607
                },
                {
                    "name": "HOMESTEAD VA CLINIC",
                    "stationNumber": "546GC",
                    "npi": "1366497737",
                    "visn": "8",
                    "ien": 54608
                },
                {
                    "name": "PEMBROKE PINES VA CLINIC",
                    "stationNumber": "546GD",
                    "npi": "1134168529",
                    "visn": "8",
                    "ien": 54609
                },
                {
                    "name": "KEY LARGO VA CLINIC",
                    "stationNumber": "546GE",
                    "npi": "1043266968",
                    "visn": "8",
                    "ien": 54610
                },
                {
                    "name": "HOLLYWOOD VA CLINIC",
                    "stationNumber": "546GF",
                    "npi": "1003861071",
                    "visn": "8",
                    "ien": 54611
                },
                {
                    "name": "DEERFIELD BEACH CLINIC",
                    "stationNumber": "546GH",
                    "npi": "1073565875",
                    "visn": "8",
                    "ien": 54612
                },
                {
                    "name": "MIA-PRRTP",
                    "stationNumber": "546PA",
                    "visn": "8",
                    "ien": 54613
                },
                {
                    "name": "MIAMI SPINAL CORD OUTCOMES",
                    "stationNumber": "546SCI",
                    "officialName": "MIAMI SPINAL CORD OUTCOMES (SCI)",
                    "visn": "8",
                    "ien": 54614
                }
            ],
            "573": [
                {
                    "name": "N. FLORIDA/S. GEORGIA HCS",
                    "stationNumber": "573",
                    "officialName": "MALCOM RANDALL DEPT OF VAMC",
                    "npi": "1063462000",
                    "domain": "NORTH-FLORIDA.MED.VA.GOV",
                    "visn": "8",
                    "ien": "573",
                    "mn": "NOR"
                },
                {
                    "name": "GAINESVILLE NHC",
                    "stationNumber": "5739A",
                    "visn": "8",
                    "ien": 57301
                },
                {
                    "name": "GAINESVILLE (NHCU)",
                    "stationNumber": "5739AA",
                    "officialName": "GAINESVILLE NHCU",
                    "visn": "8",
                    "ien": 57302
                },
                {
                    "name": "LAKE CITY NHC",
                    "stationNumber": "5739B",
                    "visn": "8",
                    "ien": 57303
                },
                {
                    "name": "SVH ST AUGUSTINE FLORIDA",
                    "stationNumber": "5739F",
                    "visn": "8",
                    "ien": 57304
                },
                {
                    "name": "LAKE CITY VA MEDICAL CENTER",
                    "stationNumber": "573A4",
                    "npi": "1508815630",
                    "visn": "8",
                    "ien": 57305
                },
                {
                    "name": "GAINESVILLE VA DOMICILIARY",
                    "stationNumber": "573BU",
                    "npi": "1174983480",
                    "visn": "8",
                    "ien": 57306
                },
                {
                    "name": "LAKE CITY VA FACILITY DOM",
                    "stationNumber": "573BV",
                    "officialName": "LAKE CITY VA FACILITY DOMICILIARY",
                    "visn": "8",
                    "ien": 57307
                },
                {
                    "name": "JACKSONVILLE 1 VA CLINIC",
                    "stationNumber": "573BY",
                    "npi": "1578506655",
                    "visn": "8",
                    "ien": 57308
                },
                {
                    "name": "DAYTONA VETERANS NURSING HOME",
                    "stationNumber": "573DT",
                    "officialName": "Emory L. Bennett State Veterans Nursing Home",
                    "visn": "8",
                    "ien": 57309
                },
                {
                    "name": "LAKE CITY STATE HOME",
                    "stationNumber": "573DU",
                    "visn": "8",
                    "ien": 57310
                },
                {
                    "name": "VALDOSTA VA CLINIC",
                    "stationNumber": "573GA",
                    "npi": "1306886791",
                    "visn": "8",
                    "ien": 57311
                },
                {
                    "name": "JACKSONVILLE, FL CBOC",
                    "stationNumber": "573GB",
                    "officialName": "JACKSONVILLE CBOC",
                    "visn": "8",
                    "ien": 57312
                },
                {
                    "name": "OCALA VA CLINIC",
                    "stationNumber": "573GD",
                    "npi": "1295772747",
                    "visn": "8",
                    "ien": 57313
                },
                {
                    "name": "ST. AUGUSTINE VA CLINIC",
                    "stationNumber": "573GE",
                    "officialName": "SAINT AUGUSTINE VA CLINIC",
                    "npi": "1922041003",
                    "visn": "8",
                    "ien": 57314
                },
                {
                    "name": "SGT E.I. BOOTS THOMAS VA CLIN",
                    "stationNumber": "573GF",
                    "npi": "1902830508",
                    "visn": "8",
                    "ien": 57315
                },
                {
                    "name": "LECANTO VA CLINIC (NFL)",
                    "stationNumber": "573GG",
                    "npi": "1538211545",
                    "visn": "8",
                    "ien": 57316
                },
                {
                    "name": "THE VILLAGES VA CLINIC",
                    "stationNumber": "573GI",
                    "npi": "1346397379",
                    "visn": "8",
                    "ien": 57317
                },
                {
                    "name": "ST. MARYS VA CLINIC",
                    "stationNumber": "573GJ",
                    "npi": "1710137203",
                    "visn": "8",
                    "ien": 57318
                },
                {
                    "name": "MARIANNA VA CLINIC",
                    "stationNumber": "573GK",
                    "visn": "8",
                    "ien": 57319
                },
                {
                    "name": "PALATKA VA CLINIC",
                    "stationNumber": "573GL",
                    "visn": "8",
                    "ien": 57320
                },
                {
                    "name": "WAYCROSS VA CLINIC",
                    "stationNumber": "573GM",
                    "npi": "1407263452",
                    "visn": "8",
                    "ien": 57321
                },
                {
                    "name": "PERRY VA CLINIC",
                    "stationNumber": "573GN",
                    "npi": "1376910968",
                    "visn": "8",
                    "ien": 57322
                },
                {
                    "name": "IDES NAVAL HOSPTL JACKSONVILLE",
                    "stationNumber": "573MA",
                    "visn": "8",
                    "ien": 57323
                },
                {
                    "name": "IDES MOODY AIR FORCE BASE",
                    "stationNumber": "573MB",
                    "visn": "8",
                    "ien": 57324
                },
                {
                    "name": "GAINESVILLE (PRRTP)",
                    "stationNumber": "573PA",
                    "officialName": "GAINESVILLE PRRTP",
                    "visn": "8",
                    "ien": 57325
                },
                {
                    "name": "SIXTEENTH STREET VA CLINIC",
                    "stationNumber": "573QA",
                    "npi": "1194123430",
                    "visn": "8",
                    "ien": 57326
                },
                {
                    "name": "GAINESVILLE VA CLINIC",
                    "stationNumber": "573QB",
                    "officialName": "GAINESVILLE 98TH ST VA CLINIC",
                    "npi": "1003214347",
                    "visn": "8",
                    "ien": 57327
                },
                {
                    "name": "SIXTY FOURTH STREET VA CLINIC",
                    "stationNumber": "573QC",
                    "npi": "1326436551",
                    "visn": "8",
                    "ien": 57328
                },
                {
                    "name": "ALACHUA COUNTY VA CLINIC",
                    "stationNumber": "573QD",
                    "officialName": "GAINESVILLE 64TH ST 2 VA CLINIC",
                    "npi": "1780072918",
                    "visn": "8",
                    "ien": 57329
                },
                {
                    "name": "GAINESVILLE SOUTWEST VA CLINIC",
                    "stationNumber": "573QE",
                    "officialName": "GAINESVILLE 64TH ST 3 VA CLINIC",
                    "npi": "1811395130",
                    "visn": "8",
                    "ien": 57330
                },
                {
                    "name": "TWENTY THIRD AVE VA CLINIC",
                    "stationNumber": "573QF",
                    "npi": "1912305251",
                    "visn": "8",
                    "ien": 57331
                },
                {
                    "name": "JACKSONVILLE SO POINT VA CLINC",
                    "stationNumber": "573QG",
                    "npi": "1649678988",
                    "visn": "8",
                    "ien": 57332
                },
                {
                    "name": "OCALA WEST VA CLINIC",
                    "stationNumber": "573QH",
                    "npi": "1992103238",
                    "visn": "8",
                    "ien": 57333
                },
                {
                    "name": "MAHAN BOULEVARD VA CLINIC",
                    "stationNumber": "573QI",
                    "npi": "1861880098",
                    "visn": "8",
                    "ien": 57334
                },
                {
                    "name": "JACKSONVILLE 2 VA CLINIC-FL",
                    "stationNumber": "573QJ",
                    "npi": "1821464314",
                    "visn": "8",
                    "ien": 57335
                },
                {
                    "name": "LAKE CITY VA CLINIC",
                    "stationNumber": "573QK",
                    "npi": "1437627015",
                    "visn": "8",
                    "ien": 57336
                }
            ],
            "673": [
                {
                    "name": "TAMPA FL VAMC",
                    "stationNumber": "673",
                    "officialName": "JAMES A. HALEY VETERANS HOSP",
                    "npi": "1386694701",
                    "domain": "TAMPA.MED.VA.GOV",
                    "visn": "8",
                    "ien": "673",
                    "mn": "TAM"
                },
                {
                    "name": "TAMPA NHCU",
                    "stationNumber": "6739AA",
                    "officialName": "JAMES A. HALEY NHCU",
                    "visn": "8",
                    "ien": 67301
                },
                {
                    "name": "TAMPA VA DOMICILIARY",
                    "stationNumber": "673BV",
                    "npi": "1699135061",
                    "visn": "8",
                    "ien": 67302
                },
                {
                    "name": "PASCO OPC",
                    "stationNumber": "673BZ",
                    "officialName": "PORT RICHEY OUTPATIENT CLINIC",
                    "npi": "1669419149",
                    "visn": "8",
                    "ien": 67303
                },
                {
                    "name": "LAKELAND VA CLINIC",
                    "stationNumber": "673GB",
                    "npi": "1811944796",
                    "visn": "8",
                    "ien": 67304
                },
                {
                    "name": "BROOKSVILLE CBOC",
                    "stationNumber": "673GC",
                    "npi": "1437100179",
                    "visn": "8",
                    "ien": 67305
                },
                {
                    "name": "ZEPHYRHILLS VA CLINIC",
                    "stationNumber": "673GF",
                    "npi": "1720027600",
                    "visn": "8",
                    "ien": 67306
                },
                {
                    "name": "SOUTH HILLSBOROUGH VA CLINIC",
                    "stationNumber": "673GG",
                    "visn": "8",
                    "ien": 67307
                },
                {
                    "name": "LECANTO VA CLINIC (TAMPA)",
                    "stationNumber": "673GH",
                    "npi": "1669031126",
                    "visn": "8",
                    "ien": 67308
                },
                {
                    "name": "46TH STREET NORTH VA CLINIC",
                    "stationNumber": "673QA",
                    "npi": "1629476965",
                    "visn": "8",
                    "ien": 67309
                },
                {
                    "name": "HILLBOROUGH COUNTY VA CLINIC",
                    "stationNumber": "673QB",
                    "officialName": "46TH STREET SOUTH VA CLINIC",
                    "npi": "1720486970",
                    "visn": "8",
                    "ien": 67310
                },
                {
                    "name": "LAKELAND VA CLINIC",
                    "stationNumber": "673QC",
                    "officialName": "WEST LAKELAND VA CLINIC",
                    "npi": "1922406180",
                    "visn": "8",
                    "ien": 67311
                },
                {
                    "name": "DEER PARK VA CLINIC",
                    "stationNumber": "673QD",
                    "npi": "1548668791",
                    "visn": "8",
                    "ien": 67312
                },
                {
                    "name": "HIGHWAY NINETEEN VA CLINIC",
                    "stationNumber": "673QE",
                    "npi": "1255729505",
                    "visn": "8",
                    "ien": 67313
                },
                {
                    "name": "TAMPA VA CLINIC",
                    "stationNumber": "673QF",
                    "officialName": "WINNERS CIRCLE VA CLINIC",
                    "npi": "1467850628",
                    "visn": "8",
                    "ien": 67314
                },
                {
                    "name": "LITTLE ROAD VA CLINIC",
                    "stationNumber": "673QG",
                    "npi": "1902204167",
                    "visn": "8",
                    "ien": 67315
                },
                {
                    "name": "BRUCE B DOWNS BLVD VA CLINIC",
                    "stationNumber": "673QH",
                    "npi": "1811395072",
                    "visn": "8",
                    "ien": 67316
                },
                {
                    "name": "PASCO COUNTY VA CLINIC",
                    "stationNumber": "673QI",
                    "officialName": "MEDICAL VIEW LANE VA CLINIC",
                    "visn": "8",
                    "ien": 67317
                },
                {
                    "name": "HIDDEN RIVER VA CLINIC",
                    "stationNumber": "673QJ",
                    "npi": "1609264951",
                    "visn": "8",
                    "ien": 67318
                },
                {
                    "name": "TAMPA 1 VA MOBILE CLINIC",
                    "stationNumber": "673QK",
                    "npi": "1548668700",
                    "visn": "8",
                    "ien": 67319
                },
                {
                    "name": "TAMPA 2 VA MOBILE CLINIC",
                    "stationNumber": "673QL",
                    "npi": "1508264763",
                    "visn": "8",
                    "ien": 67320
                },
                {
                    "name": "TAMPA SPINAL CORD OUTCOMES",
                    "stationNumber": "673SCI",
                    "officialName": "TAMPA SPINAL CORD OUTCOMES (SCI)",
                    "visn": "8",
                    "ien": 67321
                },
                {
                    "name": "TAMPA SPINAL CORD INJURY LTC",
                    "stationNumber": "673SLT",
                    "visn": "8",
                    "ien": 67322
                }
            ]
        },
        "2": {
            "526": [
                {
                    "name": "BRONX VAMC",
                    "stationNumber": "526",
                    "officialName": "BRONX VA HOSPITAL",
                    "npi": "1316996630",
                    "domain": "BRONX.MED.VA.GOV",
                    "visn": "2",
                    "ien": "526",
                    "mn": "BRO"
                },
                {
                    "name": "BRONX (NHCU)",
                    "stationNumber": "5269AA",
                    "officialName": "BRONX VA NURSING HOME CARE UNIT",
                    "visn": "2",
                    "ien": 52601
                },
                {
                    "name": "WHITE PLAINS CBOC",
                    "stationNumber": "526GA",
                    "officialName": "WHITE PLAINS CLINIC",
                    "npi": "1245270065",
                    "visn": "2",
                    "ien": 52602
                },
                {
                    "name": "YONKERS CBOC",
                    "stationNumber": "526GB",
                    "officialName": "YONKERS CLINIC",
                    "npi": "1023057965",
                    "visn": "2",
                    "ien": 52603
                },
                {
                    "name": "QUEENS CBOC",
                    "stationNumber": "526GD",
                    "officialName": "NORTH QUEENS CLINIC",
                    "npi": "1558309047",
                    "visn": "2",
                    "ien": 52604
                },
                {
                    "name": "BRONX VA MOBILE OOS",
                    "stationNumber": "526QA",
                    "npi": "1700284353",
                    "visn": "2",
                    "ien": 52605
                },
                {
                    "name": "BRONX SPINAL CORD OUTCOMES",
                    "stationNumber": "526SCI",
                    "officialName": "BRONX SPINAL CORD OUTCOMES (SCI)",
                    "visn": "2",
                    "ien": 52606
                }
            ],
            "528": [
                {
                    "name": "UPSTATE NEW YORK HCS",
                    "stationNumber": "528",
                    "officialName": "BUFFALO VA MEDICAL CENTER",
                    "npi": "1154370112",
                    "domain": "V02.MED.VA.GOV",
                    "visn": "2",
                    "ien": "528",
                    "mn": "V02"
                },
                {
                    "name": "BATAVIA PRRP",
                    "stationNumber": "5284PA",
                    "visn": "2",
                    "ien": 52801
                },
                {
                    "name": "BUFFALO NHCU",
                    "stationNumber": "5289AA",
                    "officialName": "BUFFALO, NY NHCU",
                    "visn": "2",
                    "ien": 52802
                },
                {
                    "name": "BATAVIA NHCU",
                    "stationNumber": "5289AB",
                    "visn": "2",
                    "ien": 52803
                },
                {
                    "name": "CANANDAIGUA NHCU",
                    "stationNumber": "5289AC",
                    "visn": "2",
                    "ien": 52804
                },
                {
                    "name": "SYRACUSE NHCU",
                    "stationNumber": "5289AD",
                    "visn": "2",
                    "ien": 52805
                },
                {
                    "name": "BATH NHCU",
                    "stationNumber": "5289AK",
                    "visn": "2",
                    "ien": 52806
                },
                {
                    "name": "ALBANY, NY NHCU",
                    "stationNumber": "5289AL",
                    "visn": "2",
                    "ien": 52807
                },
                {
                    "name": "BATAVIA VAMC",
                    "stationNumber": "528A4",
                    "officialName": "VA HEALTHCARE NETWORK UPSTATE NEW YORK SYSTEM VISN 2 - BATAVIA DIVISION",
                    "npi": "1669416624",
                    "visn": "2",
                    "ien": 52808
                },
                {
                    "name": "CANANDAIGUA VA MEDICAL CENTER",
                    "stationNumber": "528A5",
                    "npi": "1427006758",
                    "visn": "2",
                    "ien": 52809
                },
                {
                    "name": "BATH VA MEDICAL CENTER",
                    "stationNumber": "528A6",
                    "npi": "1609825702",
                    "visn": "2",
                    "ien": 52810
                },
                {
                    "name": "SYRACUSE VA MEDICAL CENTER",
                    "stationNumber": "528A7",
                    "npi": "1619926086",
                    "visn": "2",
                    "ien": 52811
                },
                {
                    "name": "SAMUEL S. STRATTON VAMC",
                    "stationNumber": "528A8",
                    "npi": "1831148139",
                    "visn": "2",
                    "ien": 52812
                },
                {
                    "name": "BATAVIA VA FACILITY DOM",
                    "stationNumber": "528B1",
                    "officialName": "BATAVIA VA FACILITY DOMICILIARY",
                    "visn": "2",
                    "ien": 52813
                },
                {
                    "name": "CANANDAIGUA DOMICILIARY",
                    "stationNumber": "528BU",
                    "visn": "2",
                    "ien": 52814
                },
                {
                    "name": "BATH DOMICILIARY",
                    "stationNumber": "528BV",
                    "officialName": "BATH VA DOMICILIARY",
                    "visn": "2",
                    "ien": 52815
                },
                {
                    "name": "ALBANY VA FACILITY DOM",
                    "stationNumber": "528BW",
                    "officialName": "ALBANY VA FACILITY DOMICILIARY",
                    "visn": "2",
                    "ien": 52816
                },
                {
                    "name": "BUFFALO VA FACILITY DOM",
                    "stationNumber": "528BX",
                    "officialName": "BUFFALO VA FACILITY DOMICILIARY",
                    "visn": "2",
                    "ien": 52817
                },
                {
                    "name": "BATAVIA STATE NURSING HOME",
                    "stationNumber": "528DT",
                    "visn": "2",
                    "ien": 52818
                },
                {
                    "name": "OXFORD",
                    "stationNumber": "528DU",
                    "officialName": "OXFORD STATE HOME",
                    "visn": "2",
                    "ien": 52819
                },
                {
                    "name": "WESTPORT CBOC",
                    "stationNumber": "528G2",
                    "npi": "1306899836",
                    "visn": "2",
                    "ien": 52820
                },
                {
                    "name": "BAINBRIDGE CBOC",
                    "stationNumber": "528G3",
                    "officialName": "SIDNEY CBOC",
                    "npi": "1952458010",
                    "visn": "2",
                    "ien": 52821
                },
                {
                    "name": "ELMIRA CBOC",
                    "stationNumber": "528G4",
                    "npi": "1467405803",
                    "visn": "2",
                    "ien": 52822
                },
                {
                    "name": "AUBURN",
                    "stationNumber": "528G5",
                    "npi": "1992756274",
                    "visn": "2",
                    "ien": 52823
                },
                {
                    "name": "FONDA CBOC",
                    "stationNumber": "528G6",
                    "npi": "1013960350",
                    "visn": "2",
                    "ien": 52824
                },
                {
                    "name": "CATSKILL CBOC",
                    "stationNumber": "528G7",
                    "npi": "1811948433",
                    "visn": "2",
                    "ien": 52825
                },
                {
                    "name": "WELLSVILLE",
                    "stationNumber": "528G8",
                    "officialName": "WELLSVILLE CBOC",
                    "npi": "1487694253",
                    "visn": "2",
                    "ien": 52826
                },
                {
                    "name": "CORTLAND",
                    "stationNumber": "528G9",
                    "npi": "1164474177",
                    "visn": "2",
                    "ien": 52827
                },
                {
                    "name": "JAMESTOWN VA CLINIC",
                    "stationNumber": "528GB",
                    "npi": "1760437917",
                    "visn": "2",
                    "ien": 52828
                },
                {
                    "name": "DUNKIRK VA CLINIC",
                    "stationNumber": "528GC",
                    "npi": "1013960574",
                    "visn": "2",
                    "ien": 52829
                },
                {
                    "name": "NIAGARA FALLS VA CLINIC",
                    "stationNumber": "528GD",
                    "npi": "1356388854",
                    "visn": "2",
                    "ien": 52830
                },
                {
                    "name": "ROCHESTER WESTFALL VA CLINIC",
                    "stationNumber": "528GE",
                    "officialName": "ROCHESTER OPC",
                    "npi": "1841233491",
                    "visn": "2",
                    "ien": 52831
                },
                {
                    "name": "LOCKPORT",
                    "stationNumber": "528GK",
                    "npi": "1801842059",
                    "visn": "2",
                    "ien": 52832
                },
                {
                    "name": "MASSENA CBOC",
                    "stationNumber": "528GL",
                    "npi": "1457399495",
                    "visn": "2",
                    "ien": 52833
                },
                {
                    "name": "ROME CBOC",
                    "stationNumber": "528GM",
                    "npi": "1366480360",
                    "visn": "2",
                    "ien": 52834
                },
                {
                    "name": "BINGHAMTON CBOC",
                    "stationNumber": "528GN",
                    "npi": "1194776781",
                    "visn": "2",
                    "ien": 52835
                },
                {
                    "name": "WATERTOWN CBOC",
                    "stationNumber": "528GO",
                    "npi": "1942357033",
                    "visn": "2",
                    "ien": 52836
                },
                {
                    "name": "OSWEGO CBOC",
                    "stationNumber": "528GP",
                    "npi": "1104864743",
                    "visn": "2",
                    "ien": 52837
                },
                {
                    "name": "LACKAWANNA CBOC",
                    "stationNumber": "528GQ",
                    "npi": "1821045915",
                    "visn": "2",
                    "ien": 52838
                },
                {
                    "name": "SPRINGVILLE OUTPATIENT CLINIC",
                    "stationNumber": "528GQ01",
                    "visn": "2",
                    "ien": 52839
                },
                {
                    "name": "OLEAN",
                    "stationNumber": "528GR",
                    "officialName": "OLEAN CBOC",
                    "npi": "1821035270",
                    "visn": "2",
                    "ien": 52840
                },
                {
                    "name": "GLENS FALLS CBOC",
                    "stationNumber": "528GT",
                    "npi": "1417901844",
                    "visn": "2",
                    "ien": 52841
                },
                {
                    "name": "PLATTSBURGH CBOC",
                    "stationNumber": "528GV",
                    "officialName": "PLATTSBURGH",
                    "npi": "1912945387",
                    "visn": "2",
                    "ien": 52842
                },
                {
                    "name": "SCHENECTADY",
                    "stationNumber": "528GW",
                    "officialName": "SCHENECTADY CBOC",
                    "npi": "1982643391",
                    "visn": "2",
                    "ien": 52843
                },
                {
                    "name": "TROY",
                    "stationNumber": "528GX",
                    "officialName": "TROY CBOC",
                    "npi": "1184664567",
                    "visn": "2",
                    "ien": 52844
                },
                {
                    "name": "CLIFTON PARK VA CLINIC",
                    "stationNumber": "528GY",
                    "officialName": "CLIFTON PARK CBOC",
                    "npi": "1831141969",
                    "visn": "2",
                    "ien": 52845
                },
                {
                    "name": "KINGSTON CBOC",
                    "stationNumber": "528GZ",
                    "npi": "1952339400",
                    "visn": "2",
                    "ien": 52846
                },
                {
                    "name": "WATERTOWN DES",
                    "stationNumber": "528MA",
                    "visn": "2",
                    "ien": 52847
                },
                {
                    "name": "BUFFALO SARRTP",
                    "stationNumber": "528PA",
                    "visn": "2",
                    "ien": 52848
                },
                {
                    "name": "CANANDAIGUA PRRTP",
                    "stationNumber": "528PC",
                    "visn": "2",
                    "ien": 52849
                },
                {
                    "name": "ALBANY PRRTP",
                    "stationNumber": "528PD",
                    "visn": "2",
                    "ien": 52850
                },
                {
                    "name": "BUFFALO VA CLINIC",
                    "stationNumber": "528QA",
                    "npi": "1225436827",
                    "visn": "2",
                    "ien": 52851
                },
                {
                    "name": "PACKARD VA CLINIC",
                    "stationNumber": "528QB",
                    "npi": "1932507530",
                    "visn": "2",
                    "ien": 52852
                },
                {
                    "name": "ROCHESTER CALKINS VA CLINIC",
                    "stationNumber": "528QC",
                    "npi": "1841698446",
                    "visn": "2",
                    "ien": 52853
                },
                {
                    "name": "MOUNT HOPE VA CLINIC",
                    "stationNumber": "528QD",
                    "npi": "1295133890",
                    "visn": "2",
                    "ien": 52854
                },
                {
                    "name": "COUDERSPORT VA CLINIC",
                    "stationNumber": "528QE",
                    "npi": "1659779254",
                    "visn": "2",
                    "ien": 52855
                },
                {
                    "name": "WELLSBORO VA CLINIC",
                    "stationNumber": "528QF",
                    "npi": "1568860161",
                    "visn": "2",
                    "ien": 52856
                },
                {
                    "name": "ERIE WEST VA CLINIC",
                    "stationNumber": "528QG",
                    "npi": "1477951077",
                    "visn": "2",
                    "ien": 52857
                },
                {
                    "name": "SOUTH SALINA VA CLINIC",
                    "stationNumber": "528QH",
                    "npi": "1366840969",
                    "visn": "2",
                    "ien": 52858
                },
                {
                    "name": "ERIE EAST VA CLINIC",
                    "stationNumber": "528QI",
                    "npi": "1033517602",
                    "visn": "2",
                    "ien": 52859
                },
                {
                    "name": "ALBANY VA CLINIC",
                    "stationNumber": "528QJ",
                    "npi": "1275931875",
                    "visn": "2",
                    "ien": 52860
                },
                {
                    "name": "SARANAC LAKE VA CLINIC",
                    "stationNumber": "528QK",
                    "npi": "1053719658",
                    "visn": "2",
                    "ien": 52861
                },
                {
                    "name": "BATH VAMC MOBILE MEDICAL UNIT",
                    "stationNumber": "528QL",
                    "npi": "1295133809",
                    "visn": "2",
                    "ien": 52862
                },
                {
                    "name": "CANANDAIGUA MED MOBILE UNIT",
                    "stationNumber": "528QM",
                    "npi": "1831594035",
                    "visn": "2",
                    "ien": 52863
                },
                {
                    "name": "WATERTOWN 2 VA CLINIC",
                    "stationNumber": "528QN",
                    "npi": "1730564980",
                    "visn": "2",
                    "ien": 52864
                },
                {
                    "name": "NATIONAL SUICIDE PREV HOTLINE",
                    "stationNumber": "528SH",
                    "officialName": "NATIONAL SUICIDE PREVENTION HOTLINE",
                    "visn": "2",
                    "ien": 52865
                }
            ],
            "561": [
                {
                    "name": "EAST ORANGE-VA NEW JERSEY HCS",
                    "stationNumber": "561",
                    "officialName": "NEW JERSEY HEALTH CARE SYSTEM - EAST ORANGE",
                    "npi": "1801845664",
                    "domain": "EAST-ORANGE.MED.VA.GOV",
                    "visn": "2",
                    "ien": "561",
                    "mn": "EAS"
                },
                {
                    "name": "EAST ORANGE, NJ NHCU",
                    "stationNumber": "5619AA",
                    "officialName": "NEW JERSEY HEALTH CARE SYSTEM - EAST ORANGE NHCU",
                    "visn": "2",
                    "ien": 56101
                },
                {
                    "name": "LYONS, NJ NHCU",
                    "stationNumber": "5619AB",
                    "officialName": "NEW JERSEY HEALTH CARE SYSTEM - LYONS NHCU",
                    "visn": "2",
                    "ien": 56102
                },
                {
                    "name": "MENLO PARK, NJ SVH",
                    "stationNumber": "5619AF",
                    "officialName": "VETERANS MEMORIAL HOME - MENLO PARK",
                    "visn": "2",
                    "ien": 56103
                },
                {
                    "name": "PARAMUS, NJ SVH",
                    "stationNumber": "5619AG",
                    "officialName": "VETERANS MEMORIAL HOME - PARAMUS",
                    "visn": "2",
                    "ien": 56104
                },
                {
                    "name": "LYONS- VA NEW JERSEY HCS",
                    "stationNumber": "561A4",
                    "officialName": "NEW JERSEY HEALTH CARE SYSTEM - LYONS CAMPUS",
                    "npi": "1396794293",
                    "visn": "2",
                    "ien": 56105
                },
                {
                    "name": "NJH-DOMICILIARY",
                    "stationNumber": "561BU",
                    "officialName": "NEW JERSEY HEALTH CARE SYSTEM DOMICILIARY",
                    "visn": "2",
                    "ien": 56106
                },
                {
                    "name": "EAST ORANGE VA FACILITY DOM",
                    "stationNumber": "561BV",
                    "visn": "2",
                    "ien": 56107
                },
                {
                    "name": "NEWARK VA CLINIC",
                    "stationNumber": "561BY",
                    "officialName": "NEWARK DTC",
                    "npi": "1053776385",
                    "visn": "2",
                    "ien": 56108
                },
                {
                    "name": "BRICK, NJ CBOC",
                    "stationNumber": "561BZ",
                    "officialName": "JAMES J HOWARD OUTPATIENT CLINIC",
                    "npi": "1205875333",
                    "visn": "2",
                    "ien": 56109
                },
                {
                    "name": "HAMILTON NJ CBOC",
                    "stationNumber": "561GA",
                    "officialName": "HAMILTON CBOC",
                    "npi": "1932148889",
                    "visn": "2",
                    "ien": 56110
                },
                {
                    "name": "ELIZABETH, NJ CBOC",
                    "stationNumber": "561GB",
                    "officialName": "ELIZABETH CBOC",
                    "npi": "1093768038",
                    "visn": "2",
                    "ien": 56111
                },
                {
                    "name": "HACKENSACK, NJ CBOC",
                    "stationNumber": "561GD",
                    "officialName": "HACKENSACK CBOC",
                    "npi": "1366497638",
                    "visn": "2",
                    "ien": 56112
                },
                {
                    "name": "JERSEY CITY, NJ CBOC",
                    "stationNumber": "561GE",
                    "officialName": "JERSEY CITY CBOC",
                    "npi": "1457307829",
                    "visn": "2",
                    "ien": 56113
                },
                {
                    "name": "PISCATAWAY CBOC",
                    "stationNumber": "561GF",
                    "npi": "1366489379",
                    "visn": "2",
                    "ien": 56114
                },
                {
                    "name": "MORRIS PLAINS CBOC",
                    "stationNumber": "561GH",
                    "npi": "1558304535",
                    "visn": "2",
                    "ien": 56115
                },
                {
                    "name": "TINTON FALLS CBOC",
                    "stationNumber": "561GI",
                    "npi": "1215982772",
                    "visn": "2",
                    "ien": 56116
                },
                {
                    "name": "PATERSON CBOC",
                    "stationNumber": "561GJ",
                    "npi": "1861439242",
                    "visn": "2",
                    "ien": 56117
                },
                {
                    "name": "SUSSEX VA CLINIC",
                    "stationNumber": "561GK",
                    "npi": "1912443243",
                    "visn": "2",
                    "ien": 56118
                },
                {
                    "name": "IDES MCGUIRE AFB",
                    "stationNumber": "561MA",
                    "visn": "2",
                    "ien": 56119
                },
                {
                    "name": "NJH-PRRTP",
                    "stationNumber": "561PA",
                    "officialName": "NEW JERSEY HEALTH CARE SYSTEM - PRRTP",
                    "visn": "2",
                    "ien": 56120
                },
                {
                    "name": "EAST ORANGE (SCI)",
                    "stationNumber": "561SCI",
                    "officialName": "EAST ORANGE SPINAL CORD OUTCOMES",
                    "visn": "2",
                    "ien": 56121
                }
            ],
            "620": [
                {
                    "name": "HUDSON VALLEY HCS VAMC",
                    "stationNumber": "620",
                    "officialName": "VA HUDSON VALLEY HEALTH CARE SYSTEM - MONTROSE DIVISION",
                    "npi": "1326097700",
                    "domain": "MONTROSE.MED.VA.GOV",
                    "visn": "2",
                    "ien": "620",
                    "mn": "MON"
                },
                {
                    "name": "MONTROSE",
                    "stationNumber": "6209AA",
                    "officialName": "Montrose Nursing Home",
                    "visn": "2",
                    "ien": 62001
                },
                {
                    "name": "CASTLE POINT",
                    "stationNumber": "620A4",
                    "officialName": "VA HUDSON VALLEY HEALTH CARE - CASTLE POINT DIVISION",
                    "npi": "1639128473",
                    "visn": "2",
                    "ien": 62002
                },
                {
                    "name": "MONTROSE VA FACILITY DOM",
                    "stationNumber": "620BU",
                    "npi": "1477913788",
                    "visn": "2",
                    "ien": 62003
                },
                {
                    "name": "NEW YORK STATE VET HOME",
                    "stationNumber": "620DT",
                    "officialName": "NEW YORK STATE VETERANS HOME",
                    "visn": "2",
                    "ien": 62004
                },
                {
                    "name": "NEW CITY",
                    "stationNumber": "620GA",
                    "npi": "1851339261",
                    "visn": "2",
                    "ien": 62005
                },
                {
                    "name": "PUTNAM COUNTY",
                    "stationNumber": "620GB",
                    "npi": "1437100203",
                    "visn": "2",
                    "ien": 62006
                },
                {
                    "name": "GOSHEN CBOC",
                    "stationNumber": "620GD",
                    "npi": "1851338776",
                    "visn": "2",
                    "ien": 62007
                },
                {
                    "name": "PORT JERVIS",
                    "stationNumber": "620GE",
                    "npi": "1104863760",
                    "visn": "2",
                    "ien": 62008
                },
                {
                    "name": "HARRIS",
                    "stationNumber": "620GF",
                    "npi": "1144275553",
                    "visn": "2",
                    "ien": 62009
                },
                {
                    "name": "POUGHKEEPSIE",
                    "stationNumber": "620GG",
                    "npi": "1033153770",
                    "visn": "2",
                    "ien": 62010
                },
                {
                    "name": "EASTERN DUTCHESS CBOC",
                    "stationNumber": "620GH",
                    "npi": "1467649897",
                    "visn": "2",
                    "ien": 62011
                },
                {
                    "name": "IDES WEST POINT NY",
                    "stationNumber": "620MA",
                    "visn": "2",
                    "ien": 62012
                },
                {
                    "name": "MONTROSE 1 VA MOBILE CLINIC",
                    "stationNumber": "620QA",
                    "npi": "1528466174",
                    "visn": "2",
                    "ien": 62013
                },
                {
                    "name": "MONTROSE 2 VA MOBILE CLINIC",
                    "stationNumber": "620QB",
                    "npi": "1437557089",
                    "visn": "2",
                    "ien": 62014
                },
                {
                    "name": "MONTROSE 3 VA MOBILE CLINIC",
                    "stationNumber": "620QC",
                    "npi": "1073911624",
                    "visn": "2",
                    "ien": 62015
                },
                {
                    "name": "HUDSON VALLEY (SCI)",
                    "stationNumber": "620SCI",
                    "officialName": "HUDSON VALLEY SPINAL CORDS OUTCOMES (SCI)",
                    "visn": "2",
                    "ien": 62016
                }
            ],
            "630": [
                {
                    "name": "NEW YORK HHS",
                    "stationNumber": "630",
                    "officialName": "VA NEW YORK HARBOR HEALTHCARE SYSTEM - NEW YORK DIVISION",
                    "npi": "1083663124",
                    "domain": "NY-HARBOR.MED.VA.GOV",
                    "visn": "2",
                    "ien": "630",
                    "mn": "NYH"
                },
                {
                    "name": "ST. ALBANS NHCU",
                    "stationNumber": "6309AB",
                    "visn": "2",
                    "ien": 63001
                },
                {
                    "name": "BROOKLYN HHS",
                    "stationNumber": "630A4",
                    "officialName": "VA NEW YORK HARBOR HEALTHCARE SYSTEM - BROOKLYN DIVISION",
                    "npi": "1033168356",
                    "visn": "2",
                    "ien": 63002
                },
                {
                    "name": "ST. ALBANS HHS",
                    "stationNumber": "630A5",
                    "officialName": "ST ALBANS EXTENDED CARE CNTR",
                    "npi": "1457395139",
                    "visn": "2",
                    "ien": 63003
                },
                {
                    "name": "BROOKLYN PRRTP",
                    "stationNumber": "630BU",
                    "visn": "2",
                    "ien": 63004
                },
                {
                    "name": "BROOKLYN VA FACILITY DOM",
                    "stationNumber": "630BV",
                    "officialName": "BROOKLYN VA FACILITY DOMICILIARY",
                    "visn": "2",
                    "ien": 63005
                },
                {
                    "name": "NYS VETERANS HOME",
                    "stationNumber": "630DT",
                    "visn": "2",
                    "ien": 63006
                },
                {
                    "name": "HARLEM CBOC",
                    "stationNumber": "630GA",
                    "officialName": "HARLEM CBOC-NYHHS",
                    "npi": "1992750392",
                    "visn": "2",
                    "ien": 63007
                },
                {
                    "name": "STATEN ISLAND CBOC",
                    "stationNumber": "630GB",
                    "officialName": "STATEN ISLAND CBOC-NYHHCS",
                    "npi": "1487693362",
                    "visn": "2",
                    "ien": 63008
                },
                {
                    "name": "CHAPEL ST",
                    "stationNumber": "630GC",
                    "officialName": "CHAPEL STREET OPC",
                    "npi": "1912958711",
                    "visn": "2",
                    "ien": 63009
                },
                {
                    "name": "NEW YORK PRRTP",
                    "stationNumber": "630PA",
                    "visn": "2",
                    "ien": 63010
                },
                {
                    "name": "NY HARBOR 1 VA MOBILE CLINIC",
                    "stationNumber": "630QA",
                    "npi": "1609274257",
                    "visn": "2",
                    "ien": 63011
                },
                {
                    "name": "NY HARBOR 2 VA MOBILE CLINIC",
                    "stationNumber": "630QB",
                    "npi": "1063810612",
                    "visn": "2",
                    "ien": 63012
                }
            ],
            "632": [
                {
                    "name": "NORTHPORT VAMC",
                    "stationNumber": "632",
                    "officialName": "NORTHPORT VAMC NY",
                    "npi": "1033169412",
                    "domain": "NORTHPORT.MED.VA.GOV",
                    "visn": "2",
                    "ien": "632",
                    "mn": "NOR"
                },
                {
                    "name": "NORTHPORT NHCU",
                    "stationNumber": "6329AA",
                    "officialName": "VANURS NORTHPORT",
                    "visn": "2",
                    "ien": 63201
                },
                {
                    "name": "NORTHPORT VA FACILITY DOM",
                    "stationNumber": "632BU",
                    "officialName": "NORTHPORT VA FACILITY DOMILICIARY",
                    "visn": "2",
                    "ien": 63202
                },
                {
                    "name": "STONY BROOK",
                    "stationNumber": "632DT",
                    "officialName": "Stony Brook,NY STNB",
                    "visn": "2",
                    "ien": 63203
                },
                {
                    "name": "EAST MEADOW CBOC",
                    "stationNumber": "632GA",
                    "npi": "1013955566",
                    "visn": "2",
                    "ien": 63204
                },
                {
                    "name": "VALLEY STREAM CBOC",
                    "stationNumber": "632HA",
                    "npi": "1821044116",
                    "visn": "2",
                    "ien": 63205
                },
                {
                    "name": "RIVERHEAD",
                    "stationNumber": "632HB",
                    "officialName": "Riverhead,NY ORC",
                    "npi": "1841238862",
                    "visn": "2",
                    "ien": 63206
                },
                {
                    "name": "BAY SHORE CBOC",
                    "stationNumber": "632HC",
                    "npi": "1215982897",
                    "visn": "2",
                    "ien": 63207
                },
                {
                    "name": "PATCHOGUE",
                    "stationNumber": "632HD",
                    "officialName": "Patchogue,NY ORC",
                    "npi": "1700820438",
                    "visn": "2",
                    "ien": 63208
                },
                {
                    "name": "NOR-PRRTP",
                    "stationNumber": "632PA",
                    "officialName": "Northport,NY PRRTP",
                    "visn": "2",
                    "ien": 63209
                },
                {
                    "name": "NORTHPORT 1 VA MOBILE CLINIC",
                    "stationNumber": "632QA",
                    "npi": "1508264151",
                    "visn": "2",
                    "ien": 63210
                },
                {
                    "name": "NORTHPORT 2 VA MOBILE CLINIC",
                    "stationNumber": "632QB",
                    "npi": "1326446972",
                    "visn": "2",
                    "ien": 63211
                }
            ]
        },
        "20": {
            "531": [
                {
                    "name": "BOISE VA MEDICAL CENTER",
                    "stationNumber": "531",
                    "npi": "1861441867",
                    "domain": "BOISE.MED.VA.GOV",
                    "visn": "20",
                    "ien": "531",
                    "mn": "BOI"
                },
                {
                    "name": "BOISE NURSING HOME",
                    "stationNumber": "5319AA",
                    "visn": "20",
                    "ien": 53101
                },
                {
                    "name": "BOISE STATE NURSING HOME",
                    "stationNumber": "5319AF",
                    "visn": "20",
                    "ien": 53102
                },
                {
                    "name": "BOISE VA FACILITY DOM",
                    "stationNumber": "531BU",
                    "visn": "20",
                    "ien": 53103
                },
                {
                    "name": "BOISE STATE HOME",
                    "stationNumber": "531DT",
                    "visn": "20",
                    "ien": 53104
                },
                {
                    "name": "TWIN FALLS VA CLINIC",
                    "stationNumber": "531GE",
                    "npi": "1164461042",
                    "visn": "20",
                    "ien": 53105
                },
                {
                    "name": "CALDWELL VA CLINIC",
                    "stationNumber": "531GG",
                    "npi": "1124152921",
                    "visn": "20",
                    "ien": 53106
                },
                {
                    "name": "BURNS VA CLINIC",
                    "stationNumber": "531GH",
                    "npi": "1548677602",
                    "visn": "20",
                    "ien": 53107
                },
                {
                    "name": "MOUNTAIN HOME VA CLINIC",
                    "stationNumber": "531GI",
                    "npi": "1003223165",
                    "visn": "20",
                    "ien": 53108
                },
                {
                    "name": "SALMON VA CLINIC",
                    "stationNumber": "531GJ",
                    "npi": "1750798815",
                    "visn": "20",
                    "ien": 53109
                },
                {
                    "name": "IDES MOUNTAIN HOME AFB",
                    "stationNumber": "531MA",
                    "visn": "20",
                    "ien": 53110
                },
                {
                    "name": "IDES IDAHO ARMY NATL GUARD",
                    "stationNumber": "531MB",
                    "visn": "20",
                    "ien": 53111
                }
            ],
            "648": [
                {
                    "name": "PORTLAND VA MEDICAL CENTER",
                    "stationNumber": "648",
                    "npi": "1962452441",
                    "domain": "PORTLAND.MED.VA.GOV",
                    "visn": "20",
                    "ien": "648",
                    "mn": "POR"
                },
                {
                    "name": "VANCOUVER VNC",
                    "stationNumber": "6489AA",
                    "officialName": "VANCOUVER VANC",
                    "visn": "20",
                    "ien": 64801
                },
                {
                    "name": "EDWARD ALLWORTH VETERANS HOME",
                    "stationNumber": "6489AF",
                    "visn": "20",
                    "ien": 64802
                },
                {
                    "name": "VANCOUVER,WA DIV PORTLAND VAMC",
                    "stationNumber": "648A4",
                    "npi": "1487698601",
                    "visn": "20",
                    "ien": 64803
                },
                {
                    "name": "PORTLAND DRRTP",
                    "stationNumber": "648BU",
                    "visn": "20",
                    "ien": 64804
                },
                {
                    "name": "OREGON STATE HOME",
                    "stationNumber": "648DT",
                    "visn": "20",
                    "ien": 64805
                },
                {
                    "name": "BEND, OR CBOC",
                    "stationNumber": "648GA",
                    "npi": "1184675803",
                    "visn": "20",
                    "ien": 64806
                },
                {
                    "name": "SALEM VA CLINIC",
                    "stationNumber": "648GB",
                    "npi": "1376581371",
                    "visn": "20",
                    "ien": 64807
                },
                {
                    "name": "NORTH COAST VA CLINIC",
                    "stationNumber": "648GD",
                    "npi": "1720022940",
                    "visn": "20",
                    "ien": 64808
                },
                {
                    "name": "EAST PORTLAND VA CLINIC",
                    "stationNumber": "648GE",
                    "npi": "1710174974",
                    "visn": "20",
                    "ien": 64809
                },
                {
                    "name": "HILLSBORO VA CLINIC",
                    "stationNumber": "648GF",
                    "npi": "1114151248",
                    "visn": "20",
                    "ien": 64810
                },
                {
                    "name": "WEST LINN VA CLINIC",
                    "stationNumber": "648GG",
                    "npi": "1982957510",
                    "visn": "20",
                    "ien": 64811
                },
                {
                    "name": "NEWPORT VA CLINIC",
                    "stationNumber": "648GH",
                    "npi": "1497162556",
                    "visn": "20",
                    "ien": 64812
                },
                {
                    "name": "PORTLAND VA CLINIC",
                    "stationNumber": "648GI",
                    "npi": "1346657491",
                    "visn": "20",
                    "ien": 64813
                },
                {
                    "name": "THE DALLES VA CLINIC",
                    "stationNumber": "648GJ",
                    "officialName": "LOREN R KAUFMAN VA CLINIC",
                    "npi": "1104233295",
                    "ien": 64814,
                    "visn": "20"
                },
                {
                    "name": "LINCOLN CITY VA CLINIC",
                    "stationNumber": "648GK",
                    "npi": "1639722275",
                    "visn": "20",
                    "ien": 64815
                }
            ],
            "653": [
                {
                    "name": "ROSEBURG VA MEDICAL CENTER",
                    "stationNumber": "653",
                    "npi": "1174572549",
                    "domain": "ROSEBURG.MED.VA.GOV",
                    "visn": "20",
                    "ien": "653",
                    "mn": "ROS"
                },
                {
                    "name": "ROSEBURG NHC",
                    "stationNumber": "6539AA",
                    "visn": "20",
                    "ien": 65301
                },
                {
                    "name": "ROSEBURG VA FACILITY DOM",
                    "stationNumber": "653BV",
                    "officialName": "ROSEBURG VA FACILITY DOMICILIARY",
                    "visn": "20",
                    "ien": 65302
                },
                {
                    "name": "EUGENE, OR CBOC",
                    "stationNumber": "653BY",
                    "npi": "1023061322",
                    "visn": "20",
                    "ien": 65303
                },
                {
                    "name": "NORTH BEND VA CLINIC",
                    "stationNumber": "653GA",
                    "npi": "1841241932",
                    "visn": "20",
                    "ien": 65304
                },
                {
                    "name": "BROOKINGS VA CLINIC",
                    "stationNumber": "653GB",
                    "npi": "1992756639",
                    "visn": "20",
                    "ien": 65305
                },
                {
                    "name": "ROSEBURG SARRTP",
                    "stationNumber": "653PA",
                    "visn": "20",
                    "ien": 65306
                },
                {
                    "name": "DOWNTOWN EUGENE VA CLINIC",
                    "stationNumber": "653QA",
                    "npi": "1841714607",
                    "visn": "20",
                    "ien": 65307
                }
            ],
            "663": [
                {
                    "name": "SEATTLE VA MEDICAL CENTER",
                    "stationNumber": "663",
                    "npi": "1770533218",
                    "domain": "PUGET-SOUND.MED.VA.GOV",
                    "visn": "20",
                    "ien": "663",
                    "mn": "PUG"
                },
                {
                    "name": "SEATTLE",
                    "stationNumber": "6639AA",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - SEATTLE NHC PTF",
                    "visn": "20",
                    "ien": 66301
                },
                {
                    "name": "SEATTLE",
                    "stationNumber": "6639AF",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - SEATTLE STNB PTF",
                    "visn": "20",
                    "ien": 66302
                },
                {
                    "name": "SEATTLE",
                    "stationNumber": "6639AG",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - SEATTLE STNB PTF",
                    "visn": "20",
                    "ien": 66303
                },
                {
                    "name": "AMERICAN LAKE (VAMC)",
                    "stationNumber": "663A4",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - AMERICAN LAKE DIVISION",
                    "npi": "1477502946",
                    "visn": "20",
                    "ien": 66304
                },
                {
                    "name": "AMERICAN LAKE",
                    "stationNumber": "663BU",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - AMERICAN LAKE DOM PTF",
                    "visn": "20",
                    "ien": 66305
                },
                {
                    "name": "SEATTLE",
                    "stationNumber": "663C5",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - SEATTLE PUB HEALTH PTF",
                    "visn": "20",
                    "ien": 66306
                },
                {
                    "name": "MADIGAN",
                    "stationNumber": "663CN",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - MADIGAN PTF",
                    "visn": "20",
                    "ien": 66307
                },
                {
                    "name": "SEATTLE",
                    "stationNumber": "663CT",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - SEATTLE PTF",
                    "visn": "20",
                    "ien": 66308
                },
                {
                    "name": "MADIGAN ARMY MEDICAL CENTER",
                    "stationNumber": "663DO",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - MADIGAN ARMY MEDICAL CENTER",
                    "visn": "20",
                    "ien": 66309
                },
                {
                    "name": "ORTING",
                    "stationNumber": "663DT",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - ORTING STATE HOME",
                    "visn": "20",
                    "ien": 66310
                },
                {
                    "name": "RETSIL",
                    "stationNumber": "663DU",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - RETSIL STATE HOME",
                    "visn": "20",
                    "ien": 66311
                },
                {
                    "name": "BELLEVUE VA CLINIC",
                    "stationNumber": "663GA",
                    "npi": "1477509933",
                    "visn": "20",
                    "ien": 66312
                },
                {
                    "name": "SILVERDALE VA CLINIC",
                    "stationNumber": "663GB",
                    "npi": "1649221375",
                    "visn": "20",
                    "ien": 66313
                },
                {
                    "name": "MOUNT VERNON VA CLINIC",
                    "stationNumber": "663GC",
                    "npi": "1689847345",
                    "visn": "20",
                    "ien": 66314
                },
                {
                    "name": "SOUTH SOUND VA CLINIC",
                    "stationNumber": "663GD",
                    "npi": "1346564614",
                    "visn": "20",
                    "ien": 66315
                },
                {
                    "name": "PORT ANGELES VA CBOC",
                    "stationNumber": "663GE",
                    "npi": "1629485875",
                    "visn": "20",
                    "ien": 66316
                },
                {
                    "name": "FEDERAL WAY VA CLINIC",
                    "stationNumber": "663GF",
                    "npi": "1255981452",
                    "visn": "20",
                    "ien": 66317
                },
                {
                    "name": "NORTH SEATTLE VA CLINIC",
                    "stationNumber": "663GG",
                    "npi": "1407498009",
                    "visn": "20",
                    "ien": 66318
                },
                {
                    "name": "PUGET SOUND VA MOBILE CLINIC",
                    "stationNumber": "663HK",
                    "npi": "1780059964",
                    "visn": "20",
                    "ien": 66319
                },
                {
                    "name": "IDES FORT LEWIS ",
                    "stationNumber": "663MA",
                    "visn": "20",
                    "ien": 66320
                },
                {
                    "name": "IDES MCCHORD AFB",
                    "stationNumber": "663MB",
                    "visn": "20",
                    "ien": 66321
                },
                {
                    "name": "IDES BREMERTON NAVAL HOSPITAL",
                    "stationNumber": "663MC",
                    "visn": "20",
                    "ien": 66322
                },
                {
                    "name": "IDES FORT OAK HARBR NAVL HSPTL",
                    "stationNumber": "663MD",
                    "visn": "20",
                    "ien": 66323
                },
                {
                    "name": "PUG-PRRTP",
                    "stationNumber": "663PA",
                    "officialName": "PUGET SOUND HEALTH CARE SYSTEM - PRRTP DIVISION",
                    "visn": "20",
                    "ien": 66324
                },
                {
                    "name": "PUGET SOUND (SCI)",
                    "stationNumber": "663SCI",
                    "officialName": "PUGET SOUND SPINAL CORD OUTCOMES (SCI)",
                    "visn": "20",
                    "ien": 66325
                }
            ],
            "668": [
                {
                    "name": "MANN-GRANDSTAFF VAMC",
                    "stationNumber": "668",
                    "npi": "1124078969",
                    "domain": "SPOKANE.MED.VA.GOV",
                    "visn": "20",
                    "ien": "668",
                    "mn": "SPO"
                },
                {
                    "name": "SPOKANE NHC",
                    "stationNumber": "6689AA",
                    "visn": "20",
                    "ien": 66801
                },
                {
                    "name": "LEWISTON",
                    "stationNumber": "668DT",
                    "visn": "20",
                    "ien": 66802
                },
                {
                    "name": "WASHINGTON STATE VET HOME",
                    "stationNumber": "668DU",
                    "officialName": "WASHINGTON STATE VETERANS HOME",
                    "domain": "SPOKANE.MED.VA.GOV",
                    "visn": "20",
                    "ien": 66803
                },
                {
                    "name": "WENATCHEE VA CLINIC",
                    "stationNumber": "668GA",
                    "npi": "1740477009",
                    "visn": "20",
                    "ien": 66804
                },
                {
                    "name": "COEUR D ALENE VA CLINIC",
                    "stationNumber": "668GB",
                    "npi": "1609044478",
                    "visn": "20",
                    "ien": 66805
                },
                {
                    "name": "SPOKANE MORC",
                    "stationNumber": "668HK",
                    "npi": "1871968040",
                    "domain": "SPOKANE.MED.VA.GOV",
                    "visn": "20",
                    "ien": 66806
                },
                {
                    "name": "IDES FAIRCHILD AFB",
                    "stationNumber": "668MA",
                    "visn": "20",
                    "ien": 66807
                },
                {
                    "name": "COLVILLE VA CLINIC",
                    "stationNumber": "668QA",
                    "npi": "1811388036",
                    "ien": 66808,
                    "visn": "20"
                },
                {
                    "name": "LIBBY VA CLINIC",
                    "stationNumber": "668QB",
                    "npi": "1437540507",
                    "ien": 66809,
                    "visn": "20"
                },
                {
                    "name": "SANDPOINT VA CLINIC",
                    "stationNumber": "668QD",
                    "npi": "1316338486",
                    "ien": 66810,
                    "visn": "20"
                },
                {
                    "name": "SPOKANE VA CLINIC",
                    "stationNumber": "668QE",
                    "npi": "1821489907",
                    "ien": 66811,
                    "visn": "20"
                }
            ],
            "687": [
                {
                    "name": "JONATHAN M. WAINWRIGHT VAMC",
                    "stationNumber": "687",
                    "npi": "1356393805",
                    "domain": "WALLA-WALLA.MED.VA.GOV",
                    "visn": "20",
                    "ien": "687",
                    "mn": "WWW"
                },
                {
                    "name": "WALLA WALLA NHC",
                    "stationNumber": "6879AA",
                    "visn": "20",
                    "ien": 68701
                },
                {
                    "name": "WALLA WALLA VETERANS HOME",
                    "stationNumber": "6879AF",
                    "visn": "20",
                    "ien": 68702
                },
                {
                    "name": "WALLA WALLA VA FACILITY DOM",
                    "stationNumber": "687BU",
                    "officialName": "WALLA WALLA VA FACILITY DOMICILIARY",
                    "visn": "20",
                    "ien": 68703
                },
                {
                    "name": "RICHLAND VA CLINIC",
                    "stationNumber": "687GA",
                    "npi": "1659311009",
                    "visn": "20",
                    "ien": 68704
                },
                {
                    "name": "LEWISTON VA CLINIC",
                    "stationNumber": "687GB",
                    "npi": "1588610398",
                    "visn": "20",
                    "ien": 68705
                },
                {
                    "name": "LA GRANDE CBOC",
                    "stationNumber": "687GC",
                    "npi": "1184812943",
                    "visn": "20",
                    "ien": 68706
                },
                {
                    "name": "YAKIMA CBOC",
                    "stationNumber": "687HA",
                    "npi": "1992745723",
                    "visn": "20",
                    "ien": 68707
                },
                {
                    "name": "WAL-PRRTP",
                    "stationNumber": "687PA",
                    "officialName": "NURSING HOME",
                    "visn": "20",
                    "ien": 68708
                },
                {
                    "name": "GRANGEVILLE VA CLINIC",
                    "stationNumber": "687QA",
                    "npi": "1093106171",
                    "visn": "20",
                    "ien": 68709
                },
                {
                    "name": "MORROW COUNTY VA CLINIC",
                    "stationNumber": "687QB",
                    "npi": "1457742538",
                    "visn": "20",
                    "ien": 68710
                },
                {
                    "name": "WALLOWA COUNTY VA CLINIC",
                    "stationNumber": "687QC",
                    "npi": "1710378898",
                    "visn": "20",
                    "ien": 68711
                }
            ],
            "692": [
                {
                    "name": "WHITE CITY VA MEDICAL CENTER",
                    "stationNumber": "692",
                    "npi": "1003917592",
                    "domain": "WHITE-CITY.MED.VA.GOV",
                    "visn": "20",
                    "ien": "692",
                    "mn": "WCO"
                },
                {
                    "name": "WHITE CITY, CIVH",
                    "stationNumber": "692DS",
                    "visn": "20",
                    "ien": 69201
                },
                {
                    "name": "KLAMATH FALLS VA CLINIC",
                    "stationNumber": "692GA",
                    "npi": "1093761108",
                    "visn": "20",
                    "ien": 69202
                },
                {
                    "name": "GRANTS PASS VA CLINIC",
                    "stationNumber": "692GB",
                    "officialName": "GRANTS PASS CBOC",
                    "npi": "1790038339",
                    "visn": "20",
                    "ien": 69203
                }
            ]
        },
        "12": {
            "537": [
                {
                    "name": "JESSE BROWN VAMC",
                    "stationNumber": "537",
                    "officialName": "JESSE BROWN VA MEDICAL CENTER",
                    "npi": "1215986047",
                    "domain": "CHICAGO-WEST.MED.VA.GOV",
                    "visn": "12",
                    "ien": "537",
                    "mn": "CHI"
                },
                {
                    "name": "JESSE BROWN NHCU",
                    "stationNumber": "5379AA",
                    "visn": "12",
                    "ien": 53701
                },
                {
                    "name": "CHICAGO VA FACILITY DOM",
                    "stationNumber": "537BU",
                    "officialName": "CHICAGO VA FACILITY DOMICILIARY",
                    "visn": "12",
                    "ien": 53702
                },
                {
                    "name": "CROWNPOINT CL,IN (EFF. 1/1/97)",
                    "stationNumber": "537BY",
                    "officialName": "ADAM BENJAMIN JR OPC",
                    "npi": "1952354672",
                    "visn": "12",
                    "ien": 53703
                },
                {
                    "name": "CHICAGO HEIGHTS, IL",
                    "stationNumber": "537GA",
                    "officialName": "CHICAGO HEIGHTS CBOC",
                    "npi": "1982656112",
                    "visn": "12",
                    "ien": 53704
                },
                {
                    "name": "LAKESIDE CBOC",
                    "stationNumber": "537GD",
                    "npi": "1639126535",
                    "visn": "12",
                    "ien": 53705
                },
                {
                    "name": "BEVERLY CLINIC",
                    "stationNumber": "537HA",
                    "officialName": "BEVERLY CBOC",
                    "npi": "1568402378",
                    "visn": "12",
                    "ien": 53706
                },
                {
                    "name": "CHI-PRRTP",
                    "stationNumber": "537PA",
                    "visn": "12",
                    "ien": 53707
                },
                {
                    "name": "CHICAGO VA CLINIC",
                    "stationNumber": "537QA",
                    "visn": "12",
                    "ien": 53708
                }
            ],
            "550": [
                {
                    "name": "ILLIANA HCS",
                    "stationNumber": "550",
                    "officialName": "DANVILLE VA MEDICAL CENTER",
                    "npi": "1326097163",
                    "domain": "DANVILLE.MED.VA.GOV",
                    "visn": "12",
                    "ien": "550",
                    "mn": "DAN"
                },
                {
                    "name": "DANVILLE (NHC)",
                    "stationNumber": "5509AA",
                    "visn": "12",
                    "ien": 55001
                },
                {
                    "name": "DANVILLE DOM",
                    "stationNumber": "550BU",
                    "visn": "12",
                    "ien": 55002
                },
                {
                    "name": "BOB MICHEL VA OPC",
                    "stationNumber": "550BY",
                    "npi": "1134163736",
                    "visn": "12",
                    "ien": 55003
                },
                {
                    "name": "DECATUR VA CLINIC",
                    "stationNumber": "550GA",
                    "npi": "1124071881",
                    "visn": "12",
                    "ien": 55004
                },
                {
                    "name": "SPRINGFIELD VA CLINIC",
                    "stationNumber": "550GD",
                    "npi": "1235177973",
                    "visn": "12",
                    "ien": 55005
                },
                {
                    "name": "MATTOON CBOC",
                    "stationNumber": "550GF",
                    "npi": "1861725699",
                    "visn": "12",
                    "ien": 55006
                },
                {
                    "name": "BLOOMINGTON VA CLINIC (550GG)",
                    "stationNumber": "550GG",
                    "npi": "1609426709",
                    "visn": "12",
                    "ien": 55007
                },
                {
                    "name": "DANVILLE PRRTP",
                    "stationNumber": "550PA",
                    "visn": "12",
                    "ien": 55008
                }
            ],
            "556": [
                {
                    "name": "CAPTN JAMES LOVELL FED HLT CTR",
                    "stationNumber": "556",
                    "npi": "1487604864",
                    "domain": "N-CHICAGO.MED.VA.GOV",
                    "visn": "12",
                    "ien": "556",
                    "mn": "NCH"
                },
                {
                    "name": "NORTH CHICAGO (NHC)",
                    "stationNumber": "5569AA",
                    "visn": "12",
                    "ien": 55601
                },
                {
                    "name": "NORTH CHICAGO (DOM)",
                    "stationNumber": "556BU",
                    "visn": "12",
                    "ien": 55602
                },
                {
                    "name": "EVANSTON CBOC",
                    "stationNumber": "556GA",
                    "npi": "1013962554",
                    "visn": "12",
                    "ien": 55603
                },
                {
                    "name": "MCHENRY CBOC",
                    "stationNumber": "556GC",
                    "npi": "1811934722",
                    "visn": "12",
                    "ien": 55604
                },
                {
                    "name": "KENOSHA CBOC",
                    "stationNumber": "556GD",
                    "npi": "1144276064",
                    "visn": "12",
                    "ien": 55605
                },
                {
                    "name": "IDES JAMES A. LOVELL HCC",
                    "stationNumber": "556MA",
                    "visn": "12",
                    "ien": 55606
                },
                {
                    "name": "NORTH CHICAGO (PRRTP)",
                    "stationNumber": "556PA",
                    "visn": "12",
                    "ien": 55607
                }
            ],
            "578": [
                {
                    "name": "HINES, IL VAMC",
                    "stationNumber": "578",
                    "officialName": "EDWARD J. HINES JR. HOSPITAL",
                    "npi": "1427007863",
                    "domain": "HINES.MED.VA.GOV",
                    "visn": "12",
                    "ien": "578",
                    "mn": "HIN"
                },
                {
                    "name": "HINES, IL (NHCU)",
                    "stationNumber": "5789AA",
                    "visn": "12",
                    "ien": 57801
                },
                {
                    "name": "HINES, IL (STNB)",
                    "stationNumber": "5789AF",
                    "visn": "12",
                    "ien": 57802
                },
                {
                    "name": "HINES VA FACILITY DOM",
                    "stationNumber": "578BU",
                    "officialName": "HINES VA FACILITY DOMICILIARY",
                    "visn": "12",
                    "ien": 57803
                },
                {
                    "name": "MANTENO ILLINOIS VETERANS HOME",
                    "stationNumber": "578DT",
                    "visn": "12",
                    "ien": 57804
                },
                {
                    "name": "LASALLE ILLINOIS VETERANS HOME",
                    "stationNumber": "578DU",
                    "visn": "12",
                    "ien": 57805
                },
                {
                    "name": "JOLIET CBOC",
                    "stationNumber": "578GA",
                    "officialName": "JOLIET CBOC (578GA)",
                    "npi": "1104872761",
                    "visn": "12",
                    "ien": 57806
                },
                {
                    "name": "KANKAKEE COUNTY CBOC",
                    "stationNumber": "578GC",
                    "npi": "1780631317",
                    "visn": "12",
                    "ien": 57807
                },
                {
                    "name": "AURORA CBOC",
                    "stationNumber": "578GD",
                    "officialName": "AURORA CBOC (578GD)",
                    "npi": "1396796587",
                    "visn": "12",
                    "ien": 57808
                },
                {
                    "name": "HOFFMAN ESTATES VA CLINIC",
                    "stationNumber": "578GE",
                    "npi": "1316992894",
                    "visn": "12",
                    "ien": 57809
                },
                {
                    "name": "LASALLE CBOC",
                    "stationNumber": "578GF",
                    "officialName": "LASSALLE CBOC (578GF)",
                    "npi": "1639125479",
                    "visn": "12",
                    "ien": 57810
                },
                {
                    "name": "OAK LAWN CBOC",
                    "stationNumber": "578GG",
                    "officialName": "OAK LAWN CBOC (578GG)",
                    "npi": "1154368710",
                    "visn": "12",
                    "ien": 57811
                },
                {
                    "name": "HINES, IL (PRRTP)",
                    "stationNumber": "578PA",
                    "visn": "12",
                    "ien": 57812
                },
                {
                    "name": "HINES VA MOBILE CLINIC",
                    "stationNumber": "578QA",
                    "visn": "12",
                    "ien": 57813
                },
                {
                    "name": "HINES SPINAL CORD OUTCOMES",
                    "stationNumber": "578SCI",
                    "officialName": "HINES VAMC SPINAL CORD OUTCOMES (SCI)",
                    "visn": "12",
                    "ien": 57814
                }
            ],
            "585": [
                {
                    "name": "IRON MOUNTAIN VAMC",
                    "stationNumber": "585",
                    "officialName": "Iron Mountain VA Medical Center",
                    "npi": "1679522213",
                    "domain": "IRON-MTN.MED.VA.GOV",
                    "visn": "12",
                    "ien": "585",
                    "mn": "IRO"
                },
                {
                    "name": "IRON MTN NH",
                    "stationNumber": "5859AA",
                    "visn": "12",
                    "ien": 58501
                },
                {
                    "name": "JACOBETTI NH",
                    "stationNumber": "5859AF",
                    "visn": "12",
                    "ien": 58502
                },
                {
                    "name": "JACOBETTI VETERANS HOME",
                    "stationNumber": "585DT",
                    "visn": "12",
                    "ien": 58503
                },
                {
                    "name": "HANCOCK CBOC",
                    "stationNumber": "585GA",
                    "npi": "1386699916",
                    "visn": "12",
                    "ien": 58504
                },
                {
                    "name": "RHINELANDER CBOC",
                    "stationNumber": "585GB",
                    "npi": "1164460184",
                    "visn": "12",
                    "ien": 58505
                },
                {
                    "name": "MENOMINEE CBOC",
                    "stationNumber": "585GC",
                    "npi": "1194763185",
                    "visn": "12",
                    "ien": 58506
                },
                {
                    "name": "IRONWOOD CBOC",
                    "stationNumber": "585GD",
                    "npi": "1720028319",
                    "visn": "12",
                    "ien": 58507
                },
                {
                    "name": "MANISTIQUE VA CLINIC",
                    "stationNumber": "585GF",
                    "npi": "1114334174",
                    "visn": "12",
                    "ien": 58508
                },
                {
                    "name": "MARQUETTE CBOC",
                    "stationNumber": "585HA",
                    "npi": "1437106275",
                    "visn": "12",
                    "ien": 58509
                },
                {
                    "name": "SAULT SAINT MARIE CBOC",
                    "stationNumber": "585HB",
                    "npi": "1588602254",
                    "visn": "12",
                    "ien": 58510
                },
                {
                    "name": "IRON MTN PRRTP",
                    "stationNumber": "585PA",
                    "visn": "12",
                    "ien": 58511
                }
            ],
            "607": [
                {
                    "name": "WILLIAM S. MIDDLETON VAMC",
                    "stationNumber": "607",
                    "officialName": "WILLIAM S. MIDDLETON MEMORIAL VA HOSPITAL",
                    "npi": "1275583619",
                    "domain": "MADISON.MED.VA.GOV",
                    "visn": "12",
                    "ien": "607",
                    "mn": "MAD"
                },
                {
                    "name": "MADISON VA COMMUNITY LVNG CTR",
                    "stationNumber": "6079AA",
                    "npi": "1558721860",
                    "visn": "12",
                    "ien": 60701
                },
                {
                    "name": "KING STATE NURSING HOME",
                    "stationNumber": "6079AF",
                    "visn": "12",
                    "ien": 60702
                },
                {
                    "name": "MADISON VA FACILITY DOM",
                    "stationNumber": "607BU",
                    "officialName": "MADISON VA FACILITY DOMICILIARY",
                    "visn": "12",
                    "ien": 60703
                },
                {
                    "name": "KING",
                    "stationNumber": "607DT",
                    "officialName": "WISCONSIN VETERANS HOME",
                    "visn": "12",
                    "ien": 60704
                },
                {
                    "name": "JANESVILLE, WI CBOC",
                    "stationNumber": "607GC",
                    "npi": "1124074810",
                    "visn": "12",
                    "ien": 60705
                },
                {
                    "name": "BARABOO, WI CBOC",
                    "stationNumber": "607GD",
                    "npi": "1972554913",
                    "visn": "12",
                    "ien": 60706
                },
                {
                    "name": "BEAVER DAM, WI CBOC",
                    "stationNumber": "607GE",
                    "npi": "1992758254",
                    "visn": "12",
                    "ien": 60707
                },
                {
                    "name": "FREEPORT CBOC",
                    "stationNumber": "607GF",
                    "npi": "1558414987",
                    "visn": "12",
                    "ien": 60708
                },
                {
                    "name": "MADISON WEST VA CLINIC",
                    "stationNumber": "607GG",
                    "npi": "1023425089",
                    "visn": "12",
                    "ien": 60709
                },
                {
                    "name": "ROCKFORD,IL OPC",
                    "stationNumber": "607HA",
                    "officialName": "ROCKFORD, IL OPC",
                    "npi": "1174567770",
                    "visn": "12",
                    "ien": 60710
                },
                {
                    "name": "MAD-SARTP",
                    "stationNumber": "607PA",
                    "officialName": "MADISON, WI SARTP",
                    "visn": "12",
                    "ien": 60711
                }
            ],
            "676": [
                {
                    "name": "TOMAH VAMC",
                    "stationNumber": "676",
                    "npi": "1356391775",
                    "domain": "TOMAH.MED.VA.GOV",
                    "visn": "12",
                    "ien": "676",
                    "mn": "TOM"
                },
                {
                    "name": "TOMAH NHC",
                    "stationNumber": "6769AA",
                    "visn": "12",
                    "ien": 67601
                },
                {
                    "name": "TOMAH VA FACILITY DOM",
                    "stationNumber": "676BU",
                    "officialName": "TOMAH VA FACILITY DOMICILIARY",
                    "visn": "12",
                    "ien": 67602
                },
                {
                    "name": "WAUSAU",
                    "stationNumber": "676GA",
                    "officialName": "WAUSAU CBOC",
                    "npi": "1275573677",
                    "visn": "12",
                    "ien": 67603
                },
                {
                    "name": "LEAVENWORTH, WI CBOC",
                    "stationNumber": "676GB",
                    "visn": "12",
                    "ien": 67604
                },
                {
                    "name": "LACROSS",
                    "stationNumber": "676GC",
                    "officialName": "LACROSSE CBOC",
                    "npi": "1104873298",
                    "visn": "12",
                    "ien": 67605
                },
                {
                    "name": "WISCONSIN RAPIDS",
                    "stationNumber": "676GD",
                    "officialName": "WISCONSIN RAPIDS CBOC",
                    "npi": "1881634616",
                    "visn": "12",
                    "ien": 67606
                },
                {
                    "name": "CLARK COUNTY CBOC",
                    "stationNumber": "676GE",
                    "npi": "1902853567",
                    "visn": "12",
                    "ien": 67607
                },
                {
                    "name": "TOMAH VA MOBILE CLINIC",
                    "stationNumber": "676QA",
                    "npi": "1770978637",
                    "visn": "12",
                    "ien": 67608
                }
            ],
            "695": [
                {
                    "name": "MILWAUKEE VAMC",
                    "stationNumber": "695",
                    "officialName": "Clement J Zablocki",
                    "npi": "1073563417",
                    "domain": "MILWAUKEE.MED.VA.GOV",
                    "visn": "12",
                    "ien": "695",
                    "mn": "MIL"
                },
                {
                    "name": "MILWAUKEE NHC",
                    "stationNumber": "6959AA",
                    "officialName": "Clement J Zablocki",
                    "visn": "12",
                    "ien": 69501
                },
                {
                    "name": "CLEMNT J ZABLOCKI DOM",
                    "stationNumber": "695BU",
                    "officialName": "Clement J Zablocki",
                    "visn": "12",
                    "ien": 69502
                },
                {
                    "name": "FOX VALLEY",
                    "stationNumber": "695BY",
                    "officialName": "Fox Valley",
                    "npi": "1518918705",
                    "visn": "12",
                    "ien": 69503
                },
                {
                    "name": "CLEMENT J ZABLOCKI",
                    "stationNumber": "695CA",
                    "officialName": "Clement J Zablocki",
                    "visn": "12",
                    "ien": 69504
                },
                {
                    "name": "CLEMENT J ZABLOCKI",
                    "stationNumber": "695CB",
                    "officialName": "Clement J Zablocki",
                    "visn": "12",
                    "ien": 69505
                },
                {
                    "name": "WISCONSIN SVH 695DT (DOM)",
                    "stationNumber": "695DT",
                    "officialName": "Wisconsin State Veterans Home",
                    "visn": "12",
                    "ien": 69506
                },
                {
                    "name": "WISCONSIN STATE NURSING HOME",
                    "stationNumber": "695DU",
                    "visn": "12",
                    "ien": 69507
                },
                {
                    "name": "UNION GROVE",
                    "stationNumber": "695GA",
                    "officialName": "Union Grove",
                    "npi": "1326087222",
                    "visn": "12",
                    "ien": 69508
                },
                {
                    "name": "CLEVELAND, WI",
                    "stationNumber": "695GC",
                    "officialName": "CLEVELAND CBOC",
                    "npi": "1609828748",
                    "visn": "12",
                    "ien": 69509
                },
                {
                    "name": "MILO C. HUEMPFNER CBOC",
                    "stationNumber": "695GD",
                    "npi": "1003861014",
                    "visn": "12",
                    "ien": 69510
                },
                {
                    "name": "MILWAUKEE MOC",
                    "stationNumber": "695HK",
                    "officialName": "Clement J Zablocki",
                    "visn": "12",
                    "ien": 69511
                },
                {
                    "name": "MIL-PRRTP",
                    "stationNumber": "695PA",
                    "officialName": "MILWAUKEE PSYCHOLOGICAL RESIDENTIAL REHABILITATION TREATMENT PROGRAM",
                    "visn": "12",
                    "ien": 69512
                },
                {
                    "name": "MILWAUKEE VA CLINIC",
                    "stationNumber": "695QA",
                    "npi": "1285005504",
                    "visn": "12",
                    "ien": 69513
                },
                {
                    "name": "MILWAUKEE SPINAL CORD OUTCOMES",
                    "stationNumber": "695SCI",
                    "officialName": "CLEMENT J ZABLOCKI VAMC SPINAL CORD OUTCOMES (SCI)",
                    "visn": "12",
                    "ien": 69514
                }
            ]
        },
        "6": {
            "558": [
                {
                    "name": "DURHAM VA MEDICAL CENTER",
                    "stationNumber": "558",
                    "npi": "1992754758",
                    "domain": "DURHAM.MED.VA.GOV",
                    "visn": "6",
                    "ien": "558",
                    "mn": "DUR"
                },
                {
                    "name": "DURHAM (NHC)",
                    "stationNumber": "5589AA",
                    "officialName": "DURHAM NHC",
                    "visn": "6",
                    "ien": 55801
                },
                {
                    "name": "NORTH CAROLINA SVH",
                    "stationNumber": "558DT",
                    "visn": "6",
                    "ien": 55802
                },
                {
                    "name": "GREENVILLE VA CLINIC NC",
                    "stationNumber": "558GA",
                    "npi": "1083669030",
                    "visn": "6",
                    "ien": 55803
                },
                {
                    "name": "RALEIGH VA CLINIC",
                    "stationNumber": "558GB",
                    "npi": "1760420855",
                    "visn": "6",
                    "ien": 55804
                },
                {
                    "name": "MOREHEAD CITY VA CLINIC",
                    "stationNumber": "558GC",
                    "npi": "1023056488",
                    "visn": "6",
                    "ien": 55805
                },
                {
                    "name": "DURHAM COUNTY VA CLINIC",
                    "stationNumber": "558GD",
                    "npi": "1104233204",
                    "visn": "6",
                    "ien": 55806
                },
                {
                    "name": "HILLANDALE ROAD VA CLINIC",
                    "stationNumber": "558GE",
                    "npi": "1013324110",
                    "visn": "6",
                    "ien": 55807
                },
                {
                    "name": "WAKE COUNTY VA CLINIC",
                    "stationNumber": "558GF",
                    "npi": "1972910057",
                    "visn": "6",
                    "ien": 55808
                },
                {
                    "name": "RALEIGH III VA CLINIC",
                    "stationNumber": "558GG",
                    "npi": "1124475892",
                    "visn": "6",
                    "ien": 55809
                },
                {
                    "name": "CLAYTON-EAST RALEIGH VA CLINIC",
                    "stationNumber": "558GH",
                    "npi": "1003446105",
                    "visn": "6",
                    "ien": 55810
                },
                {
                    "name": "IDES NAVEL HC CHERRY POINT",
                    "stationNumber": "558MA",
                    "visn": "6",
                    "ien": 55811
                },
                {
                    "name": "DURHAM PRRTP",
                    "stationNumber": "558PA",
                    "visn": "6",
                    "ien": 55812
                },
                {
                    "name": "BRIER CREEK VA CLINIC",
                    "stationNumber": "558QA",
                    "npi": "1609274943",
                    "visn": "6",
                    "ien": 55813
                }
            ],
            "565": [
                {
                    "name": "FAYETTEVILLE VA MEDICAL CENTER",
                    "stationNumber": "565",
                    "npi": "1578516902",
                    "domain": "FAYETTVL-NC.MED.VA.GOV",
                    "visn": "6",
                    "ien": "565",
                    "mn": "FAY"
                },
                {
                    "name": "FAYETTEVILLE CLC",
                    "stationNumber": "5659AA",
                    "npi": "1265892574",
                    "visn": "6",
                    "ien": 56501
                },
                {
                    "name": "JACKSONVILLE VA CLINIC",
                    "stationNumber": "565GA",
                    "npi": "1568417798",
                    "visn": "6",
                    "ien": 56502
                },
                {
                    "name": "WILMINGTON VA CLINIC (NC)",
                    "stationNumber": "565GC",
                    "npi": "1043253156",
                    "visn": "6",
                    "ien": 56503
                },
                {
                    "name": "HAMLET VA CLINIC",
                    "stationNumber": "565GD",
                    "npi": "1396903456",
                    "visn": "6",
                    "ien": 56504
                },
                {
                    "name": "ROBESON COUNTY VA CLINIC",
                    "stationNumber": "565GE",
                    "npi": "1912201146",
                    "visn": "6",
                    "ien": 56505
                },
                {
                    "name": "GOLDSBORO VA CLINIC",
                    "stationNumber": "565GF",
                    "npi": "1306189451",
                    "visn": "6",
                    "ien": 56506
                },
                {
                    "name": "LEE COUNTY VA CLINIC (NC)",
                    "stationNumber": "565GG",
                    "officialName": "LEE COUNTY VA CLINIC",
                    "visn": "6",
                    "ien": 56507
                },
                {
                    "name": "BRUNSWICK COUNTY VA CLINIC",
                    "stationNumber": "565GH",
                    "npi": "1962819052",
                    "visn": "6",
                    "ien": 56508
                },
                {
                    "name": "JACKSONVILLE 2 VA CLINIC-NC",
                    "stationNumber": "565GJ",
                    "npi": "1124422308",
                    "visn": "6",
                    "ien": 56509
                },
                {
                    "name": "FAYETTEVILLE VA CLINIC",
                    "stationNumber": "565GK",
                    "npi": "1013311208",
                    "visn": "6",
                    "ien": 56510
                },
                {
                    "name": "CUMBERLAND COUNTY VA CLINIC-NC",
                    "stationNumber": "565GL",
                    "officialName": "CUMBERLAND COUNTY VA CLINIC",
                    "npi": "1972974392",
                    "visn": "6",
                    "ien": 56511
                },
                {
                    "name": "JACKSONVILLE 3 VA CLINIC",
                    "stationNumber": "565GM",
                    "npi": "1659879138",
                    "visn": "6",
                    "ien": 56512
                },
                {
                    "name": "JACKSONVILLE 4 VA CLINIC",
                    "stationNumber": "565GN",
                    "npi": "1538799002",
                    "visn": "6",
                    "ien": 56513
                },
                {
                    "name": "IDES SEYMOUR JOHNSON AFB",
                    "stationNumber": "565MA",
                    "visn": "6",
                    "ien": 56514
                },
                {
                    "name": "IDES FORT BRAGG",
                    "stationNumber": "565MB",
                    "visn": "6",
                    "ien": 56515
                },
                {
                    "name": "IDES FORT POPE",
                    "stationNumber": "565MC",
                    "visn": "6",
                    "ien": 56516
                },
                {
                    "name": "IDES CAMP LEJEUNE",
                    "stationNumber": "565MD",
                    "visn": "6",
                    "ien": 56517
                },
                {
                    "name": "FAYETTEVILLE PRRTP",
                    "stationNumber": "565PA",
                    "visn": "6",
                    "ien": 56518
                },
                {
                    "name": "ROBESON STREET VA CLINIC",
                    "stationNumber": "565QA",
                    "npi": "1801294145",
                    "visn": "6",
                    "ien": 56519
                },
                {
                    "name": "FAYETTVILLE VA MOBILE CLINIC",
                    "stationNumber": "565QB",
                    "visn": "6",
                    "ien": 56520
                },
                {
                    "name": "FAYETTEVILLE 2 MOBILE CLINIC",
                    "stationNumber": "565QC",
                    "npi": "1649721309",
                    "visn": "6",
                    "ien": 56521
                },
                {
                    "name": "RAEFORD ROAD VA CLINIC",
                    "stationNumber": "565QD",
                    "visn": "6",
                    "ien": 56522
                },
                {
                    "name": "WOMACK VA CLINIC",
                    "stationNumber": "565QE",
                    "npi": "1023663085",
                    "visn": "6",
                    "ien": 56523
                }
            ],
            "590": [
                {
                    "name": "HAMPTON VA MEDICAL CENTER",
                    "stationNumber": "590",
                    "npi": "1316997356",
                    "domain": "HAMPTON.MED.VA.GOV",
                    "visn": "6",
                    "ien": "590",
                    "mn": "HAM"
                },
                {
                    "name": "HAMPTON (NHCU)",
                    "stationNumber": "5909AA",
                    "visn": "6",
                    "ien": 59001
                },
                {
                    "name": "HAMPTON (DOM)",
                    "stationNumber": "590BU",
                    "visn": "6",
                    "ien": 59002
                },
                {
                    "name": "LANGLEY",
                    "stationNumber": "590CZ",
                    "visn": "6",
                    "ien": 59003
                },
                {
                    "name": "VIRGINIA BEACH VA CLINIC",
                    "stationNumber": "590GB",
                    "npi": "1194995241",
                    "visn": "6",
                    "ien": 59004
                },
                {
                    "name": "ALBEMARLE VA CLINIC",
                    "stationNumber": "590GC",
                    "npi": "1174800585",
                    "ien": 59005,
                    "visn": "6"
                },
                {
                    "name": "CHESAPEAKE VA CLINIC",
                    "stationNumber": "590GD",
                    "npi": "1629448469",
                    "visn": "6",
                    "ien": 59006
                },
                {
                    "name": "IDES NAVAL MED CTR PORTSMOUTH",
                    "stationNumber": "590MA",
                    "visn": "6",
                    "ien": 59007
                },
                {
                    "name": "IDES USAF HOSPITAL LANGLEY",
                    "stationNumber": "590MB",
                    "visn": "6",
                    "ien": 59008
                },
                {
                    "name": "IDES FORT EUSTIS",
                    "stationNumber": "590MC",
                    "visn": "6",
                    "ien": 59009
                },
                {
                    "name": "HAMPTON (PRRTP)",
                    "stationNumber": "590PA",
                    "visn": "6",
                    "ien": 59010
                },
                {
                    "name": "HAMPTON VA MOBILE CLINIC",
                    "stationNumber": "590QA",
                    "npi": "1922496140",
                    "visn": "6",
                    "ien": 59011
                },
                {
                    "name": "HAMPTON 1 VA MOBILE CLINIC",
                    "stationNumber": "590QB",
                    "officialName": "HAMPTON CITY CNTY MOBILE CLNC",
                    "npi": "1831587054",
                    "visn": "6",
                    "ien": 59012
                }
            ],
            "637": [
                {
                    "name": "CHARLES GEORGE VAMC",
                    "stationNumber": "637",
                    "npi": "1558311357",
                    "domain": "ASHEVILLE.MED.VA.GOV",
                    "visn": "6",
                    "ien": "637",
                    "mn": "ASH"
                },
                {
                    "name": "ASHEVILLE NURSING HOME",
                    "stationNumber": "6379AA",
                    "visn": "6",
                    "ien": 63701
                },
                {
                    "name": "ASHEVILLE VA FACILITY DOM",
                    "stationNumber": "637BU",
                    "officialName": "ASHEVILLE VA FACILITY DOMICILIARY",
                    "visn": "6",
                    "ien": 63702
                },
                {
                    "name": "FRANKLIN VA CLINIC",
                    "stationNumber": "637GA",
                    "npi": "1609032135",
                    "visn": "6",
                    "ien": 63703
                },
                {
                    "name": "RUTHERFORD COUNTY VA CLINIC",
                    "stationNumber": "637GB",
                    "npi": "1811225428",
                    "visn": "6",
                    "ien": 63704
                },
                {
                    "name": "HICKORY VA CLINIC",
                    "stationNumber": "637GC",
                    "npi": "1619350261",
                    "visn": "6",
                    "ien": 63705
                },
                {
                    "name": "ASHEVILLE-PRRTP",
                    "stationNumber": "637PA",
                    "visn": "6",
                    "ien": 63706
                }
            ],
            "652": [
                {
                    "name": "HUNTER HOLMES MCGUIRE HOSPITAL",
                    "stationNumber": "652",
                    "npi": "1619927027",
                    "domain": "RICHMOND.MED.VA.GOV",
                    "visn": "6",
                    "ien": "652",
                    "mn": "RIC"
                },
                {
                    "name": "NHC RICHMOND",
                    "stationNumber": "6529AA",
                    "visn": "6",
                    "ien": 65201
                },
                {
                    "name": "RICHMOND VA FACILITY DOM",
                    "stationNumber": "652BU",
                    "officialName": "RICHMOND VA FACILITY DOMICILIARY",
                    "visn": "6",
                    "ien": 65202
                },
                {
                    "name": "SITTER AND BARFOOT VET CR CTR",
                    "stationNumber": "652DT",
                    "visn": "6",
                    "ien": 65203
                },
                {
                    "name": "FREDERICKSBURG VA CLINIC",
                    "stationNumber": "652GA",
                    "npi": "1093752099",
                    "visn": "6",
                    "ien": 65204
                },
                {
                    "name": "FREDERICKSBURG 2 VA CLINIC",
                    "stationNumber": "652GB",
                    "npi": "1790145167",
                    "visn": "6",
                    "ien": 65205
                },
                {
                    "name": "CHARLOTTESVILLE VA CLINIC",
                    "stationNumber": "652GE",
                    "npi": "1154599470",
                    "visn": "6",
                    "ien": 65206
                },
                {
                    "name": "EMPORIA VA CLINIC",
                    "stationNumber": "652GF",
                    "npi": "1598072902",
                    "visn": "6",
                    "ien": 65207
                },
                {
                    "name": "RICHMOND VA MOBILE CLINIC",
                    "stationNumber": "652GG",
                    "npi": "1528475605",
                    "visn": "6",
                    "ien": 65208
                },
                {
                    "name": "HUNTER HOLMES MCGUIRE 2 MOB CL",
                    "stationNumber": "652GH",
                    "npi": "1073920179",
                    "visn": "6",
                    "ien": 65209
                },
                {
                    "name": "IDES FORT LEE",
                    "stationNumber": "652MA",
                    "visn": "6",
                    "ien": 65210
                },
                {
                    "name": "RICHMOND PRRTP (VA)",
                    "stationNumber": "652PA",
                    "visn": "6",
                    "ien": 65211
                },
                {
                    "name": "RICHMOND SPINAL CORD OUTCOMES",
                    "stationNumber": "652SCI",
                    "officialName": "RICHMOND SPINAL CORD OUTCOMES (SCI)",
                    "visn": "6",
                    "ien": 65212
                }
            ],
            "658": [
                {
                    "name": "SALEM VA MEDICAL CENTER",
                    "stationNumber": "658",
                    "npi": "1841242377",
                    "domain": "SALEM.MED.VA.GOV",
                    "visn": "6",
                    "ien": "658",
                    "mn": "SAL"
                },
                {
                    "name": "SALEM NHC",
                    "stationNumber": "6589AA",
                    "visn": "6",
                    "ien": 65801
                },
                {
                    "name": "SALEM VA FACILITY DOM",
                    "stationNumber": "658BU",
                    "officialName": "SALEM VA FACILITY DOMICILIARY",
                    "visn": "6",
                    "ien": 65802
                },
                {
                    "name": "VIRGINIA VETERANS CARE CENTER",
                    "stationNumber": "658DT",
                    "officialName": "VA VETERANS CARE CENTER",
                    "visn": "6",
                    "ien": 65803
                },
                {
                    "name": "TAZEWELL VA CLINIC",
                    "stationNumber": "658GA",
                    "npi": "1124067012",
                    "visn": "6",
                    "ien": 65804
                },
                {
                    "name": "DANVILLE VA CLINIC",
                    "stationNumber": "658GB",
                    "npi": "1245283878",
                    "visn": "6",
                    "ien": 65805
                },
                {
                    "name": "LYNCHBURG VA CLINIC",
                    "stationNumber": "658GC",
                    "npi": "1851554232",
                    "visn": "6",
                    "ien": 65806
                },
                {
                    "name": "STAUNTON CBOC",
                    "stationNumber": "658GD",
                    "npi": "1982986006",
                    "visn": "6",
                    "ien": 65807
                },
                {
                    "name": "WYTHEVILLE VA CLINIC",
                    "stationNumber": "658GE",
                    "npi": "1154626125",
                    "visn": "6",
                    "ien": 65808
                },
                {
                    "name": "SALEM PRRTP (VA)",
                    "stationNumber": "658PA",
                    "visn": "6",
                    "ien": 65809
                }
            ],
            "659": [
                {
                    "name": "W.G. HEFNER SALISBURY VAMC",
                    "stationNumber": "659",
                    "npi": "1750333282",
                    "domain": "SALISBURY.MED.VA.GOV",
                    "visn": "6",
                    "ien": "659",
                    "mn": "SAL"
                },
                {
                    "name": "SALISBURY NHC",
                    "stationNumber": "6599AA",
                    "visn": "6",
                    "ien": 65901
                },
                {
                    "name": "SALISBURY STNH",
                    "stationNumber": "6599F",
                    "visn": "6",
                    "ien": 65902
                },
                {
                    "name": "SALISBURY VA FACILITY DOM",
                    "stationNumber": "659BU",
                    "officialName": "SALISBURY VA FACILITY DOMICILIARY",
                    "visn": "6",
                    "ien": 65903
                },
                {
                    "name": "KERNERSVILLE VA CLINIC",
                    "stationNumber": "659BY",
                    "npi": "1073557559",
                    "visn": "6",
                    "ien": 65904
                },
                {
                    "name": "SOUTH CHARLOTTE VA CLINIC",
                    "stationNumber": "659BZ",
                    "npi": "1457719312",
                    "visn": "6",
                    "ien": 65905
                },
                {
                    "name": "WINSTON",
                    "stationNumber": "659CN",
                    "visn": "6",
                    "ien": 65906
                },
                {
                    "name": "WINSTON",
                    "stationNumber": "659CT",
                    "visn": "6",
                    "ien": 65907
                },
                {
                    "name": "NORTH CHARLOTTE VA CLINIC",
                    "stationNumber": "659GA",
                    "npi": "1477504280",
                    "visn": "6",
                    "ien": 65908
                },
                {
                    "name": "PETERS CREEK VA CLINIC",
                    "stationNumber": "659GC",
                    "npi": "1184031262",
                    "visn": "6",
                    "ien": 65909
                },
                {
                    "name": "SAL-PRRTP",
                    "stationNumber": "659PA",
                    "officialName": "SALISBURY PSYCHOLOGICAL RESIDENTIAL REHABILITATION TREATMENT PROGRAM",
                    "visn": "6",
                    "ien": 65910
                },
                {
                    "name": "CHARLOTTE 5TH ST VA CLINIC",
                    "stationNumber": "659QA",
                    "npi": "1144618356",
                    "visn": "6",
                    "ien": 65911
                }
            ]
        },
        "21": {
            "570": [
                {
                    "name": "FRESNO VA MEDICAL CENTER",
                    "stationNumber": "570",
                    "npi": "1508815291",
                    "visn": "21",
                    "ien": "570"
                },
                {
                    "name": "FRESNO NHC",
                    "stationNumber": "5709AA",
                    "visn": "21",
                    "ien": 57001
                },
                {
                    "name": "STATE VETERANS HOME FRESNO",
                    "stationNumber": "570DT",
                    "visn": "21",
                    "ien": 57002
                },
                {
                    "name": "MERCED VA CLINIC",
                    "stationNumber": "570GA",
                    "npi": "1619928306",
                    "visn": "21",
                    "ien": 57003
                },
                {
                    "name": "TULARE VA CLINIC",
                    "stationNumber": "570GB",
                    "npi": "1265472658",
                    "visn": "21",
                    "ien": 57004
                },
                {
                    "name": "OAKHURST VA CLINIC",
                    "stationNumber": "570GC",
                    "npi": "1235431552",
                    "visn": "21",
                    "ien": 57005
                },
                {
                    "name": "IDES NAVAL HOSPITAL LEMOORE",
                    "stationNumber": "570MA",
                    "visn": "21",
                    "ien": 57006
                }
            ],
            "612": [
                {
                    "name": "NORTHERN CALIFORNIA HCS",
                    "stationNumber": "612",
                    "npi": "1194826685",
                    "visn": "21",
                    "ien": "612"
                },
                {
                    "name": "MARTINEZ NHCU",
                    "stationNumber": "6129AA",
                    "visn": "21",
                    "ien": 61201
                },
                {
                    "name": "SACRAMENTO VA MEDICAL CENTER",
                    "stationNumber": "612A4",
                    "npi": "1497705834",
                    "visn": "21",
                    "ien": 61202
                },
                {
                    "name": "REDDING VA CLINIC",
                    "stationNumber": "612B4",
                    "npi": "1730127762",
                    "visn": "21",
                    "ien": 61203
                },
                {
                    "name": "OAKLAND VA CLINIC",
                    "stationNumber": "612BY",
                    "npi": "1023058609",
                    "visn": "21",
                    "ien": 61204
                },
                {
                    "name": "REDDING STATE VETERANS HOME",
                    "stationNumber": "612DT",
                    "visn": "21",
                    "ien": 61205
                },
                {
                    "name": "FAIRFIELD VA CLINIC",
                    "stationNumber": "612GD",
                    "npi": "1679513089",
                    "visn": "21",
                    "ien": 61206
                },
                {
                    "name": "MARE ISLAND VA CLINIC",
                    "stationNumber": "612GE",
                    "npi": "1962443572",
                    "visn": "21",
                    "ien": 61207
                },
                {
                    "name": "MARTINEZ VA MEDICAL CENTER",
                    "stationNumber": "612GF",
                    "npi": "1417998931",
                    "visn": "21",
                    "ien": 61208
                },
                {
                    "name": "CHICO VA CLINIC",
                    "stationNumber": "612GG",
                    "npi": "1235183013",
                    "visn": "21",
                    "ien": 61209
                },
                {
                    "name": "MCCLELLAN VA CLINIC",
                    "stationNumber": "612GH",
                    "npi": "1225071756",
                    "visn": "21",
                    "ien": 61210
                },
                {
                    "name": "YUBA CITY VA CLINIC",
                    "stationNumber": "612GI",
                    "npi": "1447523196",
                    "visn": "21",
                    "ien": 61211
                },
                {
                    "name": "YREKA VA CLINIC",
                    "stationNumber": "612GJ",
                    "npi": "1144638958",
                    "visn": "21",
                    "ien": 61212
                },
                {
                    "name": "SIERRA FOOTHILLS VA CLINIC-CA",
                    "stationNumber": "612GK",
                    "npi": "1851943104",
                    "visn": "21",
                    "ien": 61213
                },
                {
                    "name": "IDES MCCLELLAN OPC",
                    "stationNumber": "612MA",
                    "visn": "21",
                    "ien": 61214
                },
                {
                    "name": "COHASSET VA CLINIC",
                    "stationNumber": "612QA",
                    "npi": "1124410550",
                    "visn": "21",
                    "ien": 61215
                },
                {
                    "name": "TWENTY FIRST STREET VA CLINIC",
                    "stationNumber": "612QB",
                    "npi": "1972995306",
                    "visn": "21",
                    "ien": 61216
                },
                {
                    "name": "CYPRESS AVENUE VA CLINIC",
                    "stationNumber": "612QC",
                    "npi": "1033501465",
                    "visn": "21",
                    "ien": 61217
                },
                {
                    "name": "HOWE ROAD VA CLINIC",
                    "stationNumber": "612QD",
                    "npi": "1073064408",
                    "visn": "21",
                    "ien": 61218
                }
            ],
            "640": [
                {
                    "name": "PALO ALTO VA MEDICAL CENTER",
                    "stationNumber": "640",
                    "npi": "1942259320",
                    "visn": "21",
                    "ien": "640"
                },
                {
                    "name": "PALO ALTO NHC",
                    "stationNumber": "6409AA",
                    "visn": "21",
                    "ien": 64001
                },
                {
                    "name": "LIVERMORE NHC",
                    "stationNumber": "6409AB",
                    "visn": "21",
                    "ien": 64002
                },
                {
                    "name": "MENLO PARK VA NHC",
                    "stationNumber": "6409AC",
                    "visn": "21",
                    "ien": 64003
                },
                {
                    "name": "MENLO PARK",
                    "stationNumber": "640A0",
                    "officialName": "PALO ALTO HEALTH CARE SYSTEM - MENLO PARK DIVSION",
                    "npi": "1750330213",
                    "visn": "21",
                    "ien": 64004
                },
                {
                    "name": "PALO ALTO VAMC-LIVERMORE",
                    "stationNumber": "640A4",
                    "npi": "1033168133",
                    "visn": "21",
                    "ien": 64005
                },
                {
                    "name": "VADOM MENLO PARK",
                    "stationNumber": "640BU",
                    "npi": "1407216880",
                    "visn": "21",
                    "ien": 64006
                },
                {
                    "name": "PALO ALTO VA DOMICILLIARY",
                    "stationNumber": "640BV",
                    "npi": "1205388394",
                    "ien": 64007,
                    "visn": "21"
                },
                {
                    "name": "SAN JOSE VA CLINIC",
                    "stationNumber": "640BY",
                    "npi": "1346283744",
                    "visn": "21",
                    "ien": 64008
                },
                {
                    "name": "CAPITOLA VA CLINIC",
                    "stationNumber": "640GA",
                    "npi": "1548213325",
                    "visn": "21",
                    "ien": 64009
                },
                {
                    "name": "SONORA VA CLINIC",
                    "stationNumber": "640GB",
                    "npi": "1346289121",
                    "visn": "21",
                    "ien": 64010
                },
                {
                    "name": "FREMONT VA CLINIC",
                    "stationNumber": "640GC",
                    "npi": "1043549215",
                    "visn": "21",
                    "ien": 64011
                },
                {
                    "name": "STOCKTON",
                    "stationNumber": "640HA",
                    "officialName": "PALO ALTO HEALTH CARE SYSTEM - STOCKTON DIVSION",
                    "npi": "1881637254",
                    "visn": "21",
                    "ien": 64012
                },
                {
                    "name": "MODESTO VA CLINIC",
                    "stationNumber": "640HB",
                    "npi": "1649218322",
                    "visn": "21",
                    "ien": 64013
                },
                {
                    "name": "MONTEREY VA CLINIC",
                    "stationNumber": "640HC",
                    "npi": "1134162688",
                    "visn": "21",
                    "ien": 64014
                },
                {
                    "name": "PAD-PRRTP",
                    "stationNumber": "640PA",
                    "officialName": "PALO ALTO HEALTH CARE SYSTEM - PAD-PRRTP DIVSION",
                    "visn": "21",
                    "ien": 64015
                },
                {
                    "name": "MPD-PRRTP",
                    "stationNumber": "640PB",
                    "officialName": "PALO ALTO HEALTH CARE SYSTEM - MPD-PRRTP DIVSION",
                    "visn": "21",
                    "ien": 64016
                },
                {
                    "name": "PALO ALTO 1 VA MOBILE CLINIC",
                    "stationNumber": "640QA",
                    "npi": "1396136263",
                    "visn": "21",
                    "ien": 64017
                },
                {
                    "name": "PALO ALTO 2 VA MOBILE CLINIC",
                    "stationNumber": "640QB",
                    "npi": "1205227170",
                    "visn": "21",
                    "ien": 64018
                },
                {
                    "name": "PALO ALTO SPINAL CORD OUTCOMES",
                    "stationNumber": "640SCI",
                    "officialName": "PALO ALTO SPINAL CORD OUTCOMES (SCI)",
                    "visn": "21",
                    "ien": 64019
                }
            ],
            "654": [
                {
                    "name": "IOANNIS A. LOUGARIS VAMC",
                    "stationNumber": "654",
                    "npi": "1639128069",
                    "visn": "21",
                    "ien": "654"
                },
                {
                    "name": "RENO VA CLC",
                    "stationNumber": "6549AA",
                    "npi": "1235581901",
                    "visn": "21",
                    "ien": 65401
                },
                {
                    "name": "NORTHERN NEVADA STATE VET HOME",
                    "stationNumber": "6549AF",
                    "visn": "21",
                    "ien": 65402
                },
                {
                    "name": "SIERRA FOOTHILLS VA CLINIC-NV",
                    "stationNumber": "654GA",
                    "npi": "1962440578",
                    "visn": "21",
                    "ien": 65403
                },
                {
                    "name": "CARSON VALLEY VA CLINIC",
                    "stationNumber": "654GB",
                    "npi": "1740227644",
                    "visn": "21",
                    "ien": 65404
                },
                {
                    "name": "LAHONTAN VALLEY VA CLINIC",
                    "stationNumber": "654GC",
                    "npi": "1811181639",
                    "visn": "21",
                    "ien": 65405
                },
                {
                    "name": "DIAMOND VIEW VA CLINIC",
                    "stationNumber": "654GD",
                    "visn": "21",
                    "ien": 65406
                },
                {
                    "name": "RENO EAST VA CLINIC",
                    "stationNumber": "654GE",
                    "npi": "1912301102",
                    "visn": "21",
                    "ien": 65407
                },
                {
                    "name": "NORTH RENO VA CLINIC",
                    "stationNumber": "654GF",
                    "npi": "1588216618",
                    "visn": "21",
                    "ien": 65408
                },
                {
                    "name": "CRAMPTON STREET VA CLINIC",
                    "stationNumber": "654QA",
                    "npi": "1841681715",
                    "visn": "21",
                    "ien": 65409
                },
                {
                    "name": "CAPITOL HILL VA CLINIC",
                    "stationNumber": "654QB",
                    "npi": "1578954442",
                    "visn": "21",
                    "ien": 65410
                },
                {
                    "name": "WINNEMUCCA VA CLINIC",
                    "stationNumber": "654QC",
                    "npi": "1639560600",
                    "visn": "21",
                    "ien": 65411
                },
                {
                    "name": "VIRGINIA STREET VA CLINIC",
                    "stationNumber": "654QD",
                    "npi": "1164979225",
                    "visn": "21",
                    "ien": 65412
                }
            ],
            "662": [
                {
                    "name": "SAN FRANCISCO VAMC",
                    "stationNumber": "662",
                    "npi": "1790735231",
                    "visn": "21",
                    "ien": "662"
                },
                {
                    "name": "SAN FRAN NHC",
                    "stationNumber": "6629AA",
                    "visn": "21",
                    "ien": 66201
                },
                {
                    "name": "SAN FRAN STNB",
                    "stationNumber": "6629AF",
                    "visn": "21",
                    "ien": 66202
                },
                {
                    "name": "SAN FRAN PHS",
                    "stationNumber": "662C5",
                    "visn": "21",
                    "ien": 66203
                },
                {
                    "name": "SAN FRAN USAH",
                    "stationNumber": "662CN",
                    "visn": "21",
                    "ien": 66204
                },
                {
                    "name": "SAN FRAN USNH",
                    "stationNumber": "662CT",
                    "visn": "21",
                    "ien": 66205
                },
                {
                    "name": "SAN FRAN USAF",
                    "stationNumber": "662CZ",
                    "visn": "21",
                    "ien": 66206
                },
                {
                    "name": "YOUNTVILLE STNB",
                    "stationNumber": "662DT",
                    "visn": "21",
                    "ien": 66207
                },
                {
                    "name": "SANTA ROSA, CA CBOC ",
                    "stationNumber": "662GA",
                    "npi": "1366480469",
                    "visn": "21",
                    "ien": 66208
                },
                {
                    "name": "VALLEJO CBOC",
                    "stationNumber": "662GB",
                    "visn": "21",
                    "ien": 66209
                },
                {
                    "name": "EUREKA VA CLINIC",
                    "stationNumber": "662GC",
                    "npi": "1669425013",
                    "visn": "21",
                    "ien": 66210
                },
                {
                    "name": "UKIAH VA CLINIC",
                    "stationNumber": "662GD",
                    "npi": "1508805409",
                    "visn": "21",
                    "ien": 66211
                },
                {
                    "name": "SAN BRUNO CBOC",
                    "stationNumber": "662GE",
                    "npi": "1033262373",
                    "visn": "21",
                    "ien": 66212
                },
                {
                    "name": "SAN FRANCISCO VA CLINIC",
                    "stationNumber": "662GF",
                    "visn": "21",
                    "ien": 66213
                },
                {
                    "name": "CLEARLAKE VA CLINIC",
                    "stationNumber": "662GG",
                    "npi": "1689976094",
                    "visn": "21",
                    "ien": 66214
                },
                {
                    "name": "SAN FRAN PRRTP",
                    "stationNumber": "662PA",
                    "visn": "21",
                    "ien": 66215
                }
            ]
        },
        "15": {
            "589": [
                {
                    "name": "VA HEARTLAND - WEST, VISN 15",
                    "stationNumber": "589",
                    "officialName": "KANSAS CITY VAMC",
                    "npi": "1447209093",
                    "domain": "KANSAS-CITY.MED.VA.GOV",
                    "visn": "15",
                    "ien": "589",
                    "mn": "KAN"
                },
                {
                    "name": "COLUMBIA, MO NURSING HOME",
                    "stationNumber": "5899AB",
                    "officialName": "COLUMBIA NURSING HOME CARE UNIT (5899AB)",
                    "visn": "15",
                    "ien": 58901
                },
                {
                    "name": "TOPEKA NHCU",
                    "stationNumber": "5899AC",
                    "visn": "15",
                    "ien": 58902
                },
                {
                    "name": "LEAVENWORTH NHCU",
                    "stationNumber": "5899AD",
                    "visn": "15",
                    "ien": 58903
                },
                {
                    "name": "WICHITA NURSING HOME CARE UNIT",
                    "stationNumber": "5899AE",
                    "visn": "15",
                    "ien": 58904
                },
                {
                    "name": "FORT DODGE STATE NURSING DOM",
                    "stationNumber": "5899AF",
                    "visn": "15",
                    "ien": 58905
                },
                {
                    "name": "COLUMBIA, MO VAMC",
                    "stationNumber": "589A4",
                    "officialName": "HARRY S. TRUMAN VAMC",
                    "npi": "1447201983",
                    "visn": "15",
                    "ien": 58906
                },
                {
                    "name": "EASTERN KS HCS TOPEKA DIV",
                    "stationNumber": "589A5",
                    "officialName": "COLMERY O NEIL VAMC",
                    "npi": "1740230226",
                    "visn": "15",
                    "ien": 58907
                },
                {
                    "name": "EASTERN KS HCS LEAVENWORTH DIV",
                    "stationNumber": "589A6",
                    "officialName": "DWIGHT D. EISENHOWER VAMC",
                    "npi": "1497704944",
                    "visn": "15",
                    "ien": 58908
                },
                {
                    "name": "ROBERT J. DOLE VAMC",
                    "stationNumber": "589A7",
                    "npi": "1609826924",
                    "visn": "15",
                    "ien": 58909
                },
                {
                    "name": "LEAVENWORTH DOMICILIARY",
                    "stationNumber": "589BU",
                    "officialName": "LEAVENWORTH VA DOMICILIARY (589BU)",
                    "visn": "15",
                    "ien": 58910
                },
                {
                    "name": "KANSAS CITY VA FACILITY DOM",
                    "stationNumber": "589BV",
                    "officialName": "KANSAS CITY VA FACILITY DOMICILIARY",
                    "visn": "15",
                    "ien": 58911
                },
                {
                    "name": "VA EASTERN KANSAS MH RRTP",
                    "stationNumber": "589BW",
                    "visn": "15",
                    "ien": 58912
                },
                {
                    "name": "HARRY S. TRUMAN DOMICILIARY",
                    "stationNumber": "589BX",
                    "npi": "1720527179",
                    "visn": "15",
                    "ien": 58913
                },
                {
                    "name": "TOPEKA (PUBLIC HOSPITAL)",
                    "stationNumber": "589DH",
                    "visn": "15",
                    "ien": 58914
                },
                {
                    "name": "COLUMBIA, MO PUBLIC HOSP",
                    "stationNumber": "589DM",
                    "officialName": "COLUMBIA NON-VA HOSPITAL (OTHER PUBLIC) (589DM)",
                    "visn": "15",
                    "ien": 58915
                },
                {
                    "name": "COLUMBIA, MO CIVILIAN HOSP",
                    "stationNumber": "589DN",
                    "officialName": "COLUMBIA NON-VA HOSPITAL (CIVIL) (589DN)",
                    "visn": "15",
                    "ien": 58916
                },
                {
                    "name": "TOPEKA (CIVILIAN HOSPITAL)",
                    "stationNumber": "589DP",
                    "visn": "15",
                    "ien": 58917
                },
                {
                    "name": "WICHITA NON-VA HOSP (CIVH)",
                    "stationNumber": "589DQ",
                    "officialName": "WICHITA (CIVIL/PUBLIC)",
                    "visn": "15",
                    "ien": 58918
                },
                {
                    "name": "COLUMBIA, MO CIVILIAN HOSP",
                    "stationNumber": "589DR",
                    "officialName": "COLUMBIA NON-VA HOSPITAL (CIVIL/PUBLIC) (589DR)",
                    "visn": "15",
                    "ien": 58919
                },
                {
                    "name": "KANSAS CITY NON-VA HOSP (CIVH)",
                    "stationNumber": "589DS",
                    "officialName": "WICHITA NON-VA HOSPITALS CIVH",
                    "visn": "15",
                    "ien": 58920
                },
                {
                    "name": "ST. JAMES STATE VETERANS HOME",
                    "stationNumber": "589DT",
                    "visn": "15",
                    "ien": 58921
                },
                {
                    "name": "CAMERON STATE VETERANS HOME",
                    "stationNumber": "589DU",
                    "visn": "15",
                    "ien": 58922
                },
                {
                    "name": "MEXICO STATE VETERANS HOME",
                    "stationNumber": "589DV",
                    "visn": "15",
                    "ien": 58923
                },
                {
                    "name": "WARRENSBURG STATE VET HOME",
                    "stationNumber": "589DW",
                    "officialName": "WARRENSBURG STATE VETERANS HOME",
                    "visn": "15",
                    "ien": 58924
                },
                {
                    "name": "KANSAS SOLDIERS HOME",
                    "stationNumber": "589DX",
                    "visn": "15",
                    "ien": 58925
                },
                {
                    "name": "KANSAS VETERANS HOME-WI",
                    "stationNumber": "589DZ",
                    "visn": "15",
                    "ien": 58926
                },
                {
                    "name": "WARRENSBURG VA CLINIC",
                    "stationNumber": "589G1",
                    "officialName": "WARRENSBURG MO CBOC",
                    "npi": "1609818459",
                    "visn": "15",
                    "ien": 58927
                },
                {
                    "name": "DODGE CITY VA CLINIC",
                    "stationNumber": "589G2",
                    "npi": "1891740361",
                    "visn": "15",
                    "ien": 58928
                },
                {
                    "name": "LIBERAL CBOC",
                    "stationNumber": "589G3",
                    "npi": "1912954942",
                    "visn": "15",
                    "ien": 58929
                },
                {
                    "name": "HAYS CBOC",
                    "stationNumber": "589G4",
                    "npi": "1326094376",
                    "visn": "15",
                    "ien": 58930
                },
                {
                    "name": "PARSONS CBOC",
                    "stationNumber": "589G5",
                    "npi": "1235171083",
                    "visn": "15",
                    "ien": 58931
                },
                {
                    "name": "HUTCHINSON CBOC",
                    "stationNumber": "589G7",
                    "npi": "1982879375",
                    "visn": "15",
                    "ien": 58932
                },
                {
                    "name": "JEFFERSON CITY, MO CBOC",
                    "stationNumber": "589G8",
                    "npi": "1134375546",
                    "visn": "15",
                    "ien": 58933
                },
                {
                    "name": "FORT RILEY CBOC",
                    "stationNumber": "589G9",
                    "npi": "1194979302",
                    "visn": "15",
                    "ien": 58934
                },
                {
                    "name": "BELTON CBOC",
                    "stationNumber": "589GB",
                    "officialName": "BELTON VA CLINIC",
                    "npi": "1184675738",
                    "visn": "15",
                    "ien": 58935
                },
                {
                    "name": "PAOLA/LOUISBURG CBOC",
                    "stationNumber": "589GC",
                    "officialName": "PAOLA VA CLINIC",
                    "npi": "1588606750",
                    "visn": "15",
                    "ien": 58936
                },
                {
                    "name": "NEVADA CBOC",
                    "stationNumber": "589GD",
                    "officialName": "NEVADA VA CLINIC",
                    "npi": "1518905140",
                    "visn": "15",
                    "ien": 58937
                },
                {
                    "name": "KIRKSVILLE CBOC",
                    "stationNumber": "589GE",
                    "officialName": "KIRKSVILLE VA CLINIC",
                    "npi": "1942256680",
                    "visn": "15",
                    "ien": 58938
                },
                {
                    "name": "FORT LEONARD WOOD CBOC",
                    "stationNumber": "589GF",
                    "officialName": "FORT LEONARD WOOD VA CLINIC",
                    "npi": "1780637918",
                    "visn": "15",
                    "ien": 58939
                },
                {
                    "name": "LAKE OF THE OZARKS VA CLINIC",
                    "stationNumber": "589GH",
                    "npi": "1508808775",
                    "visn": "15",
                    "ien": 58940
                },
                {
                    "name": "ST. JOSEPH VA CLINIC ",
                    "stationNumber": "589GI",
                    "npi": "1932141009",
                    "visn": "15",
                    "ien": 58941
                },
                {
                    "name": "WYANDOTTE COUNTY VA CLINIC",
                    "stationNumber": "589GJ",
                    "npi": "1326081845",
                    "visn": "15",
                    "ien": 58942
                },
                {
                    "name": "CHANUTE VA CLINIC",
                    "stationNumber": "589GM",
                    "npi": "1790736437",
                    "visn": "15",
                    "ien": 58943
                },
                {
                    "name": "EMPORIA VA CLINIC",
                    "stationNumber": "589GN",
                    "officialName": "EMPORIA VA CLINIC ",
                    "npi": "1730134297",
                    "visn": "15",
                    "ien": 58944
                },
                {
                    "name": "GARNETT VA CLINIC",
                    "stationNumber": "589GP",
                    "npi": "1194778936",
                    "visn": "15",
                    "ien": 58945
                },
                {
                    "name": "LT. GEN RICHARD J SEITZ CBOC",
                    "stationNumber": "589GR",
                    "officialName": "LIEUTENANT GENERAL RICHARD J. SEITZ CBOC",
                    "npi": "1487600938",
                    "visn": "15",
                    "ien": 58946
                },
                {
                    "name": "LAWRENCE VA CLINIC",
                    "stationNumber": "589GU",
                    "npi": "1326095340",
                    "visn": "15",
                    "ien": 58947
                },
                {
                    "name": "FORT SCOTT VA CLINIC",
                    "stationNumber": "589GV",
                    "npi": "1023061918",
                    "visn": "15",
                    "ien": 58948
                },
                {
                    "name": "SALINA CBOC WICHITA",
                    "stationNumber": "589GW",
                    "npi": "1235171117",
                    "visn": "15",
                    "ien": 58949
                },
                {
                    "name": "MEXICO CBOC (MO)",
                    "stationNumber": "589GX",
                    "npi": "1033156864",
                    "visn": "15",
                    "ien": 58950
                },
                {
                    "name": "ST. JAMES VA CLINIC",
                    "stationNumber": "589GY",
                    "npi": "1457394033",
                    "visn": "15",
                    "ien": 58951
                },
                {
                    "name": "CAMERON CBOC",
                    "stationNumber": "589GZ",
                    "npi": "1518918127",
                    "visn": "15",
                    "ien": 58952
                },
                {
                    "name": "KANSAS CITY MOC",
                    "stationNumber": "589HK",
                    "npi": "1942675384",
                    "visn": "15",
                    "ien": 58953
                },
                {
                    "name": "SEDALIA CBOC",
                    "stationNumber": "589JA",
                    "npi": "1720301682",
                    "visn": "15",
                    "ien": 58954
                },
                {
                    "name": "EXCELSIOR SPRINGS CBOC",
                    "stationNumber": "589JB",
                    "npi": "1891018172",
                    "visn": "15",
                    "ien": 58955
                },
                {
                    "name": "JOHNSON COUNTY VA CLINIC",
                    "stationNumber": "589JC",
                    "visn": "15",
                    "ien": 58956
                },
                {
                    "name": "MARSHFIELD CBOC",
                    "stationNumber": "589JD",
                    "npi": "1538412622",
                    "visn": "15",
                    "ien": 58957
                },
                {
                    "name": "PLATTE CITY VA CLINIC",
                    "stationNumber": "589JE",
                    "npi": "1538542188",
                    "visn": "15",
                    "ien": 58958
                },
                {
                    "name": "HONOR VA CLINIC",
                    "stationNumber": "589JF",
                    "visn": "15",
                    "ien": 58959
                },
                {
                    "name": "FORT RILEY DES CLINIC",
                    "stationNumber": "589MA",
                    "visn": "15",
                    "ien": 58960
                },
                {
                    "name": "IDES MCCONNELL AFB",
                    "stationNumber": "589MB",
                    "visn": "15",
                    "ien": 58961
                },
                {
                    "name": "IDES WHITEMAN AFB",
                    "stationNumber": "589MC",
                    "visn": "15",
                    "ien": 58962
                },
                {
                    "name": "IDES FORT LEAVENWORTH",
                    "stationNumber": "589MD",
                    "visn": "15",
                    "ien": 58963
                },
                {
                    "name": "IDES FORT LEONARD",
                    "stationNumber": "589ME",
                    "visn": "15",
                    "ien": 58964
                },
                {
                    "name": "KANSAS CITY",
                    "stationNumber": "589PA",
                    "officialName": "KANSAS CITY PSYCHOSOCIAL RESIDENTIAL REHABILITATION TREATMENT PROGRAM",
                    "visn": "15",
                    "ien": 58965
                },
                {
                    "name": "LEAVENWORTH-PRRTP",
                    "stationNumber": "589PB",
                    "visn": "15",
                    "ien": 58966
                },
                {
                    "name": "COLUMBIA PRRTP/CWT",
                    "stationNumber": "589PC",
                    "visn": "15",
                    "ien": 58967
                },
                {
                    "name": "TOPEKA PRRTP",
                    "stationNumber": "589PD",
                    "visn": "15",
                    "ien": 58968
                },
                {
                    "name": "OVERLAND PARK VA CLINIC",
                    "stationNumber": "589QA",
                    "npi": "1093104283",
                    "visn": "15",
                    "ien": 58969
                },
                {
                    "name": "SEDGWICK COUNTY VA CLINIC",
                    "stationNumber": "589QB",
                    "npi": "1912397134",
                    "visn": "15",
                    "ien": 58970
                },
                {
                    "name": "SOUTH PARKLANE VA CLINIC",
                    "stationNumber": "589QC",
                    "npi": "1649660861",
                    "visn": "15",
                    "ien": 58971
                },
                {
                    "name": "WICHITA VA MOBILE CLINIC",
                    "stationNumber": "589QD",
                    "npi": "1841680097",
                    "visn": "15",
                    "ien": 58972
                }
            ],
            "657": [
                {
                    "name": "ST. LOUIS MO VAMC-JC DIVISION",
                    "stationNumber": "657",
                    "officialName": "VA HEARTLAND-EAST, VISN 15 HCS JOHN COCHRAN MEMORIAL HOSPITAL",
                    "npi": "1720037658",
                    "domain": "ST-LOUIS.MED.VA.GOV",
                    "visn": "15",
                    "ien": "657",
                    "mn": "STL"
                },
                {
                    "name": "ST. LOUIS MO VA NHCU",
                    "stationNumber": "6579AA",
                    "officialName": "ST. LOUIS VA NURSING HOME CARE UNIT (6579AA)",
                    "visn": "15",
                    "ien": 65701
                },
                {
                    "name": "POPLAR BLUFF MO NHC",
                    "stationNumber": "6579AB",
                    "officialName": "POPLAR BLUFF NURSING HOME CARE UNIT (6579AB)",
                    "visn": "15",
                    "ien": 65702
                },
                {
                    "name": "MARION IL VA NHCU",
                    "stationNumber": "6579AC",
                    "officialName": "MARION VA NURSING HOME CARE UNIT (6579AC)",
                    "visn": "15",
                    "ien": 65703
                },
                {
                    "name": "MARION IL STATE NHCU",
                    "stationNumber": "6579AF",
                    "officialName": "MARION STATE NURSING HOME CARE UNIT (6579AF)",
                    "visn": "15",
                    "ien": 65704
                },
                {
                    "name": "ST. LOUIS MO VAMC-JB DIVISION",
                    "stationNumber": "657A0",
                    "officialName": "VA HEARTLAND-EAST, VISN 15 HCS JEFFERSON BARRACKS",
                    "npi": "1275582710",
                    "visn": "15",
                    "ien": 65705
                },
                {
                    "name": "POPLAR BLUFF MO VAMC",
                    "stationNumber": "657A4",
                    "officialName": "JOHN PERSHING VAMC",
                    "npi": "1972554566",
                    "visn": "15",
                    "ien": 65706
                },
                {
                    "name": "MARION IL VAMC",
                    "stationNumber": "657A5",
                    "officialName": "MARION VA MEDICAL CENTER",
                    "npi": "1710936174",
                    "visn": "15",
                    "ien": 65707
                },
                {
                    "name": "ST. LOUIS VA DOMICILIARY",
                    "stationNumber": "657BU",
                    "officialName": "ST LOUIS MO VA DOMICILIARY (657BU)",
                    "visn": "15",
                    "ien": 65708
                },
                {
                    "name": "MARION IL MH-RRTP",
                    "stationNumber": "657BW",
                    "visn": "15",
                    "ien": 65709
                },
                {
                    "name": "ST. LOUIS VA PHARM/DDTU MO PHA",
                    "stationNumber": "657CA",
                    "officialName": "ST. LOUIS VA PHARMACY/DDTU (657CA)",
                    "visn": "15",
                    "ien": 65710
                },
                {
                    "name": "ST. LOUIS VA PHARMACY/NUCL MED",
                    "stationNumber": "657CB",
                    "visn": "15",
                    "ien": 65711
                },
                {
                    "name": "POPLAR BLUFF NON-VA HOSP (CIV)",
                    "stationNumber": "657DN",
                    "officialName": "POPLAR BLUFF NON-VA HOSPITAL (657DN) (CIVH)",
                    "visn": "15",
                    "ien": 65712
                },
                {
                    "name": "MARION HOSPITAL (CIVIL/PUBLIC)",
                    "stationNumber": "657DP",
                    "officialName": "MARION IL HOSPITAL (657DP) (CIVIL/PUBLIC)",
                    "visn": "15",
                    "ien": 65713
                },
                {
                    "name": "ST. LOUIS NON VA HOSP (CIVH)",
                    "stationNumber": "657DS",
                    "officialName": "ST. LOUIS NON-VA HOSPITAL (CIVIL/PUBLIC)",
                    "visn": "15",
                    "ien": 65714
                },
                {
                    "name": "MT. VERNON STATE HOME",
                    "stationNumber": "657DU",
                    "officialName": "MT. VERNON MO STATE HOME (657DU)",
                    "visn": "15",
                    "ien": 65715
                },
                {
                    "name": "ST. LOUIS MO VET STATE HOME",
                    "stationNumber": "657DW",
                    "officialName": "ST. LOUIS MISSOURI VETERANS STATE HOME (657DW)",
                    "visn": "15",
                    "ien": 65716
                },
                {
                    "name": "CAPE GIRARDEAU MO STATE HOME",
                    "stationNumber": "657DX",
                    "officialName": "CAPE GIRARDEAU MO STATE HOME (657DX)",
                    "visn": "15",
                    "ien": 65717
                },
                {
                    "name": "MISSOURI STATE VET HOME-ST LOU",
                    "stationNumber": "657DY",
                    "visn": "15",
                    "ien": 65718
                },
                {
                    "name": "ILLINOIS STATE VETERANS HOME",
                    "stationNumber": "657DZ",
                    "officialName": "ILLINOIS STATE VETERANS HOME (657DZ)",
                    "visn": "15",
                    "ien": 65719
                },
                {
                    "name": "WESTERN KENTUCKY SVH",
                    "stationNumber": "657EA",
                    "visn": "15",
                    "ien": 65720
                },
                {
                    "name": "ST. CLAIR CNTY VA CLINIC",
                    "stationNumber": "657GA",
                    "npi": "1154374429",
                    "visn": "15",
                    "ien": 65721
                },
                {
                    "name": "ST. LOUIS MO CBOC",
                    "stationNumber": "657GB",
                    "officialName": "ST LOUIS COUNTY VA CLINIC",
                    "npi": "1518155878",
                    "visn": "15",
                    "ien": 65722
                },
                {
                    "name": "ST. CHARLES MO CBOC",
                    "stationNumber": "657GD",
                    "officialName": "ST CHARLES VA CLINIC",
                    "npi": "1083654453",
                    "visn": "15",
                    "ien": 65723
                },
                {
                    "name": "WEST PLAINS MO CBOC",
                    "stationNumber": "657GF",
                    "officialName": "WEST PLAINS VA CLINIC",
                    "npi": "1639113798",
                    "visn": "15",
                    "ien": 65724
                },
                {
                    "name": "PARAGOULD CBOC",
                    "stationNumber": "657GG",
                    "officialName": "PARAGOULD VA CLINIC",
                    "npi": "1508809948",
                    "visn": "15",
                    "ien": 65725
                },
                {
                    "name": "CAPE GIRARDEAU MO CBOC",
                    "stationNumber": "657GH",
                    "officialName": "CAPE GIRARDEAU VA CLINIC",
                    "npi": "1073563474",
                    "visn": "15",
                    "ien": 65726
                },
                {
                    "name": "FARMINGTON MO CBOC",
                    "stationNumber": "657GI",
                    "officialName": "FARMINTON VA CLINIC",
                    "npi": "1992745277",
                    "visn": "15",
                    "ien": 65727
                },
                {
                    "name": "EVANSVILLE CBOC",
                    "stationNumber": "657GJ",
                    "officialName": "EVANSVILLE VA CLINIC",
                    "npi": "1952356115",
                    "visn": "15",
                    "ien": 65728
                },
                {
                    "name": "MT. VERNON IL CBOC",
                    "stationNumber": "657GK",
                    "officialName": "MT. VERNON IL CBOC (657GK)",
                    "npi": "1902840622",
                    "visn": "15",
                    "ien": 65729
                },
                {
                    "name": "PADUCAH KY CBOC",
                    "stationNumber": "657GL",
                    "officialName": "PADUCAH KY CBOC (657GL)",
                    "npi": "1356384747",
                    "visn": "15",
                    "ien": 65730
                },
                {
                    "name": "EFFINGHAM VA CLINIC",
                    "stationNumber": "657GM",
                    "npi": "1629021035",
                    "visn": "15",
                    "ien": 65731
                },
                {
                    "name": "SALEM MO CBOC",
                    "stationNumber": "657GN",
                    "officialName": "SALEM VA CLINIC",
                    "npi": "1033151980",
                    "visn": "15",
                    "ien": 65732
                },
                {
                    "name": "HANSON KY CBOC",
                    "stationNumber": "657GO",
                    "officialName": "HANSON VA CLINIC",
                    "npi": "1578616892",
                    "visn": "15",
                    "ien": 65733
                },
                {
                    "name": "OWENSBORO KY CBOC",
                    "stationNumber": "657GP",
                    "officialName": "OWENSBORO VA CLINIC",
                    "npi": "1407005481",
                    "visn": "15",
                    "ien": 65734
                },
                {
                    "name": "VINCENNES IN CBOC",
                    "stationNumber": "657GQ",
                    "officialName": "VINCENNES VA CLINIC",
                    "npi": "1760631220",
                    "visn": "15",
                    "ien": 65735
                },
                {
                    "name": "MAYFIELD KY CBOC",
                    "stationNumber": "657GR",
                    "officialName": "MAYFIELD VA CLINIC",
                    "visn": "15",
                    "ien": 65736
                },
                {
                    "name": "WASHINGTON CBOC",
                    "stationNumber": "657GS",
                    "officialName": "WASHINGTON VA CLINIC",
                    "visn": "15",
                    "ien": 65737
                },
                {
                    "name": "CARBONDALE CBOC",
                    "stationNumber": "657GT",
                    "npi": "1992067995",
                    "visn": "15",
                    "ien": 65738
                },
                {
                    "name": "HARRISBURG CBOC",
                    "stationNumber": "657GU",
                    "officialName": "HARRISBURG VA CLINIC",
                    "npi": "1497012652",
                    "visn": "15",
                    "ien": 65739
                },
                {
                    "name": "SIKESTON CBOC",
                    "stationNumber": "657GV",
                    "npi": "1497988620",
                    "visn": "15",
                    "ien": 65740
                },
                {
                    "name": "POCAHONTAS VA CLINIC",
                    "stationNumber": "657GW",
                    "npi": "1033512983",
                    "visn": "15",
                    "ien": 65741
                },
                {
                    "name": "ST. LOUIS VA CLINIC",
                    "stationNumber": "657GX",
                    "npi": "1972910917",
                    "visn": "15",
                    "ien": 65742
                },
                {
                    "name": "MANCHESTER VA CLINIC",
                    "stationNumber": "657GY",
                    "npi": "1235546276",
                    "visn": "15",
                    "ien": 65743
                },
                {
                    "name": "IDES SCOTT AFB",
                    "stationNumber": "657MA",
                    "visn": "15",
                    "ien": 65744
                },
                {
                    "name": "MARION IL PRRTP",
                    "stationNumber": "657PC",
                    "officialName": "MARION PSYCHOSOCIAL RESIDENTIAL REHABILITATION TREATMENT PROGRAM",
                    "npi": "1962862342",
                    "visn": "15",
                    "ien": 65745
                },
                {
                    "name": "ST. LOUIS SARRTP",
                    "stationNumber": "657PD",
                    "npi": "1871953257",
                    "visn": "15",
                    "ien": 65746
                },
                {
                    "name": "OLIVE STREET VA CLINIC",
                    "stationNumber": "657QA",
                    "npi": "1396135513",
                    "visn": "15",
                    "ien": 65747
                },
                {
                    "name": "WASHINGTON BOULEVARD VA CLINIC",
                    "stationNumber": "657QB",
                    "npi": "1114317369",
                    "visn": "15",
                    "ien": 65748
                },
                {
                    "name": "ENTERPRISE WAY VA CLINIC",
                    "stationNumber": "657QC",
                    "npi": "1588054738",
                    "visn": "15",
                    "ien": 65749
                },
                {
                    "name": "HEARTLAND STREET VA CLINIC",
                    "stationNumber": "657QD",
                    "npi": "1558751776",
                    "visn": "15",
                    "ien": 65750
                },
                {
                    "name": "SCOTT AFB VA CLINIC",
                    "stationNumber": "657QE",
                    "npi": "1093192908",
                    "visn": "15",
                    "ien": 65751
                },
                {
                    "name": "ST LOUIS SPINAL CORD OUTCOMES",
                    "stationNumber": "657SCI",
                    "officialName": "ST. LOUIS MO - JC DIVISION SPINAL CORD OUTCOMES (SCI)",
                    "visn": "15",
                    "ien": 65752
                }
            ]
        },
        "9": {
            "596": [
                {
                    "name": "LEXINGTON VAMC-LEESTOWN",
                    "stationNumber": "596",
                    "npi": "1871684639",
                    "domain": "LEXINGTON.MED.VA.GOV",
                    "visn": "9",
                    "ien": "596",
                    "mn": "LEX"
                },
                {
                    "name": "LEXINGTON NHCU",
                    "stationNumber": "5969AA",
                    "visn": "9",
                    "ien": 59601
                },
                {
                    "name": "LEXINGTON-CDD VAMC",
                    "stationNumber": "596A4",
                    "officialName": "LEXINGTON VETERANS AFFAIRS MEDICAL CENTER",
                    "npi": "1255380242",
                    "visn": "9",
                    "ien": 59602
                },
                {
                    "name": "LEXINGTON VA FACILITY DOM",
                    "stationNumber": "596BU",
                    "officialName": "LEXINGTON VA FACILITY DOMICILIARY",
                    "visn": "9",
                    "ien": 59603
                },
                {
                    "name": "WILMORE STATE HOME",
                    "stationNumber": "596DT",
                    "officialName": "THOMPSON HOOD VETERANS CENTER",
                    "visn": "9",
                    "ien": 59604
                },
                {
                    "name": "EASTERN KENTUCKY VET CENTER",
                    "stationNumber": "596DU",
                    "officialName": "EASTERN KENTUCKY VETERANS CENTER",
                    "visn": "9",
                    "ien": 59605
                },
                {
                    "name": "SOMERSET VA CLINIC",
                    "stationNumber": "596GA",
                    "npi": "1003855966",
                    "visn": "9",
                    "ien": 59606
                },
                {
                    "name": "MOREHEAD VA CLINIC",
                    "stationNumber": "596GB",
                    "npi": "1215115696",
                    "visn": "9",
                    "ien": 59607
                },
                {
                    "name": "HAZARD VA CLINIC",
                    "stationNumber": "596GC",
                    "npi": "1689836637",
                    "visn": "9",
                    "ien": 59608
                },
                {
                    "name": "BEREA VA CLINIC",
                    "stationNumber": "596GD",
                    "npi": "1528205648",
                    "visn": "9",
                    "ien": 59609
                },
                {
                    "name": "LEXINGTON MOBILE MEDIC",
                    "stationNumber": "596HA",
                    "officialName": "LEXINGTON MOBILE OUTREACH CLINIC",
                    "npi": "1154365898",
                    "visn": "9",
                    "ien": 59610
                },
                {
                    "name": "LEXINGTON PRRTP",
                    "stationNumber": "596PA",
                    "visn": "9",
                    "ien": 59611
                },
                {
                    "name": "CUMBERLAND RIVER VA CLINIC",
                    "stationNumber": "596QA",
                    "npi": "1144618406",
                    "visn": "9",
                    "ien": 59612
                }
            ],
            "603": [
                {
                    "name": "ROBLEY REX VAMC",
                    "stationNumber": "603",
                    "npi": "1144279597",
                    "domain": "LOUISVILLE.MED.VA.GOV",
                    "visn": "9",
                    "ien": "603",
                    "mn": "LOU"
                },
                {
                    "name": "LOUISVILLE (CNH)",
                    "stationNumber": "6039AA",
                    "visn": "9",
                    "ien": 60301
                },
                {
                    "name": "LOUISVILLE VA FACILITY DOM",
                    "stationNumber": "603BU",
                    "officialName": "LOUISVILLE VA FACILITY DOMICILIARY",
                    "visn": "9",
                    "ien": 60302
                },
                {
                    "name": "LOUISVILLE, KY (USAH)",
                    "stationNumber": "603CN",
                    "officialName": "Ireland Army Community Hospital",
                    "visn": "9",
                    "ien": 60303
                },
                {
                    "name": "FORT KNOX, KY CBOC",
                    "stationNumber": "603GA",
                    "officialName": "VA Healthcare Center, Ft. Knox",
                    "npi": "1699728824",
                    "visn": "9",
                    "ien": 60304
                },
                {
                    "name": "NEW ALBANY VA CLINIC",
                    "stationNumber": "603GB",
                    "npi": "1144267071",
                    "visn": "9",
                    "ien": 60305
                },
                {
                    "name": "SHIVELY VA CLINIC",
                    "stationNumber": "603GC",
                    "npi": "1851347124",
                    "visn": "9",
                    "ien": 60306
                },
                {
                    "name": "STONYBROOK VA CLINIC",
                    "stationNumber": "603GD",
                    "npi": "1518910025",
                    "visn": "9",
                    "ien": 60307
                },
                {
                    "name": "NEWBURG VA CLINIC",
                    "stationNumber": "603GE",
                    "npi": "1750320735",
                    "visn": "9",
                    "ien": 60308
                },
                {
                    "name": "GRAYSON COUNTY VA CLINIC",
                    "stationNumber": "603GF",
                    "npi": "1063668291",
                    "visn": "9",
                    "ien": 60309
                },
                {
                    "name": "SCOTT COUNTY VA CLINIC",
                    "stationNumber": "603GG",
                    "npi": "1770732752",
                    "visn": "9",
                    "ien": 60310
                },
                {
                    "name": "CARROLLTON VA CLINIC",
                    "stationNumber": "603GH",
                    "npi": "1508015074",
                    "visn": "9",
                    "ien": 60311
                },
                {
                    "name": "IDES FORT KNOX",
                    "stationNumber": "603MA",
                    "visn": "9",
                    "ien": 60312
                },
                {
                    "name": "LOUISVILLE SARRTP",
                    "stationNumber": "603PB",
                    "npi": "1649630955",
                    "visn": "9",
                    "ien": 60313
                }
            ],
            "614": [
                {
                    "name": "MEMPHIS VA MEDICAL CENTER",
                    "stationNumber": "614",
                    "npi": "1255381695",
                    "domain": "MEMPHIS.MED.VA.GOV",
                    "visn": "9",
                    "ien": "614",
                    "mn": "MEM"
                },
                {
                    "name": "MEMPHIS VA FACILITY DOM",
                    "stationNumber": "614BU",
                    "officialName": "MEMPHIS VA FACILITY DOMICILIARY",
                    "visn": "9",
                    "ien": 61401
                },
                {
                    "name": "MEMPHIS PUB HOSP",
                    "stationNumber": "614DM",
                    "officialName": "MEMPHIS, TN PUB HOSP",
                    "visn": "9",
                    "ien": 61402
                },
                {
                    "name": "MEMPHIS CIVILIAN HOSP",
                    "stationNumber": "614DS",
                    "officialName": "MEMPHIS, TN CIVILIAN HOSP",
                    "visn": "9",
                    "ien": 61403
                },
                {
                    "name": "TENNESSEE STATE VETHOME-HMBOLT",
                    "stationNumber": "614DT",
                    "visn": "9",
                    "ien": 61404
                },
                {
                    "name": "MISSISSIPPI STATE VET HOME",
                    "stationNumber": "614DU",
                    "officialName": "MISSISSIPPI STATE VETERANS HOME",
                    "visn": "9",
                    "ien": 61405
                },
                {
                    "name": "TUPELO CBOC",
                    "stationNumber": "614GA",
                    "npi": "1073552436",
                    "visn": "9",
                    "ien": 61406
                },
                {
                    "name": "JONESBORO VA CLINIC",
                    "stationNumber": "614GB",
                    "npi": "1619923398",
                    "visn": "9",
                    "ien": 61407
                },
                {
                    "name": "MARSHALL COUNTY VA CLINIC",
                    "stationNumber": "614GC",
                    "officialName": "HOLLY SPRINGS VA CLINIC",
                    "npi": "1194778357",
                    "visn": "9",
                    "ien": 61408
                },
                {
                    "name": "SAVANNAH VA CLINIC",
                    "stationNumber": "614GD",
                    "npi": "1669427050",
                    "visn": "9",
                    "ien": 61409
                },
                {
                    "name": "COVINGTON VA CLINIC",
                    "stationNumber": "614GE",
                    "npi": "1154374726",
                    "visn": "9",
                    "ien": 61410
                },
                {
                    "name": "MEMPHIS SOUTH VA CLINIC",
                    "stationNumber": "614GF",
                    "officialName": "NONCONNAH BOULEVARD VA CLINIC",
                    "npi": "1003859588",
                    "visn": "9",
                    "ien": 61411
                },
                {
                    "name": "JACKSON VA CLINIC",
                    "stationNumber": "614GG",
                    "npi": "1205072436",
                    "visn": "9",
                    "ien": 61412
                },
                {
                    "name": "BOLIVAR CBOC",
                    "stationNumber": "614GH",
                    "npi": "1871723452",
                    "visn": "9",
                    "ien": 61413
                },
                {
                    "name": "DYERSBURG VA CLINIC",
                    "stationNumber": "614GI",
                    "npi": "1356572010",
                    "visn": "9",
                    "ien": 61414
                },
                {
                    "name": "HELENA VA CLINIC",
                    "stationNumber": "614GN",
                    "npi": "1649401316",
                    "visn": "9",
                    "ien": 61415
                },
                {
                    "name": "IDES MILLINGTON NAVAL BASE",
                    "stationNumber": "614MA",
                    "visn": "9",
                    "ien": 61416
                },
                {
                    "name": "MEM-PRRTP",
                    "stationNumber": "614PA",
                    "officialName": "MEMPHIS, TN PRRTP",
                    "visn": "9",
                    "ien": 61417
                },
                {
                    "name": "UNION AVENUE VA CLINIC",
                    "stationNumber": "614QA",
                    "visn": "9",
                    "ien": 61418
                },
                {
                    "name": "MEMPHIS SPINAL CORD OUTCOMES",
                    "stationNumber": "614SCI",
                    "officialName": "MEMPHIS SPINAL CORD OUTCOMES (SCI)",
                    "visn": "9",
                    "ien": 61419
                }
            ],
            "621": [
                {
                    "name": "JAMES H. QUILLEN VAMC",
                    "stationNumber": "621",
                    "npi": "1831148253",
                    "domain": "MTN-HOME.MED.VA.GOV",
                    "visn": "9",
                    "ien": "621",
                    "mn": "MTN"
                },
                {
                    "name": "MOUNTAIN HOME NHC",
                    "stationNumber": "6219AA",
                    "visn": "9",
                    "ien": 62101
                },
                {
                    "name": "MOUNTAIN HOME DOM",
                    "stationNumber": "621BU",
                    "visn": "9",
                    "ien": 62102
                },
                {
                    "name": "WILLIAM C. TALLENT VA OPC",
                    "stationNumber": "621BY",
                    "npi": "1235316928",
                    "visn": "9",
                    "ien": 62103
                },
                {
                    "name": "BEN ATCHLEY STATE VET HOME",
                    "stationNumber": "621DT",
                    "visn": "9",
                    "ien": 62104
                },
                {
                    "name": "ROGERSVILLE VA CLINIC",
                    "stationNumber": "621GA",
                    "npi": "1700824646",
                    "visn": "9",
                    "ien": 62105
                },
                {
                    "name": "NORTON VA CLINIC",
                    "stationNumber": "621GC",
                    "npi": "1841237385",
                    "visn": "9",
                    "ien": 62106
                },
                {
                    "name": "MORRISTOWN VA CLINIC",
                    "stationNumber": "621GG",
                    "npi": "1639342231",
                    "visn": "9",
                    "ien": 62107
                },
                {
                    "name": "DANNIE A. CARR VETERANS OPC",
                    "stationNumber": "621GI",
                    "npi": "1891059945",
                    "visn": "9",
                    "ien": 62108
                },
                {
                    "name": "BRISTOL VA CLINIC",
                    "stationNumber": "621GJ",
                    "npi": "1467601922",
                    "visn": "9",
                    "ien": 62109
                },
                {
                    "name": "CAMPBELL COUNTY VA CLINIC",
                    "stationNumber": "621GK",
                    "npi": "1285046722",
                    "visn": "9",
                    "ien": 62110
                },
                {
                    "name": "MOUNTAIN CITY VA CLINIC",
                    "stationNumber": "621GO",
                    "npi": "1053942474",
                    "visn": "9",
                    "ien": 62111
                },
                {
                    "name": "JONESVILLE VA CLINIC",
                    "stationNumber": "621QA",
                    "npi": "1316335680",
                    "visn": "9",
                    "ien": 62112
                },
                {
                    "name": "MARION VA CLINIC",
                    "stationNumber": "621QB",
                    "npi": "1497143531",
                    "visn": "9",
                    "ien": 62113
                },
                {
                    "name": "VANSANT VA CLINIC",
                    "stationNumber": "621QC",
                    "npi": "1306234448",
                    "visn": "9",
                    "ien": 62114
                },
                {
                    "name": "KNOX COUNTY VA CLINIC",
                    "stationNumber": "621QD",
                    "npi": "1417345554",
                    "visn": "9",
                    "ien": 62115
                },
                {
                    "name": "DOWNTOWN WEST VA CLINIC",
                    "stationNumber": "621QE",
                    "npi": "1922544394",
                    "visn": "9",
                    "ien": 62116
                },
                {
                    "name": "JOHNSON CITY VA CLINIC",
                    "stationNumber": "621QF",
                    "npi": "1982236907",
                    "visn": "9",
                    "ien": 62117
                }
            ],
            "626": [
                {
                    "name": "TENNESSEE VALLEY HCS",
                    "stationNumber": "626",
                    "npi": "1700835154",
                    "domain": "TENNESSEEVALLEY.MED.VA.GOV",
                    "visn": "9",
                    "ien": "626",
                    "mn": "TEN"
                },
                {
                    "name": "MURFREESBORO VANHC",
                    "stationNumber": "6269AB",
                    "visn": "9",
                    "ien": 62601
                },
                {
                    "name": "TENNESSEE STATE VETHOME-CLRKVL",
                    "stationNumber": "6269AG",
                    "visn": "9",
                    "ien": 62602
                },
                {
                    "name": "ALVIN C. YORK VAMC",
                    "stationNumber": "626A4",
                    "npi": "1851340269",
                    "visn": "9",
                    "ien": 62603
                },
                {
                    "name": "MURFREESBORO DOM",
                    "stationNumber": "626BU",
                    "officialName": "MURFREESBORO VA DOMICILIARY",
                    "visn": "9",
                    "ien": 62604
                },
                {
                    "name": "MIDDLE TENNESSEE HCS",
                    "stationNumber": "626BV",
                    "visn": "9",
                    "ien": 62605
                },
                {
                    "name": "FT. CAMPBELL ARMY HOSPITAL",
                    "stationNumber": "626CN",
                    "visn": "9",
                    "ien": 62606
                },
                {
                    "name": "TENNESSEE STATE VETHOME-M'BORO",
                    "stationNumber": "626DT",
                    "visn": "9",
                    "ien": 62607
                },
                {
                    "name": "DOVER VA CLINIC",
                    "stationNumber": "626GA",
                    "npi": "1841243300",
                    "visn": "9",
                    "ien": 62608
                },
                {
                    "name": "BOWLING GREEN",
                    "stationNumber": "626GC",
                    "officialName": "BOWLING GREEN COMMUNITY BASED OUTPATIENT CLINIC",
                    "npi": "1639120215",
                    "visn": "9",
                    "ien": 62609
                },
                {
                    "name": "CLARKSVILLE VA CLINIC",
                    "stationNumber": "626GE",
                    "npi": "1205888336",
                    "visn": "9",
                    "ien": 62610
                },
                {
                    "name": "CHATTANOOGA VA CLINIC",
                    "stationNumber": "626GF",
                    "npi": "1295903466",
                    "visn": "9",
                    "ien": 62611
                },
                {
                    "name": "TULLAHOMA",
                    "stationNumber": "626GG",
                    "officialName": "TULLAHOMA COMMUNITY BASED OUTPATIENT CLINIC",
                    "npi": "1326082546",
                    "visn": "9",
                    "ien": 62612
                },
                {
                    "name": "COOKEVILLE VA CLINIC",
                    "stationNumber": "626GH",
                    "npi": "1952353062",
                    "visn": "9",
                    "ien": 62613
                },
                {
                    "name": "HOPKINSVILLE VA CLINIC",
                    "stationNumber": "626GJ",
                    "npi": "1972739936",
                    "visn": "9",
                    "ien": 62614
                },
                {
                    "name": "MCMINNVILLE VA CLINIC",
                    "stationNumber": "626GK",
                    "npi": "1073749032",
                    "visn": "9",
                    "ien": 62615
                },
                {
                    "name": "ROANE COUNTY VA CLINIC",
                    "stationNumber": "626GL",
                    "npi": "1316267743",
                    "visn": "9",
                    "ien": 62616
                },
                {
                    "name": "MAURY COUNTY VA CLINIC",
                    "stationNumber": "626GM",
                    "npi": "1386997880",
                    "visn": "9",
                    "ien": 62617
                },
                {
                    "name": "ATHENS VA CLINIC",
                    "stationNumber": "626GN",
                    "npi": "1083079396",
                    "visn": "9",
                    "ien": 62618
                },
                {
                    "name": "INTERNATIONAL PLAZA VA CLINIC",
                    "stationNumber": "626GO",
                    "npi": "1073060133",
                    "visn": "9",
                    "ien": 62619
                },
                {
                    "name": "GALLATIN VA CLINIC",
                    "stationNumber": "626GP",
                    "npi": "1730659061",
                    "visn": "9",
                    "ien": 62620
                },
                {
                    "name": "IDES FORT CAMPBELL",
                    "stationNumber": "626MA",
                    "visn": "9",
                    "ien": 62621
                },
                {
                    "name": "MURFREESBORO PRRTP",
                    "stationNumber": "626PA",
                    "officialName": "MURFREESBORO PSYCHOSOCIAL RESIDENTIAL REHABILITATION TREATMENT PROGRAM",
                    "visn": "9",
                    "ien": 62622
                },
                {
                    "name": "ALBION STREET VA CLINIC",
                    "stationNumber": "626QA",
                    "npi": "1841688983",
                    "visn": "9",
                    "ien": 62623
                },
                {
                    "name": "CHARLOTTE AVENUE VA CLINIC",
                    "stationNumber": "626QB",
                    "npi": "1750779898",
                    "visn": "9",
                    "ien": 62624
                },
                {
                    "name": "POINTE CENTRE VA CLINIC",
                    "stationNumber": "626QC",
                    "npi": "1962958256",
                    "visn": "9",
                    "ien": 62625
                },
                {
                    "name": "GLENIS DRIVE VA CLINIC",
                    "stationNumber": "626QD",
                    "npi": "1639520281",
                    "visn": "9",
                    "ien": 62626
                },
                {
                    "name": "GLENIS DRIVE 2 VA CLINIC",
                    "stationNumber": "626QE",
                    "npi": "1447883095",
                    "visn": "9",
                    "ien": 62627
                },
                {
                    "name": "DALTON DRIVE VA CLINIC",
                    "stationNumber": "626QF",
                    "npi": "1073146627",
                    "visn": "9",
                    "ien": 62628
                }
            ]
        }
}
